package Agromodel::Loaders::Base;

use strict;
use Cwd;
use Parallel::ForkManager;
use Sys::CPU;
use Time::HiRes;

sub getProjectHome($) {
	my $pkg = shift;

        # 1. Get current directory
        my $cwd = Cwd::getcwd();

        # 2. Get script name
        my $script_name = $0;
        $script_name =~ s/^(\.*)//g;
        $script_name = $cwd . $script_name;

        # 3. Find project home based on script location
        my $index = index($script_name, "/input/Scripts");
        return substr($script_name, 0, $index);
}

sub getConfig() {
	my $pkg = shift;

        my $projectHome = $pkg->getProjectHome();
        my $configFile = $projectHome . "/config/init.properties";
        if (open(CONFIG, "<$configFile")) {
                my $config = {};
                while (<CONFIG>) {
                        my $line = $_;
                        if ($line !~ /^\s*#/) {
                                if ($line =~ /^\s*(.+?)=(.+?)\r*\n*$/) {
                                        $config->{$1} = $2;
                                }
                        }
                }

                close(CONFIG);
                return $config;
        } else {
                warn "Cannot open config file $configFile: $!";
                return undef();
        }
}

sub emptyDirectory($$) {
	my ($self, $directory) = @_;

        if (not opendir(DIRECTORY, $directory)) {
                warn "Cannot open directory $directory: $!";
                return 0;
        }

        my @files = readdir(DIRECTORY);
        foreach my $file (@files) {
                if ($file !~ /^\./) {
                        my $fullPath = $directory . "/" . $file;
			if (-d $fullPath) {
				$self->emptyDirectory($fullPath);
				if (! rmdir($fullPath)) {
					warn "Cannot delete directory $fullPath: $!";
				}
			} elsif (! unlink($fullPath)) {
                                warn "Cannot delete file $fullPath: $!";
                        }
                }
        }

        closedir(DIRECTORY);
        return 1;
}

sub executeQueryCmd($$) {
        my ($self, $cmd)= @_;
        my ($error, $warning) = (undef(), undef());

        if (open(FH, "$cmd|")) {
                my $cmdOutput = "";
                while (<FH>) {
                        $cmdOutput .= $_;
                }
                close(FH);

                if ($cmdOutput) {
                        $warning = "Warnings while executing command $cmd: " . $cmdOutput;
                }
        } else {
                $error = "Error executing command $cmd: $!";
        }

        return ($error, $warning);
}

sub logInfo($$) {
	my ($self, $message) = @_;
	print "$message... \n";	
}

sub new($) {
	my $pkg = shift;
	die "Cannot instantiate abstract class $pkg";
}

sub run($) {
	my $self = shift;

	# Start timers
	my $time = [ Time::HiRes::gettimeofday() ];

	# Initialization handlers
	$self->runHandlers("Init", $self->{init_handlers});

	# Fork handlers
	my $forkHandlers = $self->{fork_handlers};
	if (defined($forkHandlers)) {
		my $processes = Sys::CPU::cpu_count();
		my $pm = Parallel::ForkManager->new($processes);
		foreach my $element (@{$self->{fork_elements}}) {
			my $pid = $pm->start() and next();

			foreach my $handlerMethodName (@$forkHandlers) {
				my $result = $self->$handlerMethodName($element);
				if (! $result) {
					die("Fork handler execution failed");
				}
			}
			$self->report("Fork");

			$pm->finish();
		}

		$pm->wait_all_children();
	}

	# Join handlers
	$self->runHandlers("Join", $self->{join_handlers});

	# End timers
	my $wallclock_time = Time::HiRes::tv_interval($time);
	my @times = times();
	my $process_time = 0.0;
	foreach my $t (@times) {
		$process_time += $t;
	}
	$self->logInfo("Wallclock time = $wallclock_time; Process time = $process_time; Parallelization ratio = " . ($process_time / $wallclock_time));
}

sub runHandlers($$$) {
	my ($self, $handlersType, $handlers) = @_;	

	if (defined($handlers)) {
		foreach my $handlerMethodName (@$handlers) {
			my $result = $self->$handlerMethodName();
			if (! $result) {
				die("$handlersType handler $handlerMethodName execution failed");
			}
		}
		
		$self->report($handlersType);
	}
}

sub report($$) {
	my ($self, $handlersType) = @_;

	if (defined($self->{errors}) and defined($self->{warnings})) {
		if (@{$self->{errors}}) {
			$self->logInfo("$handlersType handlers execution failed:\n" . join("\n", @{$self->{errors}}));
		} elsif (@{$self->{warnings}}) {
			$self->logInfo("$handlersType handlers execution finished with warnings...\n" . join("\n", @{$self->{warnings}}));
		} elsif ($handlersType ne "Fork") {
			$self->logInfo("$handlersType handlers execution was successful!");
		}
	} elsif ($handlersType ne "Fork") {
		$self->logInfo("$handlersType handlers execution was successful!");
	}

	$self->{errors}   = undef();
	$self->{warnings} = undef();
}

1;
