package Agromodel::Loaders::CommonTables;

use base qw(Agromodel::Loaders::Base);
use strict;
use Archive::Zip;
use Cwd;

sub new($) {
	my $pkg  = shift;
	my $func = $pkg . "::new";

	my $tables = {
		activity                        =>      1,
		activity_group                  =>      1,
		activitymgmt_inputs             =>      1,
		activitymgmt_labors             =>      1,
		activitymgmt_specifications     =>      1,
		agvarcostinfo                   =>      1,
		context_type                    =>      1,
		context_type_status             =>      1,
		context_type_adjustment         =>      1,
		creagroup                       =>      1,
		crearegion                      =>      1,
		inputtype                       =>      1,
		innovatortype                   =>      1,
		labortype                       =>      1,
		parameter                       =>      1,
		region                          =>      1,
		rotatortype                     =>      1,
		soiltype                        =>      1,
	};
	my $config = $pkg->getConfig();
	if (not defined $config) {
		warn "$func: Undefined configuration object";
		return undef();
	}

	my $data = {
		config	=>	$config,
		tables	=>	$tables,
		init_handlers	=>	[
			'validateInputFile',
			'decompress',
			'validateInputFileContent',
			'load',
			'cleanup'
		]
	};

	return bless($data, $pkg);
}

sub validateInputFile($) {
	my $self = shift;
	my $func = ref($self) . "::validateInputFile";
	$self->logInfo("Validating common tables file");	

	my $currentDir = getcwd();
	my $commonTablesDir = $currentDir . "/CommonTables";
	if (not opendir(COMMON_TABLES_DIR, $commonTablesDir)) {
		warn "$func: Cannot open common tables dir ($commonTablesDir): $!";
		return 0;
	} else {
		$self->{commonTablesDir} = $commonTablesDir;
		closedir(COMMON_TABLES_DIR);
	}

	my $commonTablesFile = $commonTablesDir . "/CommonTables.zip";
	if (not (-e $commonTablesFile)) {
		warn "$func: Common tables file ($commonTablesFile) not found";
		return 0;
	} else {
		$self->{commonTablesFile} = $commonTablesFile;
		return 1;
	}
}

sub decompress($) {
	my $self = shift;
	my $func = ref($self) . "::decompress";
	$self->logInfo("Creating working directory and extractig files");

	my $tmpDir = $self->{commonTablesDir} . "/tmp";
	if (not (-e $tmpDir)) {
		warn "$func: Cannot open temporary dir ($tmpDir): $!";
		return 0;
	} else {
		# Empty directory
		if (! $self->emptyDirectory($tmpDir)) {
			warn "$func: Cannot empty directory $tmpDir";
			return 0;
		}
	}

	# 3.2 files extraction
	my $zip = Archive::Zip->new();
	if ($zip->read($self->{commonTablesFile}) == $Archive::Zip::AZ_OK ) {
		my @files = $zip->memberNames;
		foreach my $file (@files) {
			# Check if file matches any table
			if ($file =~ m!^(\w+)\.txt$!) {
				my $tableName = lc($1);
				my $outputFile = $tmpDir . "/" . $file;
				if (defined $self->{tables}->{$tableName}) {
					$self->{tables}->{$tableName} = $outputFile;
					$zip->extractMemberWithoutPaths($file, $outputFile);
				} else {
					warn "$func: File $outputFile is not associated with any common table";
					return 0;
				}
			}
		}
	} else {
		warn "$func: File " . $self->{commonTablesFile} . " is not a valid zip file";
		return 0;
	}

	return 1;
}

sub validateInputFileContent($) {
	my $self = shift;
	my $func = ref($self) . "::validateInputFileContent";
	$self->logInfo("Validating input file content");

	my $error = 0;
	foreach my $tableName (keys %{$self->{tables}}) {
		if (not $self->{tables}->{$tableName}) {
			warn "$func: Input file not found for table $tableName";
			$error = 1;
		} else {
			open(my $fh, "<", $self->{tables}->{$tableName}) or die "Can't open file " . $self->{tables}->{$tableName} . " for table $tableName: $!";
			my $content = "";
			while(<$fh>) {
				$_ =~ s/\r//;
				$content .= $_;
			}
			close($fh);

			# Rewrite file
			open($fh, ">", $self->{tables}->{$tableName}) or die "Can't open file " . $self->{tables}->{$tableName} . ": $!";
			print $fh $content;
			close($fh);
		}
	}

	return (! $error);
}

sub load($) {
	my $self = shift;
	my $func = ref($self) . "::load";
	$self->logInfo("Loading common tables");

	my $config = $self->{config};
	my $osType = $config->{'fiuba.lamm.agromodel.run.os.type'};
	my $mySqlBin = $config->{'fiuba.lamm.agromodel.db.mysql.bin'};
	my $mysqlUser = $config->{'fiuba.lamm.agromodel.db.user'};
	my $mysqlPassword = $config->{'fiuba.lamm.agromodel.db.password'};
	my $mysqlUrl = $config->{'fiuba.lamm.agromodel.db.url'};
	my ($mysqlHost, $mysqlDB);
	if ($mysqlUrl =~ /^jdbc:mysql:\/\/(.+?)(:\d+){0,1}\/(.+?)$/) {
		$mysqlHost = $1;
		$mysqlDB = $3;
	}

	my $errors   = [];
	my $warnings = [];
	foreach my $tableName (keys %{$self->{tables}}) {
		my $inputFile = $self->{tables}->{$tableName};
		my $deleteQuery = "delete from $mysqlDB.$tableName";

		$self->logInfo("Deleting data from table $tableName");
		my $cmdDelete = "$mySqlBin --show-warnings -u" . $mysqlUser . " -p" . $mysqlPassword . " -h" . $mysqlHost . " -e \"" . $deleteQuery . "\"";
		my ($errorDelete, $warningDelete) = $self->executeQueryCmd($cmdDelete);
		push(@$errors, $errorDelete) if (defined($errorDelete));
		push(@$warnings, $warningDelete) if (defined($warningDelete));

		if (defined($inputFile) and !defined($errorDelete) and !defined($warningDelete)) {
			$self->logInfo("Inserting data into table $tableName");
			my $loadQuery = "load data local infile '$inputFile' into table $mysqlDB.$tableName";
			if ($osType eq 'Windows') {
				$loadQuery .= " lines terminated by " . '\'\r\n\'';
			}
			my $cmdLoad = "$mySqlBin --show-warnings -u" . $mysqlUser . " -p" . $mysqlPassword . " -h" . $mysqlHost . " --local-infile -e \"" . $loadQuery . "\"";
			my ($errorLoad, $warningLoad) = $self->executeQueryCmd($cmdLoad);
			push(@$errors, $errorLoad) if (defined($errorLoad));
			# push(@$warnings, $warningLoad) if (defined($warningLoad)); # TODO: Rehabilitar cuando se eliminen los warnings
		}
	}

	$self->{errors}   = $errors;
	$self->{warnings} = $warnings;

	return 1;
}

sub cleanup($) {
	my $self = shift;
	my $func = ref($self) . "::cleanup";
	$self->logInfo("Cleaning up");

	my $tmpDir = $self->{commonTablesDir} . "/tmp";
	$self->emptyDirectory($tmpDir);
	
	return 1;
}

1;
