package Agromodel::Loaders::Scenario;

use base qw(Agromodel::Loaders::Base);
use strict;
use Archive::Zip;
use Cwd;

sub new($$) {
	my ($pkg, $scenarioIdString) = @_;
	my $func = $pkg . "::new";

	my $tables = {
		'activitymgmt_scenario'                 =>      1,
		'croppingcycle'                         =>      1,
		'croppingcycle_activity'                =>      1,
		'croppingcycle_contexts'                =>      1,
		'croppingcycle_historicaladjustment'    =>      1,
		'croppingcycle_inputprices'             =>      1,
		'croppingcycle_region'                  =>      1,
		'croppingcycle_soiltype'                =>      1,
		'farm'                                  =>      1,
		'farmer'                                =>      1,
		'plot'                                  =>      1,
		'scenario'                              =>      1,
		'scenario_parameter'                    =>      0,
		'yieldlookup'                           =>      1,
	};
	my $config = $pkg->getConfig();
	if (not defined $config) {
		warn "$func: Undefined configuration object";
		return undef();
	}

	my $data = {
		config	=>	$config,
		tables	=>	$tables,
		scenarioIdString	=>	$scenarioIdString,
		init_handlers	=>	[
			'validateScenarioIds'
		],
		fork_handlers	=>	[
			'validateInputFile',
			'decompress',
			'validateInputFileContent',
			'load',
			'cleanup',
		],
	};

	return bless($data, $pkg);
}

sub validateScenarioIds($) {
        my $self = shift;
	my $func = ref($self) . "::validateScenarioIds";

        my @parts = split(",", $self->{scenarioIdString});
        my $scenarioIds = [];
        foreach my $part (@parts) {
                if ($part =~ m!^(\d+)\-(\d+)$!) {
                        foreach my $scenarioId ($1..$2) {
                                push(@$scenarioIds, $scenarioId);
                        }
                } elsif ($part =~ m!^\d+$!) {
                        push(@$scenarioIds, $part);
                } else {
                        warn "$func: Invalid scenario specification: $part";
			return 0;
                }
        }

	$self->{fork_elements} = $scenarioIds;
	return 1;
}

sub validateInputFile($$) {
	my ($self, $scenarioId) = @_;
	my $func = ref($self) . "::validateInputFile";
	$self->logInfo("Validating files for scenario $scenarioId");	

        if ($scenarioId !~ /^\d+$/) {
                warn "Invalid scenarioId: $scenarioId";
		return 0;
        }
        my $currentDir = getcwd();
        my $scenarioDir = $currentDir . "/Scenarios";
        if (not opendir(SCENARIO_DIR, $scenarioDir)) {
		warn "$func: Cannot open scenario dir ($scenarioDir): $!";
		return 0;
        } else {
		$self->{scenarioDir} = $scenarioDir;
		closedir(SCENARIO_DIR);
	}

        my $scenarioFile = $scenarioDir . "/" . $scenarioId . ".zip";
        if (not (-e $scenarioFile)) {
		warn "$func: Scenario file ($scenarioFile) not found";
		return 0;
        } else {
		$self->{scenarioFile} = $scenarioFile;
		return 1;
	}
}

sub decompress($$) {
	my ($self, $scenarioId) = @_;
	my $func = ref($self) . "::decompress";
	$self->logInfo("Creating working directory and extractig files for scenario $scenarioId");

	my $baseTmpDir = $self->{scenarioDir} . "/tmp";
	my $tmpDir = $baseTmpDir . "/" . $scenarioId;
	if (not (-e $tmpDir)) {
		if (! mkdir($tmpDir)) {
			warn "$func: Cannot open temporary dir ($tmpDir): $!";
			return 0;
		}
	} else {
		# Empty directory
		if (! $self->emptyDirectory($tmpDir)) {
			warn "$func: Cannot empty directory $tmpDir";
			return 0;
		}
	}

	# 3.2 files extraction
	$self->{scenarioTables} = {};
	my $zip = Archive::Zip->new();
	if ($zip->read($self->{scenarioFile}) == $Archive::Zip::AZ_OK ) {
		my @files = $zip->memberNames;
		foreach my $file (@files) {
			# Check if file matches any table
			if ($file =~ m!^(\w+)\.txt$!) {
				my $tableName = lc($1);
				my $outputFile = $tmpDir . "/" . $file;
				if (defined $self->{tables}->{$tableName}) {
					$self->{scenarioTables}->{$tableName} = $outputFile;
					$zip->extractMemberWithoutPaths($file, $outputFile);
				} else {
					warn "$func: File $outputFile is not associated with any scenario table";
					return 0;
				}
			}
		}
	} else {
		warn "$func: File " . $self->{scenarioFile} . " is not a valid zip file";
		return 0;
	}

	return 1;
}

sub validateInputFileContent($$) {
	my ($self, $scenarioId) = @_;
	my $func = ref($self) . "::validateInputFileContent";
	$self->logInfo("Validating input file content for scenario $scenarioId");

	my $error = 0;
	foreach my $tableName (keys %{$self->{scenarioTables}}) {
		if ($self->{scenarioTables}->{$tableName} == 1) {
                        warn "$func: Input file not found for table $tableName";
                        $error = 1;
                }
        }

	return (! $error);
}

sub load($$) {
	my ($self, $scenarioId) = @_;
	my $func = ref($self) . "::load";
	$self->logInfo("Loading scenario $scenarioId");

	my $config = $self->{config};
	my $mySqlBin = $config->{'fiuba.lamm.agromodel.db.mysql.bin'};
	my $mysqlUser = $config->{'fiuba.lamm.agromodel.db.user'};
	my $mysqlPassword = $config->{'fiuba.lamm.agromodel.db.password'};
	my $mysqlUrl = $config->{'fiuba.lamm.agromodel.db.url'};
	my ($mysqlHost, $mysqlDB);
	if ($mysqlUrl =~ /^jdbc:mysql:\/\/(.+?)(:\d+){0,1}\/(.+?)$/) {
		$mysqlHost = $1;
		$mysqlDB = $3;
	}

	my $errors   = [];
	my $warnings = [];
	foreach my $tableName (keys %{$self->{scenarioTables}}) {
		my $inputFile = $self->{scenarioTables}->{$tableName};

		if ($inputFile) {
			$self->logInfo("Inserting data into table $tableName for scenario $scenarioId");
			my $loadQuery = "load data local infile '$inputFile' into table $mysqlDB.$tableName lines terminated by " . '\'\r\n\'';
			my $cmdLoad = "$mySqlBin --show-warnings -u" . $mysqlUser . " -p" . $mysqlPassword . " -h" . $mysqlHost . " --local-infile -e \"" . $loadQuery . "\"";
			my ($errorLoad, $warningLoad) = $self->executeQueryCmd($cmdLoad);
			push(@$errors, $errorLoad) if (defined($errorLoad));
			push(@$warnings, $warningLoad) if (defined($warningLoad));
		}
	}

	$self->{errors}   = $errors;
	$self->{warnings} = $warnings;

	return 1;
}

sub cleanup($$) {
        my ($self, $scenarioId) = @_;
        my $func = ref($self) . "::cleanUpScenario";
        $self->logInfo("Cleaning up scenario $scenarioId temporary files");

        my $tmpDir = $self->{scenarioDir} . "/tmp/" . $scenarioId;
        $self->emptyDirectory($tmpDir);
        rmdir($tmpDir);

        return 1;
}

1;
