package Agromodel::Loaders::ScenarioEnsemble;

use base qw(Agromodel::Loaders::Base);
use strict;
use Archive::Zip;
use Cwd;
use File::Copy;
use Tie::File;

sub new($$) {
	my ($pkg, $scenarioEnsembleId) = @_;
	my $func = $pkg . "::new";

	my $tables = {
		'activitymgmt_scenario'                 =>      1,
		'croppingcycle'                         =>      1,
		'croppingcycle_activity'                =>      1,
		'croppingcycle_contexts'                =>      1,
		'croppingcycle_historicaladjustment'    =>      1,
		'croppingcycle_inputprices'             =>      1,
		'croppingcycle_region'                  =>      1,
		'croppingcycle_soiltype'                =>      1,
		'farm'                                  =>      1,
		'farmer'                                =>      1,
		'plot'                                  =>      1,
		'scenario'                              =>      1,
		'scenario_ensemble'			=>	1,
		'scenario_parameter'                    =>      0,
		'yieldlookup'                           =>      1,
	};
	my $config = $pkg->getConfig();
	if (not defined $config) {
		warn "$func: Undefined configuration object";
		return undef();
	}

	my $data = {
		config	=>	$config,
		tables	=>	$tables,
		realizations	=>	{},
		scenarioEnsembleId	=>	$scenarioEnsembleId,
		init_handlers	=>	[
			'validateScenarioEnsembleId',
			'validateInputFile',
			'decompress',
			'validateInputFileContent',
		],
		fork_handlers	=>	[
			'combine',
			'load',
			'cleanUpScenario'
		],
		join_handlers	=>	[
			'cleanUpEnsemble'
		]
	};

	return bless($data, $pkg);
}

sub validateScenarioEnsembleId($) {
	my $self = shift;
        my $func = ref($self) . "::validateScenarioEnsembleId";

	if ($self->{scenarioEnsembleId} !~ m!^\d+$!) {
		warn "$func: Invalid scenario ensemble: " . $self->{scenarioEnsembleId};
		return 0;
        }

	return 1;
}

sub validateInputFile($) {
	my $self = shift;
	my $func = ref($self) . "::validateInputFile";
	$self->logInfo("Validating files for scenario ensemble " . $self->{scenarioEnsembleId});	

        my $currentDir = getcwd();
        my $scenarioDir = $currentDir . "/Scenarios";
        if (not opendir(SCENARIO_DIR, $scenarioDir)) {
		warn "$func: Cannot open scenario dir ($scenarioDir): $!";
		return 0;
        } else {
		$self->{scenarioDir} = $scenarioDir;
		closedir(SCENARIO_DIR);
	}

        my $scenarioFile = $scenarioDir . "/" . $self->{scenarioEnsembleId} . ".zip";
        if (not (-e $scenarioFile)) {
		warn "$func: Scenario file ($scenarioFile) not found";
		return 0;
        } else {
		$self->{scenarioFile} = $scenarioFile;
		return 1;
	}
}

sub decompress($) {
	my $self = shift;
	my $func = ref($self) . "::decompress";
	$self->logInfo("Creating working directory and extractig files for scenario ensemble " . $self->{scenarioEnsembleId});

	my $tmpDir = $self->{scenarioDir} . "/tmp";
	if (not (-e $tmpDir)) {
		warn "$func: Cannot open temporary dir ($tmpDir): $!";
		return 0;
	} else {
		# Empty directory
		if (! $self->emptyDirectory($tmpDir)) {
			warn "$func: Cannot empty directory $tmpDir";
			return 0;
		}
	}

	# Files extraction
	my $zip = Archive::Zip->new();
	if ($zip->read($self->{scenarioFile}) == $Archive::Zip::AZ_OK ) {
		my @files = $zip->memberNames;
		foreach my $file (@files) {
			my $member = $zip->memberNamed($file);
	
			if (! $member->isDirectory()) {
				# Check if file matches any table
				my ($tableName, $scenarioId);
				if ($file =~ m!^realizations/(\w+?)_(\d+)\.txt$!) {
					($tableName, $scenarioId) = (lc($1), $2);
				} elsif ($file =~ m!^(\w+)\.txt$!) {
					$tableName = lc($1);
				}
				my $outputFile = $tmpDir . "/" . $file;

				if (defined($scenarioId)) {
					if ($scenarioId !~ /^\d+$/) {
						warn "$func: Invalid scenarioId ($scenarioId) for realization file $file";
						return 0;
					} elsif (not defined($self->{tables}->{$tableName})) {
						warn "$func: File $outputFile is not associated with any scenario ensemble table";
						return 0;
					} else {
						if (not defined($self->{realizations}->{$tableName})) {
							$self->{realizations}->{$tableName} = {};
						}

						$self->{realizations}->{$tableName}->{$scenarioId} = $outputFile;
						$zip->extractMemberWithoutPaths($file, $outputFile);
					}
				} elsif (defined $self->{tables}->{$tableName}) {
					$self->{tables}->{$tableName} = $outputFile;
					$zip->extractMemberWithoutPaths($file, $outputFile);
				} else {
					warn "$func: File $outputFile is not associated with any common table";
					return 0;
				}
			}
		}
	} else {
		warn "$func: File " . $self->{scenarioFile} . " is not a valid zip file";
		return 0;
	}

	return 1;
}

sub validateInputFileContent($) {
	my $self = shift;
	my $func = ref($self) . "::validateInputFileContent";
	$self->logInfo("Validating input file content for scenario ensemble " . $self->{scenarioEnsembleId});

	# Common scenario tables
	foreach my $tableName (keys %{$self->{tables}}) {
		if (($self->{tables}->{$tableName} == 1) and (not defined $self->{realizations}->{$tableName})) {
                        warn "$func: Input file not found for table $tableName";
			return 0;
                }
        }

	# Realization tables
	my $scenarioIds = undef();
	foreach my $realizationTable (keys(%{$self->{realizations}})) {
		my $rScenarioIds = [ sort { $a <=> $b } keys(%{$self->{realizations}->{$realizationTable}}) ];

		if (not defined($scenarioIds)) {
			$scenarioIds = $rScenarioIds;
		} else {
			if (scalar(@$scenarioIds) != scalar(@$rScenarioIds)) {
				warn "$func: Realizations mismatch for table $realizationTable";
				return 0;
			} else {
				my $i = 0;
				while ($i < scalar(@$scenarioIds)) {
					if ($scenarioIds->[$i] != $rScenarioIds->[$i]) {
						warn "$func: Realizations mismatch for table $realizationTable";
						return 0;
					}
					$i++;
				}
			}	
		}
	}

	if (defined($scenarioIds)) {
		$self->{fork_elements} = $scenarioIds;
		return 1;
	} else {
		warn "$func: No realizations found";
		return 0;
	}
}

sub combine($$) {
	my ($self, $scenarioId) = @_;	
	my $func = ref($self) . "::combine";
	
	# 1. Create temporary directory
	my $tmpScenarioPath = $self->{scenarioDir} . "/tmp/tmp_" . $scenarioId;
	if (-d $tmpScenarioPath) {
		$self->emptyDirectory($tmpScenarioPath);
	} else {
		if (! mkdir($tmpScenarioPath)) {
			warn "$func: Error creating temporary directory $tmpScenarioPath";
			return 0;
		}
	}

	# 2. Copy realizations files for scenarioId
	foreach my $realizationTable (keys(%{$self->{realizations}})) {
		my $sourceFile = $self->{realizations}->{$realizationTable}->{$scenarioId};
		my $targetFile = $tmpScenarioPath . "/" . $realizationTable . ".txt";

		if (! File::Copy::copy($sourceFile, $targetFile)) {
			warn "$func: Error copying file $sourceFile to target path";
			return 0;
		}
	}

	# 3. Copy common files for scenarioId and replace placeholders.
	foreach my $tableName (keys(%{$self->{tables}})) {
		my $sourceFile = $self->{tables}->{$tableName};
		my $targetFile = $tmpScenarioPath . "/" . $tableName . ".txt";

		if (($sourceFile ne "1") and ($sourceFile ne "0")) { 
			if (! File::Copy::copy($sourceFile, $targetFile)) {
				warn "$func: Error copying file $sourceFile to target path";
				return 0;
			} else {
				my $array = [];
				if (tie(@$array, 'Tie::File', $targetFile)) {
					foreach my $line (@$array) {
						$line =~ s/SCENARIO_ID/$scenarioId/;
					}
					untie(@$array);
				} else {
					warn "$func: Can't open $targetFile: $!";
					return 0;
				}
			}
		}
	}

	return 1;
}

sub load($$) {
	my ($self, $scenarioId) = @_;
	my $func = ref($self) . "::load";

	my $config = $self->{config};
	my $mySqlBin = $config->{'fiuba.lamm.agromodel.db.mysql.bin'};
	my $mysqlUser = $config->{'fiuba.lamm.agromodel.db.user'};
	my $mysqlPassword = $config->{'fiuba.lamm.agromodel.db.password'};
	my $mysqlUrl = $config->{'fiuba.lamm.agromodel.db.url'};
	my ($mysqlHost, $mysqlDB);
	if ($mysqlUrl =~ /^jdbc:mysql:\/\/(.+?)(:\d+){0,1}\/(.+?)$/) {
		$mysqlHost = $1;
		$mysqlDB = $3;
	}

	my $errors   = [];
	my $warnings = [];
	foreach my $tableName (keys %{$self->{tables}}) {
		my $inputFile = $self->{scenarioDir} . "/tmp/tmp_" . $scenarioId . "/" . $tableName . ".txt";

		if (-e $inputFile) {
			my $loadQuery = "load data local infile '$inputFile' into table $mysqlDB.$tableName lines terminated by " . '\'\r\n\'';
			my $cmdLoad = "$mySqlBin --show-warnings -u" . $mysqlUser . " -p" . $mysqlPassword . " -h" . $mysqlHost . " --local-infile -e \"" . $loadQuery . "\"";
			my ($errorLoad, $warningLoad) = $self->executeQueryCmd($cmdLoad);
			push(@$errors, $errorLoad) if (defined($errorLoad));
			push(@$warnings, $warningLoad) if (defined($warningLoad));
		}
	}

	$self->{errors}   = $errors;
	$self->{warnings} = $warnings;

	return 1;
}

sub cleanUpScenario($$) {
	my ($self, $scenarioId) = @_;
	my $func = ref($self) . "::cleanUpScenario";
	$self->logInfo("Finished loading of scenario $scenarioId of ensemble " . $self->{scenarioEnsembleId});

	my $tmpDir = $self->{scenarioDir} . "/tmp/tmp_" . $scenarioId;
	$self->emptyDirectory($tmpDir);
	rmdir($tmpDir);
	
	return 1;
}

sub cleanUpEnsemble($) {
	my $self = shift;
	my $func = ref($self) . "::cleanUpEnsemble";
	$self->logInfo("Cleaning up ensemble " . $self->{scenarioEnsembleId});

	my $tmpDir = $self->{scenarioDir} . "/tmp";
	$self->emptyDirectory($tmpDir);
	
	return 1;
}

1;
