package fiuba.lamm.agromodel;

import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.decision.*;
import fiuba.lamm.agromodel.execution.Scenario;
import fiuba.lamm.agromodel.factories.*;
import fiuba.lamm.agromodel.output.*;

import java.util.*;

@ExportableClass(minimumAllowedVerbosity = Verbosity.Reduced)
public class SoilType {
	protected Byte soilTypeId;
	protected Region region;
	protected String soilTypeName;
	protected String soilTexture;
	protected Double soilTypeProportionOfRegion;
	
	/* Attributes that are time-dependent */
	protected Double landSalePrice;
	protected Double rentalPrice;
	protected Double expectedGrossMargin;
	protected Double farmTaxes;
	protected Double organicMatter;
	protected Double waterStorageCapacity;
	protected Double soilCurveNumber;
	
	public SoilType(Byte soilTypeId, Region region, String soilTypeName, String soilTexture, Double soilTypeProportionOfRegion) {
		this.soilTypeId = soilTypeId;
		this.region = region;
		this.soilTypeName = soilTypeName;
		this.soilTexture = soilTexture;
		this.soilTypeProportionOfRegion = soilTypeProportionOfRegion;		
	}
	
	@ExportableField(title = "Soil type Id", order = 1, minimumAllowedVerbosity = Verbosity.Reduced)
	public Byte getSoilTypeId() {
		return this.soilTypeId;
	}
	public Region getRegion() {
		return this.region;
	}
	
	@ExportableField(title = "Soil type name", order = 2, minimumAllowedVerbosity = Verbosity.Reduced)
	public String getSoilTypeName() {
		return this.soilTypeName;
	}
	public String getSoilTexture() {
		return this.soilTexture;
	}
	public Double getSoilTypeProportionOfRegion() {
		return this.soilTypeProportionOfRegion;
	}	
	public Double getLandSalePrice() {
		return this.landSalePrice;
	}
	
	@ExportableField(title = "Land rental price", order = 3, minimumAllowedVerbosity = Verbosity.Reduced)
	public Double getRentalPrice() {
		return this.rentalPrice;
	}
	public void setRentalPrice(Double rentalPrice) {
		this.rentalPrice = rentalPrice;
	}
	
	@ExportableField(title = "Expected gross margin", order = 4, minimumAllowedVerbosity = Verbosity.Reduced)
	public Double getExpectedGrossMargin() {
		return this.expectedGrossMargin;
	}
	/*
	 * Actualizacion de expected global gross margin:
	 * 
	 * Tomamos un amp equidistribuido con los 6 AMs standard (LO CUAL ASUME QUE EXISTEN 6!!) para la region a la que
	 * pertenece el tipo de suelo. Luego promediamos los ultimos N GGM correspondientes a ese amp para la tecnologia
	 * actual. Ese promedio es igual al expected global gross margin.
	 * 
	 * Cuando hay mas o menos de 6 AMs posibles para un lote lo que se hace es lo siguiente: Para cada AM(i) se calcula el EGGM(i)
	 * de un lote donde se siembra el 100% de ese AM(i). Luego el EGGM es el promedio de los EGGM(i)
	 * 
	 * Este proceso solo se realiza en los ciclos ACTIVOS
	 * 
	 */	
	public void setCurrentExpectedGrossMargin() {
		Manager manager = Manager.getInstance();
		List<ActivityManagement> ams = ActivityManagementFactory.getInstance(manager.getScenario()).findByRegion(this.region);
		List<ActivityManagement> availableAms = new ArrayList<ActivityManagement>();
		for (ActivityManagement am: ams) {
			if (am.isAvailable()) {
				availableAms.add(am);
			}
		}

		Scenario scenario = manager.getScenario();
		Integer farmMaxNumberOfPlots = (Integer) scenario.getParameter("Farm.NumberOfPlots");
		Integer croppingCyclesForEGMCalculation = (Integer) scenario.getParameter("SoilType.CroppingCyclesForExpectedGrossMarginCalculation");
		if (availableAms.size() == farmMaxNumberOfPlots) {
			ActivityManagementProportion amp = new ActivityManagementProportion();			
			for (ActivityManagement am: availableAms) {
				amp.setProportion(am, 1);
			}
			
			double ggms[] = manager.getLastGlobalGrossMargins(amp, this, croppingCyclesForEGMCalculation);
			double sum = 0.0;
			for (double ggm: ggms) {
				sum += ggm;
			}
			
			this.expectedGrossMargin = (sum / (double) ggms.length);
		} else {
			double eggm = 0.0;
			for (ActivityManagement am: availableAms) {
				ActivityManagementProportion amp = new ActivityManagementProportion();				
				amp.setProportion(am, farmMaxNumberOfPlots);
				double ggms[] = manager.getLastGlobalGrossMargins(amp, this, croppingCyclesForEGMCalculation);
				double sum = 0.0;
				for (double ggm: ggms) {
					sum += ggm;
				}
				eggm += (sum / (double) ggms.length);
			}
			
			this.expectedGrossMargin = (eggm / (double) availableAms.size());
		}
	}
	
	public Double getFarmTaxes() {
		return this.farmTaxes;
	}
	public Double getOrganicMatter() {
		return this.organicMatter;
	}
	public Double getWaterStorageCapacity() {
		return this.waterStorageCapacity;
	}
	public Double getSoilCurveNumber() {
		return this.soilCurveNumber;
	}
	
	public void update(Double landSalePrice, Double farmTaxes, Double organicMatter, Double waterStorageCapacity, Double soilCurveNumber) {
		this.landSalePrice = landSalePrice;		
		this.farmTaxes = farmTaxes;
		this.organicMatter = organicMatter;
		this.waterStorageCapacity = waterStorageCapacity;
		this.soilCurveNumber = soilCurveNumber;
	}
		
	public void update(Double landSalePrice, Double rentalPrice, Double farmTaxes, Double organicMatter, Double waterStorageCapacity, Double soilCurveNumber) {
		this.landSalePrice = landSalePrice;		
		this.rentalPrice = rentalPrice;
		this.farmTaxes = farmTaxes;
		this.organicMatter = organicMatter;
		this.waterStorageCapacity = waterStorageCapacity;
		this.soilCurveNumber = soilCurveNumber;
	}
	
	public boolean equals(Object o) {
		if (o instanceof SoilType) {
			SoilType s = (SoilType) o;
			return this.soilTypeId.equals(s.getSoilTypeId());
		}
		
		return false;
	}
	
	public int hashCode() {
		return this.soilTypeId.hashCode();
	}
	
	public String toString() {
		return this.soilTypeName;
	}
}