package fiuba.lamm.agromodel.agents;

import java.util.*;
import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.decision.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.finance.*;
import fiuba.lamm.agromodel.output.*;
import fiuba.lamm.agromodel.utils.*;

@ExportableClass(minimumAllowedVerbosity = Verbosity.Reduced)
public class Farm extends Agent implements Comparable<Farm> {
	protected Integer farmId;		// Farm ID number
	protected Integer gisFarmId;    // ID used in Shape File
	protected Region region;		// Region where farm is located
	protected Farmer owner;			// ID of Farm owner
	protected Farmer operator;		// ID of agent who operates farm
	protected Double farmSize;		// Farm size (in has)
	protected SoilType soilType;    // Farm soil type
	protected CreaGroup creaGroup;	// ID of CREA group to which farm belongs
	protected Byte cattleStructure;	// Does farm have infrastructure for cattle farming?
	protected List<Plot> plots;		// Plots (or lots) in this farm
	protected Double expectedGrossMargin;
	protected Double expectedOperationCost;
	protected Double expectedRentalCost;
	protected Double expectedUtility;
	protected Double lastGlobalGrossMargin;
	protected AspirationLevel aspirationLevel;
	protected List<Double> ggmPerturbationFactorHistory;
	protected List<Farmer> operatorsHistory;
	
	// Factores para reflejar incertidumbre en los resultados
	protected Double outcomeFactor;
	protected List<Double> outcomeFactorHistory;
	
	// Decision variables
	protected ActivityManagementProportion nextAMP;
	protected LandUseSelectionMechanism lastLandUseSelectionMechanism;
	protected PlotAssignmentMechanism plotAssignmentMechanism;
	
	// Diffusion of innovations
	protected List<ActivityManagement> newUnknownManagements;
	protected List<ActivityManagement> unknownManagements;
	protected List<Advisor> advisors;
		
	public Farm(Integer farmId, Region region, Farmer owner, SoilType soilType, Double farmSize, CreaGroup creaGroup, Byte cattleStructure) {
		this.farmId = farmId;
		this.gisFarmId = null;
		this.region = region;
		this.owner = owner;
		this.operator = null;
		this.soilType = soilType;
		this.farmSize = farmSize;
		this.creaGroup = creaGroup;
		this.cattleStructure = cattleStructure;
		this.plots = new ArrayList<Plot>();
		this.expectedGrossMargin = 0.0;
		this.expectedOperationCost = 0.0;
		this.expectedRentalCost = 0.0;
		this.expectedUtility = null;
		this.lastGlobalGrossMargin = null;
		this.outcomeFactor = null;
		this.outcomeFactorHistory = new ArrayList<Double>();
		this.aspirationLevel = null;
		this.ggmPerturbationFactorHistory = new ArrayList<Double>();
		this.operatorsHistory = new ArrayList<Farmer>();
		this.nextAMP = null;
		this.lastLandUseSelectionMechanism = null;
		this.plotAssignmentMechanism = new RandomPlotAssignmentMechanism(); // TODO: Por ahora es fijo, mas adelante puede haber diversas estrategias de asignacion de lotes.
		this.advisors = new ArrayList<Advisor>();
		this.unknownManagements = new ArrayList<ActivityManagement>();
		this.newUnknownManagements = new ArrayList<ActivityManagement>();		
	}
	
	@ExportableField(title = "Farm Id", order = 1, minimumAllowedVerbosity = Verbosity.Reduced)
	public Integer getFarmId() {
		return this.farmId;
	}
	
	public Integer getGISFarmId() {
		return this.gisFarmId;
	}
	public void setGISFarmId(Integer gisFarmId) {
		this.gisFarmId = gisFarmId;
	}
	
	@ExportableField(title = "Farm size", order = 2, minimumAllowedVerbosity = Verbosity.Reduced)
	public Double getFarmSize() {
		return this.farmSize;
	}
	public void setFarmSize(Double farmSize) {
		Double plotSize = farmSize / (double) this.plots.size();
		for (Plot plot: plots) {
			plot.setPlotSize(plotSize);
		}
		
		this.farmSize = farmSize;
	}
	
	public Region getRegion() {
		return this.region;
	}
	@ExportableField(title = "Region Id", order = 3, minimumAllowedVerbosity = Verbosity.Reduced)
	public Byte getRegionId() {
		return this.region.getRegionId();
	}
	
	@ExportableField(title = "Owner Id", order = 4, minimumAllowedVerbosity = Verbosity.Reduced)
	public Integer getOwnerId() {
		return this.owner.getFarmerId();
	}
	public Farmer getOwner() {
		return this.owner;
	}
	public void setOwner(Farmer owner) {
		this.owner = owner;
	}
	
	@ExportableField(title = "Operator Id", order = 5, minimumAllowedVerbosity = Verbosity.Reduced)	
	public Integer getOperatorId() {
		return (this.operator != null) ? this.operator.getFarmerId() : null;
	}
	public Farmer getOperator() {
		return this.operator;
	}
	public void setOperator(Farmer operator) {
		this.operator = operator;
	}
	
	@ExportableField(title = "Global gross margin", order = 6, minimumAllowedVerbosity = Verbosity.Reduced)
	public Double getLastGlobalGrossMargin() {
		return this.lastGlobalGrossMargin;
	}	
		
	public SoilType getSoilType() {
		return this.soilType;
	}
	
	@ExportableField(title = "Soil Type Id", order = 13, minimumAllowedVerbosity = Verbosity.Reduced)
	public Byte getSoilTypeId() {
		return this.soilType.getSoilTypeId();
	}
	
	public CreaGroup getCreaGroup() {
		return this.creaGroup;
	}
	public Byte getCattleStructure() {
		return this.cattleStructure;
	}
	
	public List<Plot> getPlots() {
		return this.plots;
	}
	public void addPlot(Plot plot) {
		this.plots.add(plot);
	}
	
	public ActivityManagementProportion getCurrentActivityManagementProportion() {
		return new ActivityManagementProportion(this.plots);
	}
	
	public boolean equals(Object o) {
		if (o instanceof Farm) {
			Farm f = (Farm) o;
			return this.farmId.equals(f.getFarmId());
		}
		
		return false;
	}
	public int hashCode() {
		return this.farmId.hashCode();
	}
	
	public String toString() {
		return this.farmId.toString();
	}

	public Double getExpectedGrossMargin() {
		return this.expectedGrossMargin;
	}

	// Calculates the expected operation cost assuming that the farm is being rented
	public Double getExpectedOperationCost() {
		return (this.expectedOperationCost + this.expectedRentalCost);
	}
	
	// Calculates the expected operation cost
	public Double getExpectedOperationCost(Farmer farmer) {
		double expectedTotalCost = this.expectedOperationCost;
		if (!farmer.equals(this.owner)) expectedTotalCost += this.expectedRentalCost;
		
		return expectedTotalCost;
	}
	
	public void setDefaultAspirationLevel() {
		this.aspirationLevel = AspirationLevel.getDefaultAspirationLevel(this.soilType);
	}
	
	public AspirationLevel getAspirationLevel() {
		return this.aspirationLevel;
	}
	
	public void updateAspirationLevelBeforeCroppingCycle() {
		this.aspirationLevel.updateBeforeCroppingCycle(this);
	}
	
	public void updateAspirationLevelAfterCroppingCycle() {
		this.aspirationLevel.updateAfterCroppingCycle(this);
	}
	
	@ExportableField(title = "Aspiration level adjusted by context comparison", order = 9, minimumAllowedVerbosity = Verbosity.Verification)
	public Double getLastAspirationLevelAdjustedByContextComparison() {
		List<Double> aspirationLevels = this.aspirationLevel.getAspirationLevelsAdjustedByContextComparison();
		if (aspirationLevels.isEmpty()) {
			return this.aspirationLevel.getLevel();
		} else {
			return aspirationLevels.get(aspirationLevels.size() - 1);
		}
	}
	
	@ExportableField(title = "Last land use selection mechanism", order = 10, minimumAllowedVerbosity = Verbosity.Verification)
	public LandUseSelectionMechanism getLastLandUseSelectionMechanism() {
		return this.lastLandUseSelectionMechanism;
	}
	public void setLastLandUseSelectionMechanism(LandUseSelectionMechanism mechanism) {
		this.lastLandUseSelectionMechanism = mechanism;
	}
	
	@ExportableField(title = "AMP Id", order = 11, minimumAllowedVerbosity = Verbosity.Reduced)
	public Integer getNextAMPId() {
		AMPSortedTable table = Manager.getInstance().getAMPSortedTable();
		return table.getActivityManagementProportionId(this.nextAMP);
	}
	public ActivityManagementProportion getNextAMP() {
		return this.nextAMP;
	}
	public void setNextAMP(ActivityManagementProportion nextAMP) {
		this.nextAMP = nextAMP;
	}
	
	@ExportableField(title = "Expected utility", order = 12, minimumAllowedVerbosity = Verbosity.Verification)	
	public Double getExpectedUtility() {
		return this.expectedUtility;
	}
	public void setExpectedUtility(Double expectedUtility) {
		this.expectedUtility = expectedUtility;
	}	
	
	/**
	 * Sets last global gross margin (only for testing purpose)
	 */
	public void setLastGlobalGrossMargin(Double lastGlobalGrossMargin) {
		this.lastGlobalGrossMargin = lastGlobalGrossMargin;
	}
	
	// Calculates gross operation margin for a given year
	public double getGrossOperationMargin(Short croppingCycleId, double fixedCostScaleFactor) {
		Double meanGlobalGrossMargin = Accountant.getGlobalGrossMargin(this.plots, this.operator.getManagerialAbility(), fixedCostScaleFactor, croppingCycleId, croppingCycleId);
		if (this.outcomeFactor != null) {
			this.lastGlobalGrossMargin = this.outcomeFactor * meanGlobalGrossMargin;
		} else {
			this.lastGlobalGrossMargin = meanGlobalGrossMargin;
		}
		
		return this.lastGlobalGrossMargin;
	}
	
	/*
	 * This piece returns the expected cost of operating a given farm (COOF).
	 * The COOF is used during the "area update" stage, to decide if an agent can
	 * afford to keep operating a farm,. or if she can rent an additional farm.
	 * The COOF is equal to the fixed production costs of the Activity/Management
	 * with the highest cost (this is to ensure that an agent has sufficient funds
	 * to do any AM in this farm). For tenants, the cost of leasing the farm is
	 * added to the COOF.
	 */
	public void calculateExpectedValues (CroppingCycle croppingCycle, double fixedCostScaleFactor) {
		try {
			// 1. Calculate expected cost of operating a given farm (COOF)				
			Double managerialAbility = (this.operator != null) ? operator.getManagerialAbility() : 1.0;
			this.expectedOperationCost = this.region.getExpectedOperationCost(croppingCycle) * managerialAbility * fixedCostScaleFactor; 
					
			// 2. Calculate leasing cost (only applicable if farm is operated by a tenant)
			Manager manager = Manager.getInstance();
			this.expectedRentalCost = IndirectCost.LeasingCost.getCost(manager.getScenario(), null, this, true);
					
			/* 3. Calculate expected gross margin for a farm. 
			 * This quantity is a factor of the soil type in the farm and
			 * is read from the data base during initialization.
			 */
			this.expectedGrossMargin = 0.0;
			for (int i = 0; i < plots.size(); i++) {
				Plot plot = plots.get(i);
				this.expectedGrossMargin += this.soilType.getExpectedGrossMargin() * plot.getPlotSize();
			}
		} catch (Exception e) {
			throw new FulminantError("Error calculating expected valued for farm: " + e);
		}
	}
	
	/**
	 * Sets expected operation cost (COP), expected rental cost (COA - COP) and expected gross margin (MBE) for
	 * the current cropping cycle. Use of this function is only recommended for testing purposes. calculateExpectedValues()
	 * should be used during simulation
	 */
	public void setExpectedValues(double expectedOperationCost, double expectedRentalCost, double expectedGrossMargin) {
		this.expectedOperationCost = expectedOperationCost;
		this.expectedRentalCost = expectedRentalCost;
		this.expectedGrossMargin = expectedGrossMargin;
	}
	
	/**
	 * Farms are sorted by farmId
	 */
	public int compareTo(Farm aFarm) {
		return this.farmId.compareTo(aFarm.farmId);
	}
	
	// Asigna el nuevo AMP haciendo la mejor rotacion posible
	public void assignPlots() {		
		this.plotAssignmentMechanism.assignLandUseToPlots(this);
	}
	
	
	public List<ActivityManagement> getUnknownManagements() {
		return this.unknownManagements;
	}
	public void setUnknownManagements(List<ActivityManagement> unknownManagements) {
		this.unknownManagements = unknownManagements;
	}
	
	public List<ActivityManagement> getNewUnknownManagements() {
		return this.newUnknownManagements;
	}
	public void setNewUnknownManagements(List<ActivityManagement> newUnknownManagements) {
		this.newUnknownManagements = newUnknownManagements;
	}	
	
	public List<Advisor> getAdvisors() {
		return this.advisors;
	}
	public void setAdvisors(List<Advisor> advisors) {
		List<Advisor> newAdvisors = new ArrayList<Advisor>(advisors);
		List<Advisor> currentAdvisors = getAdvisors();
		
		for (Advisor advisor: currentAdvisors) {
			if (! advisors.contains(advisor)) {
				currentAdvisors.remove(advisor); // Remove the advisor from the current list
				advisor.removeCustomer(this);
			} else {
				newAdvisors.remove(advisor); // The advisor is actually in the list. Remove from new list
			}
		}
		
		for (Advisor advisor: newAdvisors) {
			currentAdvisors.add(advisor);
			advisor.addCustomer(this);
		}
		
		this.advisors = currentAdvisors;
	}
	
	@ExportableField(title = "Outcome factor", order = 13, minimumAllowedVerbosity = Verbosity.Verification)
	public Double getOutcomeFactor() {
		return this.outcomeFactor;
	}
	public Double getOutcomeFactor(int croppingCycleOffset) {
		if (croppingCycleOffset == 0) {
			return getOutcomeFactor();
		} else if (croppingCycleOffset > 0){
			if (croppingCycleOffset <= this.outcomeFactorHistory.size()) {
				return this.outcomeFactorHistory.get(this.outcomeFactorHistory.size() - croppingCycleOffset);
			} else {
				return null;
			}
		} else {
			throw new FulminantError("Esta intentanto obtener un outcome factor en el futuro (offset = " + croppingCycleOffset + ")");
		}
	}	
	public void setNewOutcomeFactor() {
		// TODO: Hardcodeado. Corregir.
		Double MinOutcomeFactor = 0.9;
		Double MaxOutcomeFactor = 1.1;
		
		if (this.outcomeFactor != null) {
			this.outcomeFactorHistory.add(new Double(this.outcomeFactor));
		}
		
		Randomizer randomizer = Manager.getInstance().getScenario().getRandomizer();
		this.outcomeFactor = MinOutcomeFactor + randomizer.getRandomDouble() * (MaxOutcomeFactor - MinOutcomeFactor);		
	}
	
	@ExportableField(title = "Adoption status", order = 14, minimumAllowedVerbosity = Verbosity.Reduced)
	// TODO: Se asume que hay una sola innovacion, por eso devuelve true o false
	public Boolean getAdoptionStatus() {
		return this.unknownManagements.isEmpty();
	}
}