package fiuba.lamm.agromodel.agents;

import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.areaupdate.mcp.wta.*;
import fiuba.lamm.agromodel.decision.*;
import fiuba.lamm.agromodel.diffusion.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.math.*;
import fiuba.lamm.agromodel.math.custom.*;
import fiuba.lamm.agromodel.factories.*;
import fiuba.lamm.agromodel.finance.*;
import fiuba.lamm.agromodel.output.*;
import fiuba.lamm.agromodel.utils.*;

import java.util.*;

import repast.simphony.engine.schedule.*;

/**
 * This class represents the most important agent of the model: the farmer. 
 * <p>
 * Paragraph 2
 * 
 * @author		Santiago Rovere
 * @since		1.0
 * 
 */

@ExportableClass(minimumAllowedVerbosity = Verbosity.Reduced)
public class Farmer extends Agent implements Comparable<Farmer> {
	protected Integer farmerId;
	protected String farmerName;
	protected Byte age;
	protected String gender;
	protected String educationalLevel;
	protected Byte familyGroupSize;
	protected OperationalStatus operationalStatus;
	protected Double workingCapital;
	protected SearchTriggeringMechanism searchTriggeringMechanism;
	protected LandUseSelectionMechanism landUseSelectionMechanism;
	protected Double managerialAbility;
	protected Double offFarmIncome;
	protected Short yrsExperience;
	protected List<Farm> ownedFarms;
	protected List<Farm> operatedFarms;
	protected List<Double> totalIncome;
	protected Double peersAdjFactor;
	protected Double lambdaUp;
	protected Double lambdaDown;
	protected Double optimism;
	protected RotatorType ownerRotatorType;
	protected RotatorType tenantRotatorType;
	protected Double riskAversion; // For EU and WTA calculation
	protected int liveOffCroppingCycles; // Number of consecutive cropping cycles as a landlord
	
	// Cropping cycle dependent attributes
	protected Double lastOperatedFarmSize;
	
	// Area update related attribues
	protected WTACalculationMechanism wtaCalculationMechanism;
	protected Double ownMinimumProgressRate;
	protected List<Double> minimumProgressRates;
	protected List<Double> progressRates;
	protected List<Double> totalNetFarmingIncomes;
	protected Map<Farm,Double> willingnessToAccept;
	protected Map<SoilType,Double> willingnessToPay;
	protected Map<Farm,List<Double>> ownFarmsIndirectCosts;
	protected Set<Region> preferredRegions;
	protected boolean isMigrationProne;
	protected Map<Region,Double> minimumLeasableAreas;
	
	// Diffusion of innovations
	protected InnovationDiffusionModel diffusionModel;
	protected InnovatorType innovatorType;
	
	public Farmer(Integer farmerId, String farmerName, Byte age, String gender, String educationalLevel, Byte familyGroupSize, Double workingCapital, SearchTriggeringMechanism searchTriggeringMechanism, LandUseSelectionMechanism landUseSelectionMechanism, WTACalculationMechanism wtaCalculationMechanism, Double managerialAbility, Double offFarmIncome, Short yrsExperience, Double peersAdjFactor, Double lambdaUp, Double lambdaDown, Double optimism, RotatorType ownerRotatorType, RotatorType tenantRotatorType, Double riskAversion, Double ownMinimumProgressRate, InnovatorType innovatorType, InnovationDiffusionModel diffusionModel) {
		this.farmerId = farmerId;
		this.farmerName = farmerName;
		this.age = age;
		this.gender = gender;
		this.educationalLevel = educationalLevel;
		this.familyGroupSize = familyGroupSize;
		this.operationalStatus = OperationalStatus.Inactive;
		this.workingCapital = workingCapital;
		this.totalNetFarmingIncomes = new ArrayList<Double>();
		this.searchTriggeringMechanism = searchTriggeringMechanism;
		this.landUseSelectionMechanism = landUseSelectionMechanism;
		this.wtaCalculationMechanism = wtaCalculationMechanism;
		this.managerialAbility = managerialAbility;
		this.offFarmIncome = offFarmIncome;
		this.yrsExperience = yrsExperience;
		this.ownedFarms = new ArrayList<Farm>();
		this.operatedFarms = new ArrayList<Farm>();
		this.totalIncome = new ArrayList<Double>();
		this.peersAdjFactor = peersAdjFactor;
		this.lambdaUp = lambdaUp;
		this.lambdaDown = lambdaDown;
		this.optimism = optimism;
		this.ownerRotatorType = ownerRotatorType;
		this.tenantRotatorType = tenantRotatorType;
		this.riskAversion = riskAversion;
		this.lastOperatedFarmSize = 0.0;
		this.liveOffCroppingCycles = this.operationalStatus.isLiveOff() ? 1 : 0;
		this.willingnessToAccept = null;
		this.willingnessToPay = null;
		this.ownFarmsIndirectCosts = new HashMap<Farm,List<Double>>();
		this.preferredRegions = new HashSet<Region>();
		this.minimumLeasableAreas = new HashMap<Region,Double>();
		this.isMigrationProne = false;
		this.ownMinimumProgressRate = ownMinimumProgressRate;
		this.minimumProgressRates = new ArrayList<Double>();
		this.progressRates = new ArrayList<Double>();
		this.innovatorType = innovatorType;
		this.diffusionModel = diffusionModel;
	}
		
	/* 
	 * Accessors: metodos get y set para acceder y modificar atributos
	 *            solo existen metodos set de aquellos campos que 
	 *            tienen sentido que sean modificados	
	 */
	@ExportableField(title = "Farmer Id", order = 1, minimumAllowedVerbosity = Verbosity.Reduced)
	public Integer getFarmerId() {
		return this.farmerId;
	}
	public String getFarmerName() {
		return this.farmerName;
	}
	public Byte getAge() {
		return this.age;
	}
	public String getGender() {
		return this.gender;
	}
	public String getEducationalLevel() {
		return this.educationalLevel;
	}
	public Byte getFamilyGroupSize() {
		return this.familyGroupSize;
	}
	
	@ExportableField(title = "Operational Status", order = 3, minimumAllowedVerbosity = Verbosity.Reduced)
	public OperationalStatus getOperationalStatus() {
		return this.operationalStatus;
	}
	public void setOperationalStatus(OperationalStatus operationalStatus) {
		if (operationalStatus.isLiveOff() && !this.operationalStatus.isLiveOff()) {
			// Operational status changed to LiveOff. Initialize the counter
			this.liveOffCroppingCycles = 1;
		} else {
			this.liveOffCroppingCycles = 0;
		}
		
		this.operationalStatus = operationalStatus;		
	}
	
	@ExportableField(title = "Working capital", order = 2, minimumAllowedVerbosity = Verbosity.Reduced)
	public Double getWorkingCapital() {
		return this.workingCapital;
	}
	public void setWorkingCapital(Double workingCapital) {
		this.workingCapital = workingCapital;
	}
	public void addWorkingCapital(Double workingCapital) {
		this.workingCapital += workingCapital;
	}
	/* Calculates working capital at the end of a cropping cycle
	 * substracting expected operation costs of operated farms
	 */
	public Double getExpectedWorkingCapital() {
		Scenario scenario = Manager.getInstance().getScenario();
		double expectedWorkingCapital = this.workingCapital;
		for (int i = 0; i < this.operatedFarms.size(); i++) {
			Farm farm = this.operatedFarms.get(i);
			expectedWorkingCapital -= farm.getExpectedOperationCost(this);
		}
		
		// Optimismo
		if (expectedWorkingCapital > 0) {
			// Para darle mas expectativa, le agrego un plus
			expectedWorkingCapital *= (Double) scenario.getParameter("Farmer.ExpectedLiquidCapitalOptimismFactor");
		}
		
		return expectedWorkingCapital;
	}
		
	public Double getManagerialAbility() {
		return this.managerialAbility;
	}
	public void setManagerialAbility(Double managerialAbility) {
		this.managerialAbility = managerialAbility;
	}

	public SearchTriggeringMechanism getSearchTriggeringMechanism() {
		return this.searchTriggeringMechanism;
	}
	
	public LandUseSelectionMechanism getLandUseSelectionMechanism() {
		return this.landUseSelectionMechanism;
	}
	
	public WTACalculationMechanism getWTACalculationMechanism() {
		return this.wtaCalculationMechanism;
	}
	
	public Double getOffFarmIncome() {
		return this.offFarmIncome;
	}
	public void setOffFarmIncome(Double offFarmIncome) {
		this.offFarmIncome = offFarmIncome;
	}
	
	public Short getYrsExperience() {
		return this.yrsExperience;
	}
	public void addExperience() {
		this.yrsExperience++;
	}
	public Double getPeersAdjFactor() {
		return this.peersAdjFactor;
	}
	public Double getLambdaUp() {
		return this.lambdaUp;
	}
	public Double getLambdaDown() {
		return this.lambdaDown;
	}
	public Double getOptimism() {
		return this.optimism;
	}
	
	public RotatorType getOwnerRotatorType() {
		return this.ownerRotatorType;
	}
	
	public RotatorType getTenantRotatorType() {
		return this.tenantRotatorType;
	}
	
	public Double getRiskAversion() {
		return this.riskAversion;
	}
	
	public Integer getLiveOffCroppingCycles() {
		return this.liveOffCroppingCycles;
	}
	public void addLiveOffCroppingCycle() {
		this.liveOffCroppingCycles++;
	}
	
	@ExportableField(title = "WTA", order = 5, minimumAllowedVerbosity = Verbosity.Verification)
	public Map<Farm,Double> getWillingnessToAccept() {
		return this.willingnessToAccept;
	}
	public void setWillingnessToAccept(Map<Farm,Double> willingnessToAccept) {
		this.willingnessToAccept = willingnessToAccept;
	}
	
	@ExportableField(title = "WTP", order = 6, minimumAllowedVerbosity = Verbosity.Verification)
	public Map<SoilType,Double> getWillingnessToPay() {
		return this.willingnessToPay;
	}
	public void setWillingnesstoPay(Map<SoilType,Double> willingnessToPay) {
		this.willingnessToPay = willingnessToPay;
	}
	
	public Double getOwnMinimumProgressRate() {
		return this.ownMinimumProgressRate;
	}
	
	public InnovatorType getInovatorType() {
		return this.innovatorType;
	}
	
	public List<Double> getTotalNetFarmingIncomes() {
		return this.totalNetFarmingIncomes;
	}
	
	public List<Double> getProgressRates() {
		return this.progressRates;
	}
	public void setCurrentProgressRate(Double progressRate) {
		this.progressRates.add(progressRate);		
	}
	
	public List<Double> getMinimumProgressRates() {
		return this.minimumProgressRates;
	}
	public void setCurrentMinimumProgressRate(Double minimumProgressRate) {
		this.minimumProgressRates.add(minimumProgressRate);		
	}
		
	public List<Farm> getOwnedFarms() {
		return new ArrayList<Farm>(this.ownedFarms);
	}
	public void addOwnedFarm(Farm farm) {
		this.ownedFarms.add(farm);
		this.updateOperationalStatus(false);
		farm.setOwner(this);
	}
	
	
	@ExportableField(title = "Last operated farm size", order = 4, minimumAllowedVerbosity = Verbosity.Verification)
	public Double getLastOperatedFarmSize() {
		return this.lastOperatedFarmSize;
	}	
	
	public List<Farm> getOperatedFarms() {
		return new ArrayList<Farm>(this.operatedFarms);
	}
	public void addOperatedFarm(Farm farm) {
		operatedFarms.add(farm);
		this.updateOperationalStatus(false);
		farm.setOperator(this);
	}
	public void removeOperatedFarm(Farm farm, boolean forcedRemoval) {
		operatedFarms.remove(farm);
		this.updateOperationalStatus(forcedRemoval);
		farm.setOperator(null);
	}
	public void clearOperatedFarmList() {
		while (! this.operatedFarms.isEmpty()) removeOperatedFarm(this.operatedFarms.get(0), true);		
	}
	
	public void updateOperationalStatus(boolean forcedRemoval) {
		boolean ownsAnyFarm = !this.ownedFarms.isEmpty();
		
		if (this.operatedFarms.isEmpty()) {
			/* Agent does not operate any farm.
			 * If removal was forced, then the farmer should not become inactive
			 */
			this.setOperationalStatus(ownsAnyFarm ? OperationalStatus.LandLord : forcedRemoval ? OperationalStatus.TenantOnly : OperationalStatus.Broke);
		} else {
			boolean operatesOtherFarms = false;
			boolean operatesOwnFarms   = false;
			for (Farm farm: this.operatedFarms) {
				if (this.ownedFarms.contains(farm)) {
					operatesOwnFarms = true;
				} else {
					operatesOtherFarms = true;
				}
			}
			
			if (operatesOwnFarms && operatesOtherFarms) {
				this.setOperationalStatus(OperationalStatus.OwnerTenant);
			} else if (operatesOwnFarms && !operatesOtherFarms) {
				this.setOperationalStatus(OperationalStatus.OwnerOnly);
			} else {
				this.setOperationalStatus(OperationalStatus.TenantOnly);
			}
		}
	}
	
	public void retire() {
		// 1. Check if farmers owns any farm
		boolean ownsAnyFarm = !this.ownedFarms.isEmpty();
		
		// 2. Release all operated farms
		while (! this.operatedFarms.isEmpty()) {
			Farm farm = this.operatedFarms.get(0);
			farm.setOperator(null);
			this.operatedFarms.remove(farm);
		}
		
		// 3. Update operational status
		this.operationalStatus = ownsAnyFarm ? OperationalStatus.RetiredOwner : OperationalStatus.RetiredTenant;
	}
	
	public List<Double> getOwnFarmIndirectCosts(Farm farm) {
		return this.ownFarmsIndirectCosts.get(farm);
	}
	public void calculateAndAddOwnFarmIndirectCost() {
		if (! this.ownedFarms.isEmpty()) {
			Manager manager = Manager.getInstance();
			for (Farm farm: this.ownedFarms) {
				Double indirectCost = IndirectCost.getTotalIndirectCostsForOwnFarm(manager.getScenario(), this, farm);
				List<Double> farmIndirectCosts = this.ownFarmsIndirectCosts.get(farm);
				if (farmIndirectCosts == null) {
					farmIndirectCosts = new ArrayList<Double>();
				}
				farmIndirectCosts.add(indirectCost);
				this.ownFarmsIndirectCosts.put(farm, farmIndirectCosts);
			}
		}
	}
	
	// Returns is the farmer is migration-prone or not
	public Boolean isMigrationProne() {
		return this.isMigrationProne;
	}
	
	// Returns the candidate regions for land rental market
	public List<Region> getCandidateRegions() {
		if (this.isMigrationProne || this.preferredRegions.isEmpty()) {
			// All the regions a candidates
			return RegionFactory.getInstance().findAll();
		} else {
			// Only the preferred regions are candidates
			return new ArrayList<Region>(this.preferredRegions);
		}
	}
	
	// Returns the minimum leasable area for a give region
	public Double getMinimumLeasableArea(Region region) {
		return this.minimumLeasableAreas.get(region);
	}
	
	// Updates land-market variables before entering the Land Market Model
	public void updateLandMarketVariables(boolean allowFarmersMigration, boolean nOutOfMEvaluationMechanism) {
		// 1. Update last operated farm size
		this.lastOperatedFarmSize = 0.0;
		for (Farm farm: this.operatedFarms) {
			this.lastOperatedFarmSize += farm.getFarmSize();
		}
		
		// 1. Set WTA and WTP to null
		this.setWillingnessToAccept(null);
		this.setWillingnesstoPay(null);
		
		/*
		 * 3. Update preferred regions. The set of preferred contains:
		 * 
		 * a. All the regions where the farmer owns at least one farm
		 * b. All the regions where the farmer operated at least one farm during the last cropping cycle.
		 */
		this.preferredRegions.clear();
		for (Farm farm: this.ownedFarms) {
			Region region = farm.getRegion();
			this.preferredRegions.add(region);
		}
		for (Farm farm: this.operatedFarms) {
			Region region = farm.getRegion();
			this.preferredRegions.add(region);
		}
		
		/*
		 * 4. Update "migration-prone" flag:
		 * 
		 * a. If the scenario allows farmer migration, then the flag may be activated, otherwise it cannot.
		 * b. If the flag is activated, it will remain activated for the rest of the simulation.
		 * c. If the flag is not activated, then
		 * 		If the set of preferred regions contains more than one region, activate the flag.
		 * 		Else Let P = Probability of Activating the Migration Flag = f(Total operated area) 
		 * 			 And activate the flag with a probability of P.
		 */
		if (allowFarmersMigration) {
			if (! this.isMigrationProne) {
				if (this.preferredRegions.size() > 1) {
					this.isMigrationProne = true;
				} else {
					Function migrationProbabilityFunction = new MigrationProbabilityFunction();				
					Double p = migrationProbabilityFunction.apply(this.lastOperatedFarmSize);
					Map<Boolean, Double> events = new HashMap<Boolean, Double>();
					events.put(new Boolean(true), p);
					events.put(new Boolean(false), (1.0 - p));
					
					Randomizer randomizer = Manager.getInstance().getScenario().getRandomizer();
					this.isMigrationProne = randomizer.getRandomEvent(events);
				}
			}
		}
		
		// 5. Update MLA for each region
		this.minimumLeasableAreas.clear();
		List<Region> regions = this.getCandidateRegions();
		for (Region region: regions) {
			Function mlaFunction = new MLAScalingFunction(region, this.preferredRegions);
			this.minimumLeasableAreas.put(region, mlaFunction.apply(this.lastOperatedFarmSize));
		}		
	}
	
	// Sort farmers by farmerId
	public int compareTo(Farmer aFarmer) {
		return this.farmerId.compareTo(aFarmer.farmerId);
	}
	
	public boolean equals(Object o) {
		if ((o != null) && (o instanceof Farmer)) {
			Farmer f = (Farmer) o;
			return this.farmerId.equals(f.getFarmerId());
		}
		
		return false;
	}	
	public int hashCode() {
		return this.farmerId.hashCode();
	}
	
	public String toString() {
		return this.farmerId.toString();
	}
	
	/* Scheduled methods */
	
	/**
	 * The first action that the farmer carries out is to decide whether he retires or not
	 * This action is carried out if only if:
	 * 
	 *    a. Scenario allows farmers' retirement
	 *	  b. It is not the first cycle    
	 *    c. The farmer is active and not already retired
	 */
	@ScheduledMethod (start = 1.03, interval = 1)
	public void considerRetirement() {
		try {
			Manager manager = Manager.getInstance();
			Scenario scenario = manager.getScenario();
			CroppingCycle croppingCycle = manager.getCurrentCroppingCycle();
			boolean considerRetirement = (scenario.hasAttribute(Scenario.Attribute.AllowFarmerRetirement) && !croppingCycle.isFirstActive() && this.operationalStatus.isActive() && !this.operationalStatus.isRetired());
	
			if (considerRetirement) {
				Double retirementProbability = 0.01; // TODO: Hacer que la probabilidad de retiro sea un parametro del escenario
				Map<Boolean,Double> retirementEvents = new HashMap<Boolean,Double>();
				retirementEvents.put(new Boolean(true), retirementProbability);
				retirementEvents.put(new Boolean(false), 1.0 - retirementProbability);
				Randomizer randomizer = scenario.getRandomizer();
				Boolean retires = randomizer.getRandomEvent(retirementEvents);
				
				if (retires) {
					this.retire();
				}
			}
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}

	// Aspiration level update using expected contexts
	@ScheduledMethod (start = 1.06, interval = 1)
	public void updateAspirationLevelUsingExpectedContexts() {
		try {
			for (int i = 0; i < this.operatedFarms.size(); i++) {
				Farm farm = this.operatedFarms.get(i);
				farm.updateAspirationLevelBeforeCroppingCycle();
			}
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}
	
	// Innovation diffusion process (define adoption status)
	@ScheduledMethod (start = 1.4, interval = 1)
	public void defineAdoptionStatus() {
		try {
			Manager manager = Manager.getInstance();
			Scenario scenario = manager.getScenario();
			if (scenario.hasAttribute(Scenario.Attribute.HasTechnologyInnovations) && !this.operatedFarms.isEmpty()) {
				ActivityManagementFactory factory = ActivityManagementFactory.getInstance(scenario);
				Map<Region,List<ActivityManagement>> innovationsByRegion = new HashMap<Region,List<ActivityManagement>>();
				
				for (Farm farm: this.operatedFarms) {
					Region region = farm.getRegion();
					if (! innovationsByRegion.containsKey(region)) {
						innovationsByRegion.put(region, factory.findInnovationsByRegion(region));
					}				
					List<ActivityManagement> innovations = innovationsByRegion.get(region);
					
					List<ActivityManagement> unknownManagements = new ArrayList<ActivityManagement>();				
					if (! innovations.isEmpty()) {				
						for (ActivityManagement innovation: innovations) {
							if (! this.diffusionModel.doesInnovate(farm, innovation)) {
								unknownManagements.add(innovation);
							}						
						}
					}				
					farm.setNewUnknownManagements(unknownManagements);
				}
			}
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}
	
	// Innovation diffusion process (update adoption status)
	@ScheduledMethod (start = 1.45, interval = 1)
	public void updateAdoptionStatus() {
		try {
			Manager manager = Manager.getInstance();
			Scenario scenario = manager.getScenario();
			if (scenario.hasAttribute(Scenario.Attribute.HasTechnologyInnovations)) {
				for (Farm farm: this.operatedFarms) {
					List<ActivityManagement> unknownManagements = farm.getNewUnknownManagements();
					farm.setUnknownManagements(unknownManagements);
				}
			}
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}
	
	// Activity management selection
	@ScheduledMethod (start = 1.5, interval = 1)
	public void planCurrentCycleLandUse() {		
		try {
			if (this.operationalStatus.isOperating()) {
				// Activity management selection (the first cropping cycle is read form the database)
				Manager manager = Manager.getInstance();
				CroppingCycle currentCroppingCycle = manager.getCurrentCroppingCycle();			
				if (! currentCroppingCycle.isFirstActive()) {
					for (int i = 0; i < this.operatedFarms.size(); i++) {
						Farm farm = this.operatedFarms.get(i);
						
						// a. Check if a new search should be triggered
						Map<LandUseSelectionRestriction, Object> restrictions = new HashMap<LandUseSelectionRestriction, Object>();
						/*
						 * For now, we only have 3 restrictions:
						 * 
						 * 1) The region of the farm
						 * 2) Minimum and maximum plots per activity
						 * 3) List of activities: exclude unknown technologies
						 * 	  
						 */
						// b1.
						restrictions.put(LandUseSelectionRestriction.RegionRestriction, farm.getRegion());					
						// b2.
						RotatorType rotatorType = this.equals(farm.getOwner()) ? this.ownerRotatorType : this.tenantRotatorType;
						restrictions.put(LandUseSelectionRestriction.MinimumActivityGroupProportionRestriction, rotatorType.getMinPlotsPerActivityGroup());
						restrictions.put(LandUseSelectionRestriction.MaximumActivityGroupProportionPRestriction, rotatorType.getMaxPlotsPerActivityGroup());					
						// b3.
						restrictions.put(LandUseSelectionRestriction.TechnologyRestriction, farm.getUnknownManagements());
						
						LandUseSelectionMechanism selectionMechanism = null;					
						if (this.searchTriggeringMechanism.triggerSearch(farm, restrictions)) {
							// b. Trigger search using land use selection mechanism for this farmer (defined in initialization database)						
							selectionMechanism = this.landUseSelectionMechanism;
						} else {
							// c. Trigger habit mechanism
							selectionMechanism = new HabitLandUseSelectionMechanism();
						}
						
						ActivityManagementProportion nextAMP = selectionMechanism.search(farm, restrictions);
						farm.setLastLandUseSelectionMechanism(selectionMechanism.getLastLandUseSelectionMechanism());
						farm.setNextAMP(nextAMP);
					}
				}			
			}
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}
	
	// Ejecuta la decision de asignacion de AMP tomada anteriormente
	@ScheduledMethod (start = 1.55, interval = 1)
	public void executeCurrentCroppingCycleLandUse() {
		try {
			// Execute assignment if cropping cycle is not the first one
			Manager manager = Manager.getInstance();
			CroppingCycle currentCroppingCycle = manager.getCurrentCroppingCycle();
			if (! currentCroppingCycle.isFirstActive()) {
				for (Farm farm: this.operatedFarms) {
					farm.assignPlots();
				}
			}
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}
	
	/**
	 * Calculates the economic result for the current cropping cycle. This method should be called after 
	 * Farmer.planCurrentCycle()
	 * <p>
	 * After the computation of the economic result, this Farmer's capital will be updated as well as
	 * it's experience and age.
	 * </p> 
	 * 
	 * @see		<a href="https://www.srsoft.com.ar/svn/projects/Agromodel/docs/Documentation/Computation%20of%20Economic%20Results/" target="_blank">Computation of Economic Results</a> 
	 * @since	1.0
	 * @author	Santiago Rovere 
	 */
	@ScheduledMethod (start = 1.65, interval = 1)
	public void calculateMargins() {
		try {
			Manager manager = Manager.getInstance();
			Scenario scenario = manager.getScenario();
			
			// Do calculation if and only if farmer is active
			if (this.operationalStatus.isActive()) {						
				/* 1. Agricultural Income */
				double fixedCostScaleFactor = 1.0; // Default scale factor
				if (scenario.hasAttribute(Scenario.Attribute.HasScaleEconomics)) {
					/* 1a. Scale factor calculation for fixed costs */	
					// 1a. i. Calculation of total area operated by a farmer during this cropping cycle
					double operatorTotalFarmSize = 0.0;
					for (int i = 0; i < operatedFarms.size(); i++) {
						Farm farm = operatedFarms.get(i);
						operatorTotalFarmSize += farm.getFarmSize();
					}
					
					Function fcScalingFunction = new FixedCostScalingFunction();
					fixedCostScaleFactor = fcScalingFunction.apply(operatorTotalFarmSize);
				}
										
				/* 1.2 Global (farm-wide) gross margin calculation and indirect costs calculation for each farm */
				double totalNetFarmingIncome = 0.0;		
				for (int i = 0; i < operatedFarms.size(); i++) {
					Farm farm = operatedFarms.get(i);
					Double farmGlobalGrossMargin = farm.getGrossOperationMargin(manager.getCurrentCroppingCycle().getCroppingCycleId(), fixedCostScaleFactor);
					Double farmIndirectOperationCosts = IndirectCost.getTotalIndirectCosts(scenario, this, farm);
					totalNetFarmingIncome += (farmGlobalGrossMargin - farmIndirectOperationCosts);
				}
				
				// 1.3 Add total net farming income to list				
				this.totalNetFarmingIncomes.add(totalNetFarmingIncome);
				
				/* 1.4 If the farmer owns a farm, calculate indirect costs for that farm only in order to be used in WTA calculation */
				this.calculateAndAddOwnFarmIndirectCost();
				
				/* 
				 * 2. Household non-farming income.
				 * 
				 * This variable includes income received by an agent (or household) from
				 * non-farming sources such as:
				 * 
				 * - Leasing of farms owned (net land rental income)
				 * - Interests and dividends from investments
				 * - Off-farm income
				 * - Other income sources
				 */
				double personalProfits = 0.0; 
				
				/* 2.1 Net Land Rental Income (received by owners who lease out their farm)*/
				for (int i = 0; i < ownedFarms.size(); i++) {
					Farm farm = ownedFarms.get(i);
					personalProfits += Accountant.getRentalMargin(scenario, farm);
				}
				
				/* 2.2 Interests and dividends */
				// TODO: For now, we have not implemented this income source
				double interests = 0.0;
				personalProfits += interests;
				
				/* 2.3 Off-farm income */
				personalProfits += this.offFarmIncome;
							
				/* 3 Farmer total income is the sum of Agriculture income and household non-farming income */
				double farmerTotalIncome = totalNetFarmingIncome + personalProfits;
				
				double incomeTaxCost = 0.0;				
				if (scenario.hasAttribute(Scenario.Attribute.HasIncomeTax)) {
					/* 3.1 Income tax calculation (personal tax, paid by a farmer).
					 * Any household income lower than profitTaxMinIncome (minimum taxable income)
					 * is not subject to income tax.
					 */
					double minimumNonTaxableIncome = (Double) scenario.getParameter("Farmer.MinimumNonTaxableIncome");
					double taxableIncome = (farmerTotalIncome - minimumNonTaxableIncome);	// Taxable income
					if (taxableIncome > 0) {			// Taxable income > 0, tax must be paid...
						// Profit tax calculation
						incomeTaxCost = Accountant.getIncomeTax(taxableIncome);
					}
				}
										
				/* 
				 * 3.2 Farmer withdrawal 
				 * 
				 * farmer net income = farmer total income (explained earlier) - profit tax
				 * It is calculated upon the mean farmer net income of the last 5 years
				 * 
				 */
				double farmerNetIncome = (farmerTotalIncome - incomeTaxCost);
				totalIncome.add(farmerNetIncome);
				
				double withdrawalCost = 0.0;			
				Double minimumYearlyWithdrawal = (Double) scenario.getParameter("Farmer.MinimumYearlyWithdrawal");
				if (scenario.hasAttribute(Scenario.Attribute.HasVariableWithdrawal)) {
					// Mean farmer net income calculation
					double meanTotalIncome = 0.0;
					if (scenario.hasAttribute(Scenario.Attribute.HasMemoryForWithdrawal)) {
						// Calculate average net income for the last 5 years
						double sumTotalIncome = 0.0;
						int years = (totalIncome.size() > 5) ? 5 : totalIncome.size();
						for (int i = (totalIncome.size() - years); i < totalIncome.size(); i++) {
							sumTotalIncome += totalIncome.get(i);
						}
						meanTotalIncome = sumTotalIncome / years;
					} else {
						meanTotalIncome = farmerNetIncome;
					}
					
					// Withdrawal function calculation. Must be >= minimumYearlyWithdrawal
					if (meanTotalIncome <= minimumYearlyWithdrawal) {
						withdrawalCost = minimumYearlyWithdrawal; 
					} else {
						Double maximumYearlyWithdrawal = (Double) scenario.getParameter("Farmer.MaximumYearlyWithdrawal");
						Double withdrawalRatioForMaximumAsIncome = (Double) scenario.getParameter("Farmer.WithdrawalRatioForMaximumAsIncome");
						Function withdrawalFunction = new FarmerWithdrawalFunction(minimumYearlyWithdrawal, maximumYearlyWithdrawal, withdrawalRatioForMaximumAsIncome);
						withdrawalCost = withdrawalFunction.apply(meanTotalIncome);
					}
				} else {
					withdrawalCost = minimumYearlyWithdrawal;
				}
				
				/* 3.3 Debt service */
				// TODO: for now farmers do not borrow money and thus do not repay debt
				double debtPayment = 0.0;
				
				/* Calculate total household expenses as the sum of:
				 * 1. Household withdrawal
				 * 2. Personal (income) taxes
				 * 3. Debt service
				 * 4. Other expenses
				 */
				double personalCosts = incomeTaxCost + withdrawalCost + debtPayment;
				
				/* 4. Cropping cycle margin calculation */
				double margin = (farmerTotalIncome - personalCosts);
							
				/* 5. Update working capital and years of experience */
				addWorkingCapital(margin);
				addExperience();			
			}
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}

	// Aspiration level update using real contexts
	@ScheduledMethod (start = 1.7, interval = 1)
	public void updateAspirationLevelUsingRealContexts() {
		try {
			for (int i = 0; i < this.operatedFarms.size(); i++) {
				Farm farm = this.operatedFarms.get(i);
				farm.updateAspirationLevelAfterCroppingCycle();		
			}
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}
}