package fiuba.lamm.agromodel.agents;

import java.lang.reflect.*;
import java.io.*;
import java.sql.*;
import java.text.*;
import java.util.*;
import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.agents.factories.*;
import fiuba.lamm.agromodel.areaupdate.*;
import fiuba.lamm.agromodel.areaupdate.mcp.*;
import fiuba.lamm.agromodel.decision.*;
import fiuba.lamm.agromodel.diffusion.advisory.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.factories.*;
import fiuba.lamm.agromodel.finance.*;
import fiuba.lamm.agromodel.networks.*;
import fiuba.lamm.agromodel.output.*;
import fiuba.lamm.agromodel.utils.*;
import fiuba.lamm.agromodel.utils.compression.*;
import repast.simphony.engine.schedule.*;
import repast.simphony.engine.environment.*;
import repast.simphony.space.graph.*;

public class Manager extends Agent {	
	private Scenario scenario;
	private CroppingCycle firstActiveCroppingCycle;
	private CroppingCycle currentCroppingCycle;
	private Map<SoilType,Double> defaultAspirationLevels;
	private AdvisoryModel advisoryModel;
	private Network<Agent> neighbourhood;
	private Map<Region, List<ActivityManagementProportion>> activityManagementProportions;
	private AMPSortedTable ampSortedTable;
	private Map<ActivityManagementProportion, Map<SoilType, List<Double>>> ampGlobalGrossMargins;
	private Map<Class<?>,File> outputFiles;
	private long checkPoint;
		
	/**
	 * Gets current instance of manager
	 * 
	 * @return Current instance of manager.
	 */
	public static synchronized Manager getInstance() {
		Scenario scenario = (Scenario) RunState.getInstance().getFromRegistry("scenario");
		return scenario.getManager();
	}
	
	/**
	 * Gets a new instance of manager for simulation or testing purposes 
	 */
	public static synchronized Manager newInstance(Scenario scenario, List<Farmer> farmers) {
		Manager scenarioManager = null;
		
		try {
			scenarioManager = new Manager(scenario);
			scenario.setManager(scenarioManager);
			
			// Initialize cropping cycle
			scenarioManager.currentCroppingCycle = CroppingCycleFactory.getInstance(scenario).findFirst();
			scenarioManager.firstActiveCroppingCycle = scenarioManager.currentCroppingCycle;
			
			/*
			 * Update expected contexts and calculate list of possible AMPs 
			 * IF AND ONLY IF any of this conditions is met:
			 * 
			 * 1) Not executing a Test Unit
			 * 2) Aspiration Level Test Execution
			 * 3) Decision Making Text Execution
			 */
	
			// Generate list of possible AMPs and AMP Sorted Table
			scenarioManager.ampSortedTable.build();
			scenarioManager.generateListOfAvailableAMPs();
			
			// Busco el primer ciclo activo y calculos los GGM para los ciclos pasivos (solo de memoria)			
			while (! scenarioManager.currentCroppingCycle.isActive()) {
				scenarioManager.nextCycle();
				
				// Para el caso de los owners, calculo los costos indirectos en los ciclos de carga a los efectos de poder contar con esos datos
				// al momento de realizar el calculo de el WTA
				for (Farmer farmer: farmers) {
					if (! farmer.getOwnedFarms().isEmpty()) {
						farmer.calculateAndAddOwnFarmIndirectCost();
					}
				}
			}
			
			// Set the current cropping cycle as the first active
			scenarioManager.currentCroppingCycle.setIsFirstActive(true);
		} catch (Exception e) {
			e.printStackTrace();
			throw new FulminantError("Error while creation scenario manager: " + e);
		}
		
		return scenarioManager;
	}
		
	private Manager(Scenario scenario) {
		this.scenario = scenario;
		this.neighbourhood = null;
		this.defaultAspirationLevels = null;
		this.advisoryModel = null;
		this.activityManagementProportions = new HashMap<Region, List<ActivityManagementProportion>>();
		this.ampSortedTable = new AMPSortedTable(scenario);
		this.ampGlobalGrossMargins = new HashMap<ActivityManagementProportion, Map<SoilType, List<Double>>>();		
		this.checkPoint = java.lang.System.currentTimeMillis();
		this.outputFiles = new HashMap<Class<?>,File>();
	
		// Delete output files if exist
		File outputDir = new File("output");
		File tmpOutputDir = new File(outputDir, "tmp");
		
		Class<?> classes[] = { fiuba.lamm.agromodel.agents.Farmer.class, fiuba.lamm.agromodel.agents.Farm.class, fiuba.lamm.agromodel.Plot.class, fiuba.lamm.agromodel.SoilType.class }; 
		for (Class<?> clazz: classes) {
			String fileName = clazz.getSimpleName().toLowerCase() + "s_scenario_" + this.scenario.getScenarioId() + ".csv";
			File outputFile = new File(tmpOutputDir, fileName);
			if (outputFile.exists()) outputFile.delete();
			this.outputFiles.put(clazz, outputFile);
		}
	}
	
	public void destroy() {
		this.firstActiveCroppingCycle = null;
		this.currentCroppingCycle = null;
		this.defaultAspirationLevels = null;
		this.advisoryModel = null;
		this.neighbourhood = null;
		this.activityManagementProportions = null;
		this.ampSortedTable = null;
		this.ampGlobalGrossMargins = null;
		this.outputFiles = null;			
	}
	
	private void generateListOfAvailableAMPs() {
		RegionFactory regionFactory = RegionFactory.getInstance();
		ActivityManagementFactory activityManagementFactory = ActivityManagementFactory.getInstance(scenario);
		
		List<Region> regions = regionFactory.findAll();		
		for (int i = 0; i < regions.size(); i++) {			
			Region region = regions.get(i);
			List<ActivityManagement> regionManagements = activityManagementFactory.findByRegion(region);			
			List<ActivityManagementProportion> regionProportions = ActivityManagementProportion.getPossibleProportions(regionManagements, (Integer) scenario.getParameter("Farm.NumberOfPlots"));
			
			// Check available managements
			for (ActivityManagement management: regionManagements) {
				Short firstAvailableCroppingCycleId = management.getFirstAvailableCroppingCycleId();
				if ((firstAvailableCroppingCycleId != null) && firstAvailableCroppingCycleId.equals(this.currentCroppingCycle.getCroppingCycleId())) {
					management.setIsAvailable(true);
				}
			}
			
			this.activityManagementProportions.put(region, regionProportions);		
		}
	}
	
	public CroppingCycle getCurrentCroppingCycle() {
		return this.currentCroppingCycle;
	}
	
	public CroppingCycle getFirstActiveCroppingCycle() {
		return this.firstActiveCroppingCycle;
	}
	
	public Scenario getScenario() {
		return this.scenario;
	}
	
	public Network<Agent> getNeighbourhood() {
		return this.neighbourhood;
	}
	public void setNeighbourhood(Network<Agent> neighbourhood) {
		this.neighbourhood = neighbourhood;
	}
	
	public void setAdvisoryModel(AdvisoryModel advisoryModel) {
		this.advisoryModel = advisoryModel;
	}
	
	public Map<SoilType,Double> getDefaultAspirationLevels() {
		return this.defaultAspirationLevels;
	}
	public void setDefaultAspirationLevels(Map<SoilType,Double> defaultAspirationLevels) {
		this.defaultAspirationLevels = defaultAspirationLevels;
	}
	
	public List<ActivityManagementProportion> getActivityManagementProportions(Region region) {
		return this.activityManagementProportions.get(region);
	}
	
	public AMPSortedTable getAMPSortedTable() {
		return this.ampSortedTable;
	}

	public Map<ActivityManagementProportion,List<Double>> getLastGlobalGrossMargins(SoilType soilType, int nCycles) {
		Map<ActivityManagementProportion,List<Double>> ggms = new HashMap<ActivityManagementProportion,List<Double>>();
		Iterator<ActivityManagementProportion> amps = this.ampGlobalGrossMargins.keySet().iterator();
		while (amps.hasNext()) {
			ActivityManagementProportion amp = amps.next();
			List<Double> ampGgms = this.ampGlobalGrossMargins.get(amp).get(soilType);
			if (ampGgms != null) {
				ggms.put(amp, ampGgms.subList(ampGgms.size() - nCycles, ampGgms.size()));
			}
		}
		
		return ggms;
	}
	
	/**
	 * This method returns the last nCycles (or less) global gross margins for the given activity/management/proportion
	 * (amp) and the give soil type (soilType). This value is calculated by the manager assuming 
	 * a managerial ability of 1.0 and a fixed cost scale factor of 1.0
	 * 
	 * @param amp	activity/management/proportion
	 * @param soilType	soil type
	 * @param nCycles	number of cycles to consider
	 * @return	an array of global gross margins
	 */
	public double[] getLastGlobalGrossMargins(ActivityManagementProportion amp, SoilType soilType, int nCycles) {
		// Cuando no existen nCycles de informacion se devuelven solo los ciclos para los que hay datos
		// Si no hay ningun dato aun, se devuelve un array vacio
		Map<SoilType, List<Double>> globalGrossMarginsForAllSoilTypes = this.ampGlobalGrossMargins.get(amp);
		List<Double> globalGrossMargins = null;
		if (globalGrossMarginsForAllSoilTypes != null) {
			globalGrossMargins = globalGrossMarginsForAllSoilTypes.get(soilType);
		}
		
		double r[] = null;
		if (globalGrossMargins != null) {
			r = new double[Math.min(nCycles, globalGrossMargins.size())];
			int i = 0;
			int j = globalGrossMargins.size() - 1;
			while ((i < nCycles) && (j >= 0)) {
				r[i] = globalGrossMargins.get(j);
				i++;
				j--;
			}			
		} else {
			r = new double[0];
		}
		
		return r;
	}
	
	/*
	 * Devuelve el Global Gross Margin para un dado AMP, un tipo de suelo y un ciclo determinado. El ciclo
	 * corresponde a el ultimo ciclo - croppingCycleOffset. De modo que si croppingCycleOffset == 0, 
	 * se devuelve el Global Gross Margin para el ultimo ciclo. Si no existe valor, tira un error.
	 */
	public Double getGlobalGrossMargin(ActivityManagementProportion amp, SoilType soilType, Integer croppingCycleOffset) {
		Map<SoilType, List<Double>> globalGrossMarginsForAllSoilTypes = this.ampGlobalGrossMargins.get(amp);
		List<Double> globalGrossMargins = null;
		if (globalGrossMarginsForAllSoilTypes != null) {
			globalGrossMargins = globalGrossMarginsForAllSoilTypes.get(soilType);
		}
		
		Double ggm = null;
		if (globalGrossMargins != null) {
			int index = globalGrossMargins.size() - 1 - croppingCycleOffset;
			if ((index >= 0) && (index < globalGrossMargins.size())) {
				ggm = globalGrossMargins.get(index);
			} else {
				throw new FulminantError("El offset es invalido: " + croppingCycleOffset);
			}
		} else {
			throw new FulminantError("No existen ggms para el AMP " + amp + " para el tipo de suelo " + soilType);
		}
		
		return ggm;
	}
	
	/* Scheduled methods */
	@ScheduledMethod (start = 1.0, interval = 1)
	public void updateDefaultAspirationLevel() {
		try {
			this.defaultAspirationLevels = AspirationLevel.getDefaultAspirationLevels();
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}
	
	// Land market model execution
	@ScheduledMethod (start = 1.1, interval = 1)
	public void executeLandMarketModel() {
		try {
			if (! this.currentCroppingCycle.isFirstActive()) {
				// Crea el modelo de mercado de alquiler
				LandMarketModel landMarketModel = new MCPLandMarketModel();			
				landMarketModel.scheduleExecution();
			}
			
			// Update advisor's network after land market execution
			if (this.advisoryModel != null) {
				// TODO: Actualizar la red del asesor externo luego el mercado de alquileres
			}
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}
	
	@ScheduledMethod (start = 1.60, interval = 1)
	public void adjustYields() {
		try {
			if (scenario.hasAttribute(Scenario.Attribute.HasYieldAdjustment)) {
				// 1. Generar archivos
				FarmNetworkModel farmNetworkModel = this.scenario.getFarmNetworkModel();
				if (farmNetworkModel instanceof GISNetworkModel) {
					GISNetworkModel gisNetworkModel = (GISNetworkModel) farmNetworkModel;
					
					// 1a. Creacion de carpeta temporal de trabajo
					File tmpDir = new File(java.lang.System.getProperty("fiuba.lamm.agromodel.run.tmpdir"));
					File baseDir = new File(tmpDir, "gis_" + this.scenario.getScenarioId().toString());			
					baseDir.mkdir();
		
					// 1b. Exportacion de archivos GIS a carpeta temporal
					gisNetworkModel.exportFarmsLandUse(baseDir);
		
					// 1c. Creacion de archivo ZIP a partir de archivos GIS
					String inputFileNames[] = { "farming_regions.shp", "farming_regions.shx", "farming_regions.dbf", "farming_regions.prj" };
					File outputFile = new File(baseDir, "farming_regions.zip");
					try {
						ZipCompressor compressor = new ZipCompressor();
						compressor.compress(baseDir, inputFileNames, outputFile, true);
					} catch (IOException ioe) {
						throw new FulminantError("Error while creating ZIP GIS file: " + ioe);
					} finally {
						baseDir.delete();
					}
				} else {
					throw new FulminantError("El modelo no puede generar los archivos de shape para Mike ya que el modelo de red que se utiliza no esta basdo en un espacio GIS");
				}
				
				// TODO: Este metodo tiene que invocar al subsistema que intareactua con Mike-SHE para obtener los valores de napas.
				//       Luego, con estos datos se deben alterar los rindes con la formula empirica de Esteban.
			}
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}
		
	@ScheduledMethod (start = 1.8, interval = 1)
	public void outputResults() {
		try {
			Verbosity scenarioVerbosity = this.scenario.getVerbosityLevel();
			DecimalFormat decimalFormat = scenarioVerbosity.getDecimalFormat();		
			String delimiter = ";";
			
			try {
				// 0. Specifications for each output file			
				Iterator<Class<?>> classes = this.outputFiles.keySet().iterator();
				while (classes.hasNext()) {
					Class<?> clazz = classes.next();
					ExportableClass classAnnotation = null;
					if (clazz.isAnnotationPresent(fiuba.lamm.agromodel.output.ExportableClass.class)) {
						classAnnotation = clazz.getAnnotation(fiuba.lamm.agromodel.output.ExportableClass.class);
					} else {
						throw new FulminantError("Class is not exportable: " + clazz.getName());
					}
					
					Verbosity classVerbosity = classAnnotation.minimumAllowedVerbosity();
					if (classVerbosity.getLevel() <= scenarioVerbosity.getLevel()) {
						// 1. Create PrintWriter and file if necessary				
						File outputFile = this.outputFiles.get(clazz);
						boolean appendHeaders = false;
						if (outputFile.createNewFile()) appendHeaders = true;
						PrintWriter writer = new PrintWriter(new FileWriter(outputFile, true));
						
						// 2. Find all agents
						Factory factory = Factory.getFactoryForClass(clazz, this.scenario);
						List<?> objects = factory.findAll();
						Method allMethods[] = clazz.getDeclaredMethods();
						
						// 3. Find exportable fields and print headers
						StringBuilder buffer = new StringBuilder();
						List<Method> methods = new ArrayList<Method>();
						for (int i = 0; i < allMethods.length; i++) {
							if (allMethods[i].isAnnotationPresent(fiuba.lamm.agromodel.output.ExportableField.class)) {
								ExportableField annotation = allMethods[i].getAnnotation(fiuba.lamm.agromodel.output.ExportableField.class);
								Verbosity methodVerbosity  = annotation.minimumAllowedVerbosity();
								boolean outputMethodData   = (methodVerbosity.getLevel() <= scenarioVerbosity.getLevel());
								
								if (outputMethodData) {
									// Insert method ordered in list
									Integer pos = null;
									Integer order = annotation.order();
									Integer j = 0;
									while ((j < methods.size()) && (pos == null)) {
										Method anotherMethod = methods.get(j);
										int anotherOrder = anotherMethod.getAnnotation(fiuba.lamm.agromodel.output.ExportableField.class).order();
										if (order < anotherOrder) pos = j;
										
										j++;
									}
									if (pos == null) {
										methods.add(allMethods[i]);
									} else {
										methods.add(pos, allMethods[i]);
									}
								}
							}
						}
						
						// 4. Append headers if necessary
						if (appendHeaders) {
							buffer.append("Cropping cycle");
							for (int i = 0; i < methods.size(); i++) {
								Method method = methods.get(i);
								ExportableField annotation = method.getAnnotation(fiuba.lamm.agromodel.output.ExportableField.class);
								
								buffer.append(delimiter); 
								buffer.append(annotation.title());
							}
							
							writer.println(buffer.toString());
						}
						
						// 5. Print data
						for (int i = 0; i < objects.size(); i++) {
							Object object = objects.get(i);
							
							buffer = new StringBuilder();
							buffer.append(this.currentCroppingCycle.getCroppingCycleName());
							for (int j = 0; j < methods.size(); j++) {
								Method method = methods.get(j);
								Object value = method.invoke(object);
								boolean isDecimal = ((value instanceof Double) || (value instanceof Float));
								String outputValue = null;
								if (value != null) {
									if (isDecimal) {
										Double d = (Double) value;
										if ((decimalFormat == null) || Double.isInfinite(d) || Double.isNaN(d)) {
											outputValue = d.toString();
										} else {
											outputValue = decimalFormat.format(d);
										}
									} else {
										outputValue = value.toString();
									}
								} else {
									outputValue = new String();
								}						
								
								buffer.append(delimiter);
								buffer.append(outputValue);
							}
							writer.println(buffer.toString());					
						}
						
						// 6. Close writer
						writer.close();
					}
				}
			} catch (Exception e) {
				e.printStackTrace(java.lang.System.err);
				throw new FulminantError("Error al buscar los agentes para hacer el output de datos");			
			}
			
			/*
			 *  7. Chequeo de consistencia a posteriori: 
			 *  
			 *  a) Ningun campo puede quedar sin operador
			 *  b) Todos los farmers tienen que tener un operational status acorde a los campos operados
			 */
			List<Farm> farms = null;
			List<Farmer> farmers = null;
			try {
				farms = FarmFactory.getInstance(this.scenario).findAll();
				farmers = FarmerFactory.getInstance(this.scenario).findAll();
			} catch (java.sql.SQLException sqle) {
				throw new FulminantError("Error al buscar los campos y/o productores: " + sqle);
			}
			
			for (Farm farm: farms) {
				if (farm.getOperator() == null) {
					for (Farmer farmer: farmers) {
						java.lang.System.err.println("Farmer " + farmer.getFarmerId() + "; operational status = " + farmer.getOperationalStatus());
					}
					throw new FulminantError(this.currentCroppingCycle.getCroppingCycleName() + ": El campo " + farm.getFarmId() + " no tiene ningun operador");
				}
			}
			
			for (Farmer farmer: farmers) {
				List<Farm> operatedFarms = farmer.getOperatedFarms();
				List<Farm> ownedFarms = farmer.getOwnedFarms();
				OperationalStatus operationalStatus = farmer.getOperationalStatus();
				
				if (! ownedFarms.isEmpty()) {
					// Es un propietario
					if (operationalStatus.equals(OperationalStatus.LandLord) || operationalStatus.equals(OperationalStatus.RetiredOwner)) {
						if (ownedFarms.isEmpty()) {
							throw new FulminantError(this.currentCroppingCycle.getCroppingCycleName() + ": El farmer " + farmer.getFarmerId() + " es un Land Lord / Retired Owner pero no tiene campos propios");
						} else if (! operatedFarms.isEmpty()) {
							throw new FulminantError(this.currentCroppingCycle.getCroppingCycleName() + ": El farmer " + farmer.getFarmerId() + " es un Land Lord / Retired Owner pero esta operando algun campo = " + operatedFarms);
						}
					} else if (operationalStatus.equals(OperationalStatus.OwnerOnly)) {
						// Controlar que solo opere sus propios campos					
						if (operatedFarms.size() != ownedFarms.size()) {
							throw new FulminantError(this.currentCroppingCycle.getCroppingCycleName() + ": El farmer " + farmer.getFarmerId() + " es un Propietario puro pero la cantidad de campos operados difiere de la cantidad de campos propios; operados = " + operatedFarms + "; propios = " + ownedFarms);
						} else {
							boolean contains = true;
							int i = 0;
							while ((i < ownedFarms.size()) && contains) {
								contains = operatedFarms.contains(ownedFarms.get(i));
								i++;
							}
							
							if (! contains) {
								throw new FulminantError(this.currentCroppingCycle.getCroppingCycleName() + ": El farmer " + farmer.getFarmerId() + " es un Propietario puro no esta operando al menos uno de sus campos propios; operados = " + operatedFarms + "; propios = " + ownedFarms);
							}
						}
					} else if (operationalStatus.equals(OperationalStatus.OwnerTenant)) {
						// Controlar que opere todos sus campos y ademas algun otro.
						if (operatedFarms.size() <= ownedFarms.size()) {
							throw new FulminantError(this.currentCroppingCycle.getCroppingCycleName() + ": El farmer " + farmer.getFarmerId() + " es un Propietario/Arrendatario pero la cantidad de campos operados es menor o igual a la cantidad de campos propios; operados = " + operatedFarms + "; propios = " + ownedFarms);
						} else {
							boolean contains = true;
							int i = 0;
							while ((i < ownedFarms.size()) && contains) {
								contains = operatedFarms.contains(ownedFarms.get(i));
								i++;
							}
						
							if (! contains) {
								throw new FulminantError(this.currentCroppingCycle.getCroppingCycleName() + ": El farmer " + farmer.getFarmerId() + " es un Propietario/Arrendatario pero no esta operando al menos uno de sis campos propios; operados = " + operatedFarms + "; propios = " + ownedFarms);
							}
						}
					} else {
						throw new FulminantError(this.currentCroppingCycle.getCroppingCycleName() + ": El farmer " + farmer.getFarmerId() + " es propietario de un campo pero su status es " + operationalStatus);
					}
				} else {
					// Es un arrendatario puro
					if (operationalStatus.equals(OperationalStatus.TenantOnly)) {
						if (operatedFarms.isEmpty()) {
							throw new FulminantError(this.currentCroppingCycle.getCroppingCycleName() + ": El farmer " + farmer.getFarmerId() + " es un Arrendatario puro pero no esta operando ningun campo");
						}
					} else if (operationalStatus.equals(OperationalStatus.OutOfBusiness)) {
						if (! operatedFarms.isEmpty()) {
							throw new FulminantError(this.currentCroppingCycle.getCroppingCycleName() + ": El farmer " + farmer.getFarmerId() + " es un Arrendatario puro que quedo fuera del mercado pero esta operando algun campo = " + operatedFarms);
						} else if (farmer.getWorkingCapital() <= 0) {
							throw new FulminantError(this.currentCroppingCycle.getCroppingCycleName() + ": El farmer " + farmer.getFarmerId() + " es un Arrendatario puro que quedo fuera del mercado pero ingreso con capital negativo = " + farmer.getWorkingCapital());
						}
					} else if (operationalStatus.equals(OperationalStatus.RetiredTenant)) {
						if (! operatedFarms.isEmpty()) {
							throw new FulminantError(this.currentCroppingCycle.getCroppingCycleName() + ": El farmer " + farmer.getFarmerId() + " es un Arrendatario puro retirado pero esta operando algun campo = " + operatedFarms);
						}
					} else if (operationalStatus.equals(OperationalStatus.Broke)) {
						if (! operatedFarms.isEmpty()) {
							throw new FulminantError(this.currentCroppingCycle.getCroppingCycleName() + ": El farmer " + farmer.getFarmerId() + " es un Arrendatario puro que se fue a la quiebra pero esta operando algun campo = " + operatedFarms);
						}
					} else if (operationalStatus.equals(OperationalStatus.Inactive)) {
						if (! operatedFarms.isEmpty()) {
							throw new FulminantError(this.currentCroppingCycle.getCroppingCycleName() + ": El farmer " + farmer.getFarmerId() + " es un Inactivo pero esta operando algun campo = " + operatedFarms);
						}
					} else {
						throw new FulminantError(this.currentCroppingCycle.getCroppingCycleName() + ": El farmer " + farmer.getFarmerId() + " no es propietario de ningun campo pero su status es " + operationalStatus);
					}
				}
			}
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}
	
	/* Luego de calcular los margenes del periodo anterior, paso de ciclo */
	@ScheduledMethod (start = 1.9, interval = 1)
	public void nextCycle() {
		try {
			// 1. Calculo de Global Gross Margin para cada cultivo/manejo/proporcion y 
			//    tipo de suelo (se asume que todos los plots tienen el mismo tipo de suelo)
			Verbosity scenarioVerbosity = this.scenario.getVerbosityLevel();
			DecimalFormat decimalFormat = scenarioVerbosity.getDecimalFormat();		
			Integer farmMaxNumberOfPlots = (Integer) scenario.getParameter("Farm.NumberOfPlots");
			List<SoilType> soilTypes = SoilTypeFactory.getInstance(this.scenario).findAll();	
			for (SoilType soilType: soilTypes) {
				List<ActivityManagementProportion> proportions = this.getActivityManagementProportions(soilType.getRegion());
			
				for (ActivityManagementProportion proportion: proportions) {
					// Se asume que todos los lotes del campo tienen el mismo tamano
					double plotSize = 1.0 / farmMaxNumberOfPlots.doubleValue();
					List<Plot> plots = new ArrayList<Plot>(farmMaxNumberOfPlots);
					List<ActivityManagement> managements = proportion.getAllActivityManagements();					
					for (int i = 0; i < farmMaxNumberOfPlots; i++) {
						plots.add(new Plot(soilType, plotSize, managements.get(i)));
					}
					
					// Ahora calculo el ultimo GGM con todos los datos que necesito
					Double globalGrossMargin = Accountant.getGlobalGrossMargin(plots, 1.0, 1.0, this.currentCroppingCycle.getCroppingCycleId(), this.currentCroppingCycle.getCroppingCycleId());
					
					// Asigno ese GGM al AMP y tipo de suelo
					Map<SoilType, List<Double>> ggmsBySoilType = this.ampGlobalGrossMargins.get(proportion);
					if (ggmsBySoilType == null) {
						// No habia datos para ese AMP. Creo la tabla vacia
						ggmsBySoilType = new HashMap<SoilType, List<Double>>();
						this.ampGlobalGrossMargins.put(proportion, ggmsBySoilType);
					}
					
					List<Double> ggms = ggmsBySoilType.get(soilType);
					if (ggms == null) {
						// No habia datos para ese AMP y tipo de suelo. Creo la lista vacia
						ggms = new ArrayList<Double>();
						ggmsBySoilType.put(soilType, ggms);
					}
					
					// Agrego el GGM a la lista
					ggms.add(globalGrossMargin);
				}
			}
			
			/* 
			 * 2. Output manager's data
			 * 
			 * a. Cropping cycle
			 * b. Soil type
			 * c. AMP Id
			 * d. Estimated global gross margin
			 * 
			 */
			try {
				String delimiter = new String(";");
				File outputDir = new File("output");
				File tmpOutputDir = new File(outputDir, "tmp");
				File outputFile = new File(tmpOutputDir, "manager_scenario_" + this.scenario.getScenarioId() + ".csv");
				if (this.currentCroppingCycle.isFirstCroppingCycle() && outputFile.exists()) outputFile.delete();
				
				PrintWriter writer = new PrintWriter(new FileWriter(outputFile, true));
				// Append headers if neccesary
				if (outputFile.createNewFile()) {
					StringBuilder buffer = new StringBuilder();
					buffer.append("Cropping cycle" + delimiter);				
					buffer.append("AMP Id" + delimiter);
					buffer.append("Soil type Id" + delimiter);				
					buffer.append("Global gross margin");				
					writer.println(buffer.toString());				
				}
				
				// Append data
				List<Region> regions = RegionFactory.getInstance().findAll();
				for (Region region: regions) {
					List<SoilType> regionSoilTypes = SoilTypeFactory.getInstance(this.scenario).findByRegion(region);
					List<ActivityManagementProportion> amps = getActivityManagementProportions(region);
					
					for (ActivityManagementProportion amp: amps) {
						for (SoilType soilType: regionSoilTypes) {
							double ggms[] = this.getLastGlobalGrossMargins(amp, soilType, 1);
							StringBuilder buffer = new StringBuilder();
							buffer.append(this.currentCroppingCycle + delimiter);
							buffer.append(this.ampSortedTable.getActivityManagementProportionId(amp) + delimiter);
							buffer.append(soilType.getSoilTypeId() + delimiter);
							
							Double ggm = (Double) ggms[0];
							if (decimalFormat == null) {
								buffer.append(ggm.toString());
							} else {
								buffer.append(decimalFormat.format(ggm));
								
							}						
							writer.println(buffer.toString());
						}
					}
				}
				
				writer.close();
			} catch (Exception e) {
				throw new FulminantError("Error al generar el output del manager: " + e);
			}
			
			// 3. Ir al proximo ciclo
			long now = java.lang.System.currentTimeMillis();
			double diff = (double) (now - this.checkPoint) / 1000.0;
			java.lang.System.out.println(this.currentCroppingCycle.getCroppingCycleName() + ": Time elapsed = " + diff + " seconds. Moving to the next cropping cycle...");
			this.checkPoint = now;	
			try {
				this.currentCroppingCycle = currentCroppingCycle.nextCycle();			
			} catch (SQLException sqle) {
				throw new FulminantError("Error al buscar los datos de la proxima temporada: " + sqle);
			}
	
			// 4. Si no hay mas ciclos, finalizar la simulacion.
			//    Sino, chequear si al comienzo del ciclo hay innovaciones (en el caso de escenario con difusion de innovaciones)
			//    Para el caso de los manejos que son innovaciones, analizar condicion de saturacion
			if (this.currentCroppingCycle == null) {
				// 5. This was the last cycle. Actions to carry out: 
				/* 
				 * 5a. Output AMP Sorted Table data
				 * 
				 * a. AMP Id
				 * b. AMP Name
				 * 
				 */
				try {
					String delimiter = new String(";");
					File outputDir = new File("output");
					File tmpOutputDir = new File(outputDir, "tmp");
					File outputFile = new File(tmpOutputDir, "amp_scenario_" + this.scenario.getScenarioId() + ".csv");
					if (outputFile.exists()) outputFile.delete();
					
					PrintWriter writer = new PrintWriter(new FileWriter(outputFile, true));
					// Append headers if neccesary
					if (outputFile.createNewFile()) {
						StringBuilder buffer = new StringBuilder();
						buffer.append("AMP Id" + delimiter);
						buffer.append("AMP Name" + delimiter);				
						writer.println(buffer.toString());				
					}
					
					// Append data
					List<Integer> ampIds = this.ampSortedTable.getActivityManagementProportionIds();
					Collections.sort(ampIds);
					for (Integer ampId: ampIds) {
						StringBuilder buffer = new StringBuilder();
						buffer.append(ampId + delimiter);
						buffer.append(this.ampSortedTable.getActivityManagementProportion(ampId) + delimiter);
						writer.println(buffer.toString());
					}
					
					writer.close();
				} catch (Exception e) {
					throw new FulminantError("Error al generar el output del manager: " + e);
				}
				
				fiuba.lamm.agromodel.utils.System.finish();
			} else {
				// 4a. Actualizacion de datos relacionados a tipos de suelo
				for (SoilType soilType: soilTypes) {
					// i. Recalculo de tabla de GGMs de acuerdo a la tecnologia del nuevo ciclo
					//     Ademas, actualizo los expected gross margins para cada tipo de suelo
					if (! scenario.hasAttribute(Scenario.Attribute.DisableTechnologyAdjustmentUpdate)) {
						List<ActivityManagementProportion> proportions = this.getActivityManagementProportions(soilType.getRegion());				
						for (ActivityManagementProportion proportion: proportions) {
							Map<SoilType, List<Double>> ggmsBySoilType = this.ampGlobalGrossMargins.get(proportion);
							List<Double> ggms = ggmsBySoilType.get(soilType);
							
							// Genero un lote virtual para calcular el rendimiento del AMP
							double plotSize = 1.0 / farmMaxNumberOfPlots.doubleValue();
							List<Plot> plots = new ArrayList<Plot>(farmMaxNumberOfPlots);
							List<ActivityManagement> managements = proportion.getAllActivityManagements();					
							for (int i = 0; i < farmMaxNumberOfPlots; i++) {
								plots.add(new Plot(soilType, plotSize, managements.get(i)));
							}
							
							// Ya tengo los GGMs para un dado AMP y tipo de suelo
							// Ahora recalculo esos valores
							for (int i = 0; i < ggms.size(); i++) {							
								Short croppingCycleId = new Short((short) (this.currentCroppingCycle.getCroppingCycleId().intValue() - (i + 1)));						
								int index = ggms.size() - (i + 1);
								Double updatedGlobalGrossMargin = Accountant.getGlobalGrossMargin(plots, 1.0, 1.0, croppingCycleId, this.currentCroppingCycle.getCroppingCycleId());
								ggms.set(index, updatedGlobalGrossMargin);
							}
						}
					}
					
					// ii. Actualizacion de expected gross margin para cada tipo de suelo
					soilType.setCurrentExpectedGrossMargin();
				}
				
				if (scenario.hasAttribute(Scenario.Attribute.HasTechnologyInnovations) || scenario.hasAttribute(Scenario.Attribute.HasGrossMarginPerturbation)) {
					List<Farm> farms = null;
					try {
						farms = FarmFactory.getInstance(scenario).findAll();							
					} catch (SQLException sqle) {
						throw new FulminantError("Error finding innovators: " + sqle);
					}
					
					// 4b. Perturbacion en el outcome factor
					if (scenario.hasAttribute(Scenario.Attribute.HasGrossMarginPerturbation)) {
						for (Farm farm: farms) {
							farm.setNewOutcomeFactor();
						}
					}
					
					// 4c. Difusion de innovaciones (busqueda de nuevas tecnologias)
					if (scenario.hasAttribute(Scenario.Attribute.HasTechnologyInnovations)) {
						List<ActivityManagement> managements = ActivityManagementFactory.getInstance(scenario).findAll();
						for (ActivityManagement management: managements) {
							if (! management.isAvailable()) {
								// Check if it is available in this cropping cycle
								Short firstAvailableCroppingCycleId = management.getFirstAvailableCroppingCycleId();
								if ((firstAvailableCroppingCycleId != null) && firstAvailableCroppingCycleId.equals(this.currentCroppingCycle.getCroppingCycleId())) {
									// A new product is introduced
									management.setIsAvailable(true);
									management.setIsInnovation(true);
									
									// Lo agrego al listado de actividades/manejo desconocidos
									for (Farm farm: farms) {
										List<ActivityManagement> unknownManagements = farm.getUnknownManagements();
										unknownManagements.add(management);
										farm.setUnknownManagements(unknownManagements);
									}							
								}				
							}
						}
					}
				}
			}
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}
}