package fiuba.lamm.agromodel.agents.factories;

import java.sql.*;
import java.util.*;

import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.factories.*;

public class FarmFactory extends Factory {
	protected Map<Integer, List<Farm>> cacheByOwner;
	protected Map<Integer, Farm> cacheByPK;
	protected PlotFactory plotFactory;
	protected static Map<Scenario,FarmFactory> factories;
	
	public static synchronized FarmFactory getInstance(Scenario scenario) throws SQLException {
		if (factories == null) {
			factories = new TreeMap<Scenario,FarmFactory>();
		}
		if (! factories.containsKey(scenario)) {
			factories.put(scenario, new FarmFactory(scenario));
		}
		
		return factories.get(scenario);
	}
	
	protected FarmFactory(Scenario scenario) throws SQLException {
		super();
		this.plotFactory = new PlotFactory();
		this.cacheByPK = new TreeMap<Integer, Farm>();
		this.cacheByOwner = new TreeMap<Integer, List<Farm>>();
		
		FarmerFactory farmerFactory = FarmerFactory.getInstance(scenario);
		SoilTypeFactory soilTypeFactory = SoilTypeFactory.getInstance(scenario);
		CreaGroupFactory creaGroupFactory = CreaGroupFactory.getInstance(scenario);
		ActivityManagementFactory activityManagementFactory = ActivityManagementFactory.getInstance(scenario);
		
		Connection connection = getConnection();
		PreparedStatement st = connection.prepareStatement("select * from farm where scenarioId = ? order by farmId");
		PreparedStatement stPlots = connection.prepareStatement("select * from plot where scenarioId = ? and farmId = ? order by plotNumber");
		st.setShort(1, scenario.getScenarioId());
		ResultSet rs = st.executeQuery();
		while (rs.next()) {
			Region region = RegionFactory.getInstance().findByPrimaryKey(rs.getByte("regionId"));
			Integer ownerId = rs.getInt("ownerId");
			Double originalFarmSize = rs.getDouble("originalFarmSize");
			Farm farm = new Farm(
				rs.getInt("farmId"),
				region,
				null,
				soilTypeFactory.findByPrimaryKey(rs.getByte("soilType")),
				originalFarmSize,
				creaGroupFactory.findByPrimaryKey(rs.getShort("creaGroupId")),
				rs.getByte("cattleStructure")
			);
			
			// Assign owner
			Farmer owner = farmerFactory.findByPrimaryKey(rs.getInt("ownerId"));
			if (owner == null) {
				throw new FulminantError("Invalid farm owner: " + rs.getInt("ownerId")); 
			}
			owner.addOwnedFarm(farm);
			
			// Assign first operator. Aspiration level and plot assignment must be done afterwards
			Farmer operator = farmerFactory.findByPrimaryKey(rs.getInt("initialOperatorId"));
			if (operator != null) operator.addOperatedFarm(farm);
			
			// Set outcome factor if necessary
			if (scenario.hasAttribute(Scenario.Attribute.HasGrossMarginPerturbation)) {
				farm.setNewOutcomeFactor();
			}
			
			// Plots initialization. Check farm size consistency.
			Integer farmNumberOfPlots = (Integer) scenario.getParameter("Farm.NumberOfPlots");
			List<Plot> farmPlots = new ArrayList<Plot>();
			Double plotsTotalSize = 0.0;
			stPlots.setShort(1, scenario.getScenarioId());
			stPlots.setInt(2, farm.getFarmId());
			ResultSet rsPlot = stPlots.executeQuery();
			while (rsPlot.next()) {
				Double plotSize = rsPlot.getDouble("plotSize");
				if (rsPlot.wasNull()) {					
					plotSize = farm.getFarmSize() / farmNumberOfPlots.doubleValue();
				}
				
				Plot plot = new Plot(
					farm,
					rsPlot.getByte("plotNumber"),
					farm.getSoilType(),
					plotSize					
				);
				farmPlots.add(plot);
				plotsTotalSize += plot.getPlotSize();
				
				// Assign first ActivityManagement
				ActivityManagement initialManagement = activityManagementFactory.findByPrimaryKey(rsPlot.getShort("initialActivityMgmtId"));
				plot.setCurrentActivityManagement(initialManagement);
			}
			rsPlot.close();
			if (Math.abs(plotsTotalSize - farm.getFarmSize()) > 1.0) {
				throw new FulminantError("El tamano del farm " + farm.getFarmId() + " es de " + farm.getFarmSize() + " pero la suma de los tamanos de sus lotes es " + plotsTotalSize);
			}
		
			// Assign plots to farm
			plotFactory.setFarmPlots(farm, farmPlots);
							
			// Save cache
			cacheByPK.put(rs.getInt("farmId"), farm);
			if (! cacheByOwner.containsKey(ownerId)) cacheByOwner.put(ownerId, new ArrayList<Farm>());
			List<Farm> ownedFarms = cacheByOwner.get(ownerId);
			ownedFarms.add(farm);
		}
		rs.close();
		st.close();
		stPlots.close();
		connection.close();
	}
	
	public Farm findByPrimaryKey(Integer farmId) {
		return this.cacheByPK.get(farmId);
	}
		
	public List<Farm> findByOwner(Farmer owner)  {
		List<Farm> farms = this.cacheByOwner.get(owner.getFarmerId());
		if (farms == null) farms = new ArrayList<Farm>();
		
		return farms;
	}
	
	public static void clearScenarioCache(Scenario scenario) {
		factories.remove(scenario);
	}
		
	public List<Farm> findAll() {
		return new ArrayList<Farm>(this.cacheByPK.values());
	}
	
	public PlotFactory getPlotFactory() {
		return this.plotFactory;
	}
}