package fiuba.lamm.agromodel.agents.factories;

import java.sql.*;
import java.util.*;
import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.areaupdate.mcp.wta.*;
import fiuba.lamm.agromodel.decision.*;
import fiuba.lamm.agromodel.decision.factories.*;
import fiuba.lamm.agromodel.diffusion.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.factories.*;

public class FarmerFactory extends Factory {
	protected Map<Integer, Farmer> cache;
	protected static Map<Scenario,FarmerFactory> factories;
	
	public static synchronized FarmerFactory getInstance(Scenario scenario) throws SQLException {
		if (factories == null) {
			factories = new TreeMap<Scenario,FarmerFactory>();			
		}
		if (! factories.containsKey(scenario)) {
			factories.put(scenario, new FarmerFactory(scenario));
		}
		
		return factories.get(scenario);
	}
	
	protected FarmerFactory(Scenario scenario) throws SQLException {
		super();		
		this.cache = new TreeMap<Integer, Farmer>();
		
		RotatorTypeFactory rotatorTypeFactory = RotatorTypeFactory.getInstance();
		InnovatorTypeFactory innovatorTypeFactory = InnovatorTypeFactory.getInstance();
		
		Connection connection = getConnection();
		PreparedStatement st = connection.prepareStatement("select * from farmer where scenarioId = ? order by farmerId");
		st.setShort(1, scenario.getScenarioId());
		ResultSet rs = st.executeQuery();
		ResultSetMetaData metaData = null;
		
		while (rs.next()) {
			// Estategia para uso de la tierra (SearchTriggeringMechanism y LandUseSelectionMechanism)
			Map<LandUseSelectionMechanism.Parameter, Object> lusmParameters = new HashMap<LandUseSelectionMechanism.Parameter, Object>();
			Map<SearchTriggeringMechanism.Parameter, Object> stmParameters = new HashMap<SearchTriggeringMechanism.Parameter, Object>();
			if (metaData == null) metaData = rs.getMetaData();
			int count = metaData.getColumnCount();
			int i = 1;
			while (i <= count) {
				String columnName = metaData.getColumnName(i);
				Object columnValue = rs.getObject(i);
				if (columnValue != null) {
					String parameterName = columnName.substring(0, 1).toUpperCase() + columnName.substring(1);
					try {
						LandUseSelectionMechanism.Parameter lusmParameter = LandUseSelectionMechanism.Parameter.valueOf(parameterName);
						lusmParameters.put(lusmParameter, columnValue); 
					} catch (IllegalArgumentException iae) {
						// It's not a LandUseSelectionMechanismParameter
						try {
							SearchTriggeringMechanism.Parameter stmParameter = SearchTriggeringMechanism.Parameter.valueOf(parameterName);
							stmParameters.put(stmParameter, columnValue);
						} catch (IllegalArgumentException iae2) {
							// It's neither a LUSM nor a STM parameter
						}
					}
				}
				i++;
			}			
			SearchTriggeringMechanism searchTriggeringMechanism = SearchTriggeringMechanism.getInstance(rs.getInt("searchTriggeringMechanismId"), stmParameters);
			LandUseSelectionMechanism landUseSelectionMechanism = LandUseSelectionMechanism.getInstance(rs.getInt("landUseSelectionMechanismId"), lusmParameters);
			
			// Estrategia para el calculo del WTA
			int wtaCalculationMechanismId = rs.getInt("wtaCalculationMechanismId");
			WTACalculationMechanism wtaCalculationMechanism = null;
			if (! rs.wasNull()) {
				wtaCalculationMechanism = WTACalculationMechanism.getInstance(wtaCalculationMechanismId);			
			}
			
			// Estrategia para la difusion de innovaciones
			InnovationDiffusionModel diffusionModel = null;
			if (scenario.hasAttribute(Scenario.Attribute.HasTechnologyInnovations)) {
				Map<InnovationDiffusionModel.Parameter, Object> idmParameters = new HashMap<InnovationDiffusionModel.Parameter, Object>();
				i = 1;
				while (i <= count) {
					String columnName = metaData.getColumnName(i);
					Object columnValue = rs.getObject(i);
					if (columnValue != null) {
						String parameterName = columnName.substring(0, 1).toUpperCase() + columnName.substring(1);
						try {
							IsingDiffusionModel.Parameter idmParameter = InnovationDiffusionModel.Parameter.valueOf(parameterName);
							idmParameters.put(idmParameter, columnValue); 
						} catch (IllegalArgumentException iae) {
							// It's not an InnovationDiffusionModel parameter							
						}
					}
					i++;
				}
				diffusionModel = InnovationDiffusionModel.getInstance(scenario, rs.getInt("innovationDiffusionModelId"), idmParameters);
			} 
			
			// Lectura del minimum progress rate
			Double minimumProgressRate = rs.getDouble("minimumProgressRate");
			if (rs.wasNull()) {
				minimumProgressRate = null;
			}
			
			Farmer farmer = new Farmer(
					rs.getInt("farmerId"),
					rs.getString("farmerName"),
					rs.getByte("age"),
					rs.getString("gender"),
					rs.getString("educationalLevel"),
					rs.getByte("familyGroupSize"),
					rs.getDouble("workingCapital"),
					searchTriggeringMechanism,
					landUseSelectionMechanism,
					wtaCalculationMechanism,
					rs.getDouble("managerialAbility"),
					rs.getDouble("offFarmIncome"),
					rs.getShort("yrsExperience"),
					rs.getDouble("peersAdjFactor"),
					rs.getDouble("lambdaUp"),
					rs.getDouble("lambdaDown"),
					rs.getDouble("optimism"),
					rotatorTypeFactory.findByPrimaryKey(rs.getShort("ownerRotatorTypeId")),
					rotatorTypeFactory.findByPrimaryKey(rs.getShort("tenantRotatorTypeId")),
					rs.getDouble("riskAversion"),
					minimumProgressRate,
					innovatorTypeFactory.findByPrimaryKey(rs.getByte("innovatorTypeId")),
					diffusionModel
			);
			
			this.cache.put(farmer.getFarmerId(), farmer);
		}
		rs.close();
		st.close();
		connection.close();
	}
	
	public Farmer findByPrimaryKey(Integer farmerId) {
		return this.cache.get(farmerId);
	}
	

	public static void clearScenarioCache(Scenario scenario) {
		factories.remove(scenario);
	}
	
	public List<Farmer> findAll() {
		return new ArrayList<Farmer>(this.cache.values());
	}
	
	public List<Farmer> findByOperationalStatus(OperationalStatus operationalStatus) {
		List<Farmer> farmers = new ArrayList<Farmer>();
		List<Farmer> all = findAll();
		
		for (Farmer farmer: all) {
			if (farmer.getOperationalStatus().equals(operationalStatus)) {
				farmers.add(farmer);
			}
		}
		
		return farmers;
	}
	
	public List<Farmer> findByOperationalStatus(List<OperationalStatus> operationalStatusList) {
		List<Farmer> farmers = new ArrayList<Farmer>();
		List<Farmer> all = findAll();
		
		for (Farmer farmer: all) {
			if (operationalStatusList.contains(farmer.getOperationalStatus())) {
				farmers.add(farmer);
			}
		}
		
		return farmers;
	}
}