package fiuba.lamm.agromodel.areaupdate.mcp;

import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.agents.*;
import java.util.*;

/**
 * @author santiago
 * 
 * Esta clase se encarga de generar un listado de campos potencialmente alquilables para un dado farmer y un dado conjunto de campos ofrecidos.
 * El proceso se divide en 2 partes:
 * 
 * (1) Filtrado de campos de acuerdo a una serie de criterios.
 * (2) Ordenamiento del listado generado en (1) de acuerdo a otra serie de criterios.
 * 
 * El detalle de los 2 subprocesos es el siguiente:
 * 
 * 1. Se reciben 2 parametros (farmer, farmCandidates)
 * 	  a. Se obtiene el listado de regiones candidatas del farmer (regiones donde puedo alquilar)
 *    b. Para cada campo
 *    		Si el mismo pertenece a una region candidata y tiene una superficie mayor o igual que la MLA de su region, lo incluyo.
 *    		Sino, lo descarto
 *          
 * 2. Una vez que tengo el listado generado del punto (1), ordeno segun el el siguiente criterio:         
 *	  a. Expected gross margin (descendente)
 *    b. Expected operation cost (ascendente)
 */
public class FarmPreferenceListBuilder implements Comparator<Farm> {
	public List<Farm> buildList(Farmer farmer, List<Farm> farmCandidates) {
		List<Farm> allowedFarms = new ArrayList<Farm>();
		
		// 1. Filtrado de campos
		List<Region> candidateRegions = farmer.getCandidateRegions();
		for (Farm farm: farmCandidates) {
			if (candidateRegions.contains(farm.getRegion())) {
				// El campo esta permitido por la region a la que pertenece. Ahora me fijo si su soperficie es aceptable.
				Double mla = farmer.getMinimumLeasableArea(farm.getRegion());
				if (farm.getFarmSize() >= mla) {
					allowedFarms.add(farm);
				}
			}
		}
		
		// 2. Ordenamiento de campos segun el criterio expuesto arriba
		Collections.sort(allowedFarms, this);
		return allowedFarms;
	}
	
	public int compare(Farm farm1, Farm farm2) {
		if (farm1.getExpectedGrossMargin() != farm2.getExpectedGrossMargin()) {
			return (- farm1.getExpectedGrossMargin().compareTo(farm2.getExpectedGrossMargin()));
		} else {
			return farm1.getExpectedOperationCost().compareTo(farm2.getExpectedOperationCost());
		}
	}
}