package fiuba.lamm.agromodel.areaupdate.mcp;

import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.agents.factories.*;
import fiuba.lamm.agromodel.areaupdate.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.factories.*;
import java.util.*;
import repast.simphony.engine.environment.*;
import repast.simphony.engine.schedule.*;

public class MCPCalculationComponent extends LandMarketComponent {
	protected List<SoilType> soilTypes;
	
	private enum TableComponent {
		WTP,
		WTA,
		DemandedSize,
		SuppliedSize;
	}
		
	public MCPCalculationComponent(Double relativeStartTime, Double relativeEndTime) {
		super(relativeStartTime, relativeEndTime);
		this.soilTypes = SoilTypeFactory.getInstance(Manager.getInstance().getScenario()).findAll();				
	}
		
	public void scheduleNextAction() {
		// Esta clase es la que se encarga de hacer el calculo del MCP
		ISchedule schedule = RunEnvironment.getInstance().getCurrentSchedule();
		Double tick = schedule.getTickCount();
		ScheduleParameters parameters = ScheduleParameters.createOneTime(tick + this.relativeStartTime);		
		schedule.schedule(parameters, this, "execute");		
	}	
	
	public void execute() {
		try {
			Manager manager = Manager.getInstance();
			
			// 1. Por cada tipo de suelo, genero 2 listas
			//    Oferta: WTA y tamano del campo que se ofrece
			//    Demanda: WTP y capital del que se dispone para alquilar		
			Map<SoilType,List<Map<TableComponent,Double>>> supply = new HashMap<SoilType,List<Map<TableComponent,Double>>>();
			Map<SoilType,List<Map<TableComponent,Double>>> demand = new HashMap<SoilType,List<Map<TableComponent,Double>>>();
			List<Farmer> farmers = FarmerFactory.getInstance(manager.getScenario()).findAll();
			
			for (Farmer farmer: farmers) {
				// WTA
				Map<Farm, Double> wtas = farmer.getWillingnessToAccept();
				if (wtas != null) {
					// Tiene campos y esta dispuesto a ponerlos en alquiler
					for (Farm farm: wtas.keySet()) {
						Double wta = wtas.get(farm);
						SoilType soilType = farm.getSoilType();
						
						Map<TableComponent,Double> supplyData = new HashMap<TableComponent,Double>();
						supplyData.put(TableComponent.WTA, wta);
						supplyData.put(TableComponent.SuppliedSize, farm.getFarmSize());
						
						if (! supply.containsKey(soilType)) {
							supply.put(soilType, new ArrayList<Map<TableComponent,Double>>());
						}
						List<Map<TableComponent,Double>> data = supply.get(soilType);
						data.add(supplyData);
						supply.put(soilType, data);
					}
				}
				
				
				// WTP
				Double availableCapital = farmer.getWorkingCapital();
				for (Farm farm: farmer.getOwnedFarms()) {
					// Es un propietario. El capital disponible es el que tiene menos el costo de implantacion de sus propios campos
					availableCapital -= farm.getExpectedOperationCost(farmer);
				}
				Map<SoilType,Double> wtps = farmer.getWillingnessToPay();
				if (wtps != null) {
					// Esta dispuesto a arrendar un campo en al menos un tipo de suelo
					Set<SoilType> soilTypes = wtps.keySet();
					for (SoilType soilType: soilTypes) {
						Double wtp = wtps.get(soilType);
						if (wtp != null) {
							// Esta dispuesto a arrendar un campo en ese tipo de suelo
							Map<TableComponent,Double> demandData = new HashMap<TableComponent,Double>();
							demandData.put(TableComponent.WTP, wtp);
							demandData.put(TableComponent.DemandedSize, availableCapital / wtp);
							
							if (! demand.containsKey(soilType)) {
								demand.put(soilType, new ArrayList<Map<TableComponent,Double>>());
							}
							List<Map<TableComponent,Double>> data = demand.get(soilType);
							data.add(demandData);
							demand.put(soilType, data);
						}
					}
				}					
			}
			
			// 2. Calcular el MCP para cada tipo de suelo
			for (SoilType soilType: this.soilTypes) {
				List<Map<TableComponent,Double>> soilTypeSupply = supply.get(soilType);
				List<Map<TableComponent,Double>> soilTypeDemand = demand.get(soilType);
				
				if ((soilTypeSupply != null) && (soilTypeDemand != null)) {
					List<Double> mcp = new ArrayList<Double>();
					Double maxTransactions = 0.0;
					
					// Hay valores de oferta y demanda para ese tipo de suelo.
					List<Double> candidates = new ArrayList<Double>();
					for (Map<TableComponent,Double> supplyData: soilTypeSupply) {
						Double supplyCandidate = supplyData.get(TableComponent.WTA);
						if (! candidates.contains(supplyCandidate)) {
							candidates.add(supplyCandidate);
						}
					}
					for (Map<TableComponent,Double> demandData: soilTypeDemand) {
						Double demandCandidate = demandData.get(TableComponent.WTP);
						if (! candidates.contains(demandCandidate)) {
							candidates.add(demandCandidate);
						}
					}				
					Collections.sort(candidates);
					
					if (! candidates.isEmpty()) {
						Double minPrice = candidates.get(0);
						Double maxPrice = candidates.get(candidates.size() - 1);
						Double priceStep = 1.0;
						
						for (Double price = minPrice; price <= maxPrice; price += priceStep) {
							Double suppliedArea = 0.0;
							Double demandedArea = 0.0;
							
							for (Map<TableComponent,Double> supplyData: soilTypeSupply) {
								Double wta = supplyData.get(TableComponent.WTA);
								if (price >= wta) {
									// El precio posible de alquiler le satisface al propietario, asi que aceptaria alquilar todo su campo
									suppliedArea += supplyData.get(TableComponent.SuppliedSize);
								}
							}
							for (Map<TableComponent,Double> demandData: soilTypeDemand) {
								Double wtp = demandData.get(TableComponent.WTP);
								if (price <= wtp) {
									// El precio posible de alquiler le satisface al arrendatario, asi que aceptaria alquilar con todo su capital
									demandedArea += demandData.get(TableComponent.DemandedSize);
								}
							}
											
							// Si la cantidad de transacciones con ese precio es mayor que el maximo, entonces ese es el nuevo MCP
							Double transactions = Math.min(suppliedArea, demandedArea);
							if (transactions > maxTransactions) {
								maxTransactions = transactions;
								mcp.clear();
								mcp.add(price);
							} else if (transactions.equals(maxTransactions)) {
								mcp.add(price);
							}
						}
					}
					
					// Establezco el precio del suelo en base al MCP
					if (! mcp.isEmpty()) {
						Double averageMcp = 0.0;
						for (Double price: mcp) {
							averageMcp += price;
						}
						averageMcp /= (double) mcp.size();
						
						soilType.setRentalPrice(averageMcp);
						if (maxTransactions == 0.0) {
							System.out.println(manager.getCurrentCroppingCycle() + ": No hay match entre la oferta y la demanda para el tipo de suelo " + soilType + ". Se tomara un promedio del rango de precios.");
						}
					} else {
						throw new FulminantError("No hay candidatos de precios para el ciclo " + manager.getCurrentCroppingCycle() + " y tipo de suelo " + soilType + ". Oferta = " + soilTypeSupply + "; Demanda = " + soilTypeDemand);
					}
				}
			}
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}		
}