package fiuba.lamm.agromodel.areaupdate.mcp;

import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.agents.factories.*;
import fiuba.lamm.agromodel.areaupdate.*;
import fiuba.lamm.agromodel.execution.*;

import java.sql.*;
import java.util.*;

import repast.simphony.engine.environment.*;
import repast.simphony.engine.schedule.*;

public class SupplyDemandCreationComponent extends LandMarketComponent {
	protected List<Farm> supplyList;
	protected List<Farmer> activeDemandList;
	protected List<Farmer> inactiveDemandList;
	
	public SupplyDemandCreationComponent(Double relativeStartTime, Double relativeEndTime) {
		super(relativeStartTime, relativeEndTime);
		
		this.supplyList = new ArrayList<Farm>();
		this.activeDemandList = new ArrayList<Farmer>();
		this.inactiveDemandList = new ArrayList<Farmer>();
	}
	
	public List<Farm> getSupplyList() {
		return this.supplyList;
	}
	
	public List<Farmer> getActiveDemandList() {
		return this.activeDemandList;
	}
	
	public List<Farmer> getInactiveDemandList() {
		return this.inactiveDemandList;
	}
	
	public void scheduleNextAction() {
		Manager manager = Manager.getInstance();
		ISchedule schedule = RunEnvironment.getInstance().getCurrentSchedule();
		Double tick = schedule.getTickCount();
		Double nextTickStage1 = tick + this.relativeStartTime;
		
		// 1. Agregar a la lista de oferta todos aquellos campos que hayan quedado sin operador
		ScheduleParameters parametersStage1 = ScheduleParameters.createOneTime(nextTickStage1);
		schedule.schedule(parametersStage1, this, "executeStage1");
				
		// 2. En principio todos los farmers participan 
		List<Farmer> farmers = null;
		try {
			FarmerFactory farmerFactory = FarmerFactory.getInstance(manager.getScenario());
			farmers = farmerFactory.findAll();
		} catch (SQLException sqle) {
			throw new FulminantError("Error al buscar los productores: " + sqle);
		}
		
		Double nextTickStage2 = tick + (this.relativeStartTime + this.relativeEndTime) / 2.0;
		ScheduleParameters parametersStage2 = ScheduleParameters.createOneTime(nextTickStage2);
		for (Farmer farmer: farmers) {
			schedule.schedule(parametersStage2, this, "executeStage2", farmer);
		}		
	}

	public void executeStage1() {
		try {
			Manager manager = Manager.getInstance();
			try {
				List<Farm> farms = FarmFactory.getInstance(manager.getScenario()).findAll();
				for (Farm farm: farms) {
					if (farm.getOperator() == null) {
						this.supplyList.add(farm);
					}
				}
			} catch (SQLException sqle) {
				throw new FulminantError("Error al buscar los campos: " + sqle);
			}
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}
	
	public void executeStage2(Farmer farmer) {
		try {
			Manager manager = Manager.getInstance();
			boolean checkRentedFarms = false;
		
			if (farmer.getOperationalStatus().equals(OperationalStatus.OwnerOnly) || farmer.getOperationalStatus().equals(OperationalStatus.OwnerTenant)) {
				// Propietarios. Si formo WTA entonces ver si decide ponerlo en alquiler
				Map<Farm,Double> wtas = farmer.getWillingnessToAccept();			
				boolean becomesLandLord = false;
				if (wtas != null) {
					/*
					 * La manera de decidir si se transforma en LandLord o no es la siguiente:
					 * 
					 * expected = sumatoria(wta[i] * size[i]);
					 * offered  = sumatoria(mcp[i] * size[i]);
					 * 
					 * si (offered >= expected) {
					 *    paso a land lord
					 * } sino {
					 *    sigo explotando mis campos
					 * }
					 */
					Double expected = 0.0;
					Double offered  = 0.0;
					for (Farm farm: farmer.getOwnedFarms()) {
						SoilType soilType = farm.getSoilType();
						offered  += soilType.getRentalPrice() * farm.getFarmSize();
						expected += wtas.get(farm) * farm.getFarmSize();
					}				
					becomesLandLord = (offered >= expected);				
				}
				
				if (becomesLandLord) {
					// Deja todos sus campos y se convierte en LandLord
					List<Farm> operatedFarms = farmer.getOperatedFarms();
					for (Farm farm: operatedFarms) {
						farmer.removeOperatedFarm(farm, false);
						farm.setDefaultAspirationLevel();
						
						// El campo va a la lista de oferta
						if (! this.supplyList.contains(farm)) {
							this.supplyList.add(farm);
						}
					}
	
					// Agrego un ciclo como land lord
					farmer.addLiveOffCroppingCycle();
				} else {
					// Sigue al menos operando sus campos.
					// Si es owner-only entonces va a la lista de demanda siempre y cuando le quede capital luego de implantar sus propios campos.
					// Si el owner-tenant, primero debe definir que campos sigue alquilando, y si le queda capital va a la lista de demanda.
					if (farmer.getOperationalStatus().equals(OperationalStatus.OwnerOnly)) {
						Double workingCapital = farmer.getWorkingCapital();
						Double ownFarmsExpectedOperationCost = 0.0;
						for (Farm farm: farmer.getOwnedFarms()) {
							ownFarmsExpectedOperationCost += farm.getExpectedOperationCost(farmer);
						}
						if (workingCapital > ownFarmsExpectedOperationCost) {
							this.activeDemandList.add(farmer);
						}
					} else {
						checkRentedFarms = true;
					}
				}
			} else if (farmer.getOperationalStatus().equals(OperationalStatus.TenantOnly)) {
				// Los tenant-only solo pueden definir que hacer viendo que campos retienen y el capital sobrante
				checkRentedFarms = true;
			} else if (farmer.getOperationalStatus().equals(OperationalStatus.Inactive)) {
				// Add the farmer to the list of inactive demand
				this.inactiveDemandList.add(farmer);			
			}
			
			// Para aquellos que alquilan campos, ver con que se pueden quedar luego de modificarse el precio de alquiler
			if (checkRentedFarms) {
				CroppingCycle croppingCycle = manager.getCurrentCroppingCycle();
							
				// For now, we don't use scale functions
				double fixedCostScaleFactor = 1.0;
				
				// Calculate expected costs and gross margins for each operated farm
				List<Farm> operatedFarms = farmer.getOperatedFarms();
				List<Farm> selectableFarms = new ArrayList<Farm>(operatedFarms.size());
				double totalOperationCost = 0.0;
				for (int i = 0; i < operatedFarms.size(); i++) {
					Farm farm = operatedFarms.get(i);
					farm.calculateExpectedValues(croppingCycle, fixedCostScaleFactor);
					totalOperationCost += farm.getExpectedOperationCost(farmer);
					selectableFarms.add(farm);
				}
			
				boolean wantsToLease = false;
				List<Farm> offeredFarms = null;
				if (farmer.getWorkingCapital() > totalOperationCost) { // Le sobra plata
					// This farmer is a potential leaser
					wantsToLease = true;								
				} else if (farmer.getWorkingCapital() < totalOperationCost) {
					// This farmer must stop operating some farms
					// He must choose which farms to keep
					offeredFarms = new ArrayList<Farm>(operatedFarms);
					FarmSubSetCollectionGenerator generator = null;
					List<Farm> keptFarms = null;
					
					if (farmer.getWorkingCapital() > 0) {
						// Create every possible combination or farms
						if (operatedFarms.size() > 7) {
							// Use fast generator
							generator = new FarmSubSetFastCollectionGenerator(farmer, operatedFarms);
						} else {
							// Use brute force generator
							generator = new FarmSubSetFullCollectionGenerator(farmer, operatedFarms);
						}
						
						// If the farmer owns some farms, every possible combination MUST contain those farms
						List<FarmSubSetCollection> collection = generator.getCollectionsList();
						List<Farm> ownedFarms = farmer.getOwnedFarms();
						if (! ownedFarms.isEmpty()) {
							List<FarmSubSetCollection> allowedCollections = new ArrayList<FarmSubSetCollection>();
							for (int i = 0; i < collection.size(); i++) {
								FarmSubSetCollection c = collection.get(i);
								if (c.containsAll(ownedFarms)) allowedCollections.add(c);
							}
							
							collection = allowedCollections;
						}
						
						// Find which combinations are possible
						Iterator<FarmSubSetCollection> farms = collection.iterator();
						while (farms.hasNext() && (keptFarms == null)) {
							FarmSubSetCollection c = farms.next();
							if (c.getExpectedOperationCost() <= farmer.getWorkingCapital()) {
								totalOperationCost = c.getExpectedOperationCost();
								keptFarms = c.getFarms();
							}
						}
					}
					
					if (keptFarms != null) {
						// He keeps some farms
						
						for (int i = 0; i < selectableFarms.size(); i++) {
							Farm farm = selectableFarms.get(i);
							if (keptFarms.contains(farm)) {
								// Remove form offer list
								offeredFarms.remove(farm);							
							} else {
								// Remove from operated farms list and set default aspiration level
								farmer.removeOperatedFarm(farm, false);
								farm.setDefaultAspirationLevel();
							}
						}
						
						wantsToLease = (farmer.getWorkingCapital() > totalOperationCost);
					} else {
						// Set default aspiration level for all offered farms
						// Also, stop working all the farms
						for (int i = 0; i < offeredFarms.size(); i++) {
							Farm farm = offeredFarms.get(i);
							farmer.removeOperatedFarm(farm, false);
							farm.setDefaultAspirationLevel();						
						}
											
						// If the farmer owns some farms, he goes to LiveOff status until he can regain control of his farms
						if (farmer.getOwnedFarms().isEmpty() && (farmer.getWorkingCapital() > 0)) {
							// Farmer is not owner. If he has liquid capital, he may lease a smaller farm
							farmer.setOperationalStatus(OperationalStatus.TenantOnly);
							wantsToLease = true;						
						}					
					}				
				}
				
				if (wantsToLease) this.activeDemandList.add(farmer);
				if (offeredFarms != null) this.supplyList.addAll(offeredFarms);
			}		
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}
}