package fiuba.lamm.agromodel.areaupdate.mcp;

import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.areaupdate.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.utils.*;
import java.util.*;
import repast.simphony.engine.environment.*;
import repast.simphony.engine.schedule.*;

public class SupplyDemandMatchingComponent extends LandMarketComponent {
	protected LandMarketModel landMarketModel;
	
	public SupplyDemandMatchingComponent(LandMarketModel landMarketModel, Double relativeStartTime, Double relativeEndTime) {
		super(relativeStartTime, relativeEndTime);		
		this.landMarketModel = landMarketModel;
	}
	
	public void scheduleNextAction() {
		// Esta clase es la que se encarga de hacer el "matching" de oferta y demanda
		ISchedule schedule = RunEnvironment.getInstance().getCurrentSchedule();
		Double tick = schedule.getTickCount();
		ScheduleParameters parameters = ScheduleParameters.createOneTime(tick + this.relativeStartTime);		
		schedule.schedule(parameters, this, "execute");		
	}

	public void execute() {
		try {
			List<Farm> supplyList = this.landMarketModel.getSupplyList();
			List<Farmer> activeDemandList = this.landMarketModel.getActiveDemandList();
			List<Farmer> inactiveDemandList = this.landMarketModel.getInactiveDemandList();
			FarmPreferenceListBuilder farmPreferenceListBuilder = new FarmPreferenceListBuilder();
			Manager manager = Manager.getInstance();
			Scenario scenario = manager.getScenario();
			Randomizer randomizer = scenario.getRandomizer();
			boolean useWealthWeighedSelectionMechanism = scenario.hasAttribute(Scenario.Attribute.HasWealthWeightedSelectionMechanism);
					
			/*
			 *  1. Chequeo de consistencia: Ningun owner puede estar en la lista de demanda activa si no opera sus propios campos
			 */
			for (Farmer farmer: activeDemandList) {
				if (! farmer.getOperatedFarms().isEmpty()) {
					for (Farm farm: farmer.getOwnedFarms()) {
						if (! farmer.getOperatedFarms().contains(farm)) {
							throw new FulminantError(manager.getCurrentCroppingCycle().getCroppingCycleName() + ": El farmer " + farmer.getFarmerId() + " ha ingresado en la lista de demanda pero no esta operando su propio campo. OS = " + farmer.getOperationalStatus() + "; WTA = " + farmer.getWillingnessToAccept() + "; WTP = " + farmer.getWillingnessToPay());
						}
					}
				}
			}
			
			// 2. Match both lists using discussed criteria
			//    Stop when supply list is empty
			boolean stopAreaManagement = false;
			while (! supplyList.isEmpty() && !stopAreaManagement) {
				if (activeDemandList.isEmpty()) {
					// 3. List of potential leasers is empty
					//
					//	  i. Pick as many inactive farmers as leasable farms
					//    ii. Match each farmer with one farm (farmer.addOperatedFarm(leasableFarm))
					//    iii. Set farmer's liquid capital equals to farm expected operation cost
					//    iv. Empty list of leasable farms
	
					while (! supplyList.isEmpty() && !stopAreaManagement) {
						Farm farm = supplyList.remove(0);
						
						// Farmer is picked at random
						if (inactiveDemandList.isEmpty()) {
							// There are no inactive farmers left but, there are unassigned farms.
							// When this condition is met, simulation must end
							Map<Region, Integer> farmsByRegion = new HashMap<Region, Integer>();
							for (Farm f: supplyList) {
								Region region = f.getRegion();
								if (! farmsByRegion.containsKey(region)) {
									farmsByRegion.put(region, 1);
								} else {
									farmsByRegion.put(region, farmsByRegion.get(region) + 1);
								}
							}
							String msg = "There are no inactive farmers left, but there are still " + supplyList.size() + " unassigned farms";
							for (Region region: farmsByRegion.keySet()) {
								msg += ". " + region.getRegionName() + " farms = " + farmsByRegion.get(region);
							}
							msg += ". Outputting results and exiting.";
													
							// Output results and end simulation
							stopAreaManagement = true;
							manager.outputResults();
							RunEnvironment runEnvironment = RunEnvironment.getInstance();
							runEnvironment.endRun();
							
							throw new FulminantError(msg);
						} else {
							int index = (int) Math.floor(randomizer.getRandomDouble() * (double) inactiveDemandList.size());
							Farmer farmer = inactiveDemandList.remove(index);
							
							farmer.setWorkingCapital(farm.getExpectedOperationCost(farmer));					
							farmer.addOperatedFarm(farm);
							farm.setDefaultAspirationLevel();
						}
					}				
				} else {			
					// 4.1 Pick a Farmer.
					Farmer farmer = null;				
					if (useWealthWeighedSelectionMechanism) {
						// Each Farmer should be picked at random with a probability proportional
						// to his liquid capital
						Double totalCapital = 0.0;
						for (Farmer f: activeDemandList) {
							totalCapital += f.getExpectedWorkingCapital();
						}
						
						Map<Farmer,Double> distribution = new HashMap<Farmer,Double>();
						for (Farmer f: activeDemandList) {
							distribution.put(f, f.getExpectedWorkingCapital() / totalCapital);
						}
						
						farmer = randomizer.getRandomEvent(distribution);
					} else {
						// Pick at random
						int index = randomizer.getRandomInt(activeDemandList.size());
						farmer = activeDemandList.get(index);
					}
					
					// 4.2 After the farmed is picked, he should find a a farm the matches
					//     his expectations.
					Farm leasedFarm = null;	
					List<Farm> farmerSupplyList = farmPreferenceListBuilder.buildList(farmer, supplyList);
					int i = 0;
					boolean noFarmsLeft = (i == farmerSupplyList.size());
					double farmerExpectedWorkingCapital = farmer.getExpectedWorkingCapital();
					while ((leasedFarm == null) && !noFarmsLeft) {
						Farm farm = farmerSupplyList.get(i);
						if (farmerExpectedWorkingCapital >= farm.getExpectedOperationCost(farmer)) {
							leasedFarm = farm;
						} else {
							i++;
							noFarmsLeft = (i == farmerSupplyList.size());							
						}					
					}
					
					// 4.3 If a farm is leased, add it to the farmer's list of operated farms (also remove the farm from the list of leasable farms)
					//     If he has ran out of expected liquid capital, remove it from the list of potential leasers
					//     If no farm is leased, remove the farmer from potential leasers
					if (leasedFarm != null) {
						farmerSupplyList.remove(leasedFarm);
						supplyList.remove(leasedFarm);
						farmer.addOperatedFarm(leasedFarm);
						leasedFarm.setDefaultAspirationLevel();
						
						if (farmer.getExpectedWorkingCapital() <= 0) {
							activeDemandList.remove(farmer);
						}
					} else {
						activeDemandList.remove(farmer);
						farmer.updateOperationalStatus(false);					
					}
				}
			}
					
			// 5. Me fijo si hay farmers que no tuvieron la chance de alquilar campos.
			//    De ellos, los que quedan con 0 campos pasan a Out Of Business
			for (Farmer farmer: activeDemandList) {
				List<Farm> operatedFarms = farmer.getOperatedFarms();
				if (operatedFarms.isEmpty()) farmer.setOperationalStatus(OperationalStatus.OutOfBusiness);
			}
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}
}