package fiuba.lamm.agromodel.areaupdate.mcp;

import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.agents.factories.*;
import fiuba.lamm.agromodel.areaupdate.*;
import fiuba.lamm.agromodel.areaupdate.mcp.pe.*;
import fiuba.lamm.agromodel.areaupdate.mcp.wta.*;
import fiuba.lamm.agromodel.decision.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.factories.*;
import fiuba.lamm.agromodel.finance.*;
import fiuba.lamm.agromodel.math.*;
import fiuba.lamm.agromodel.math.custom.*;
import fiuba.lamm.agromodel.utils.*;
import java.util.*;
import repast.simphony.engine.environment.*;
import repast.simphony.engine.schedule.*;
import repast.simphony.space.graph.*;

public class WillingnessValuesFormationComponent extends LandMarketComponent {
	protected ProgressEvaluationMechanism progressEvaluationMechanism;
	protected List<SoilType> soilTypes;
	
	public WillingnessValuesFormationComponent(Double relativeStartTime, Double relativeEndTime) {
		super(relativeStartTime, relativeEndTime);
		
		// Defino el mecanismo de evaluacion de progreso para los farmers
		this.progressEvaluationMechanism = ProgressEvaluationMechanism.getInstance();
		this.soilTypes = SoilTypeFactory.getInstance(Manager.getInstance().getScenario()).findAll();					
	}
		
	public void scheduleNextAction() {
		Manager manager = Manager.getInstance();
		CroppingCycle croppingCycle = manager.getCurrentCroppingCycle();
		Scenario scenario = manager.getScenario();
		
		// 1. Modifico el valor de las variables correspondientes al mercado de alquiler para cada farmer.
		try {
			FarmerFactory farmerFactory = FarmerFactory.getInstance(manager.getScenario());
			List<Farmer> farmers = farmerFactory.findAll();
			boolean allowFarmersMigration = scenario.hasAttribute(Scenario.Attribute.AllowFarmersMigration);
			boolean nOutOfMEvaluationMechanism = scenario.hasAttribute(Scenario.Attribute.ProgressRateNOutOfM);
			
			// Actualizacion de variables de mercado
			for (Farmer farmer: farmers) {				
				farmer.updateLandMarketVariables(allowFarmersMigration, nOutOfMEvaluationMechanism);			
			}
			
			// Actualizacion de progress rates y minimum progress rates (solo para N out of M)
			if (nOutOfMEvaluationMechanism) {				
				Integer croppingCyclesForProgressEvaluation = (Integer) scenario.getParameter("ProgressEvaluation.CroppingCyclesForProgressEvaluation");
				Integer minimumPositiveCycles = (Integer) scenario.getParameter("ProgressEvaluation.MinimumPositiveCycles");
				NOutOfMEvaluationMechanism evaluationMechanism = new NOutOfMEvaluationMechanism(croppingCyclesForProgressEvaluation, minimumPositiveCycles);
				
				for (Farmer farmer: farmers) {
					if (farmer.getOperationalStatus().isActive() && (farmer.getOwnMinimumProgressRate() != null)) {
						Double progressRate = evaluationMechanism.evaluateCurrentProgress(farmer);						
						farmer.setCurrentProgressRate(progressRate);
					}
				}
				
				for (Farmer farmer: farmers) {
					if (farmer.getOperationalStatus().isActive() && (farmer.getOwnMinimumProgressRate() != null)) {						
						Double minimumProgressRate = evaluationMechanism.calculateMinimumProgressRate(farmer);
						farmer.setCurrentMinimumProgressRate(minimumProgressRate);
					}
				}
			}
		} catch (Exception e) {
			throw new FulminantError("Error al setear en null los valores de WTA y WTP: " + e);			
		}
		
		// 2. Ademas calculo los valores esperados de operacion de cada campo
		try {
			FarmFactory farmFactory = FarmFactory.getInstance(manager.getScenario());
			List<Farm> farms = farmFactory.findAll();
			
			double fixedCostScaleFactor = 1.0;			
			for (Farm farm: farms) {
				farm.calculateExpectedValues(croppingCycle, fixedCostScaleFactor);				
			}
		} catch (Exception e) {
			throw new FulminantError("Error al actualizar los niveles de aspiracion por comparacion de contextos: " + e);
		}
		
		// 3. Buscar todos los owner-tenants y tenants-only para actualizar sus expectativas
		//    Estos deben hacerlo primero ya que pueden ser echados por los land lords
		try {
			FarmerFactory farmerFactory = FarmerFactory.getInstance(manager.getScenario());
			List<Farmer> tenantFarmers = farmerFactory.findByOperationalStatus(Arrays.asList(OperationalStatus.OwnerTenant, OperationalStatus.TenantOnly));
			
			ISchedule schedule = RunEnvironment.getInstance().getCurrentSchedule();
			Double tick = schedule.getTickCount();
			ScheduleParameters parameters = ScheduleParameters.createOneTime(tick + this.relativeStartTime);
			
			for (Farmer tenantFarmer: tenantFarmers) {
				schedule.schedule(parameters, this, "executeStage1", tenantFarmer);
			}			
		} catch (Exception e) {
			throw new FulminantError("Error al buscar los owner/tenans y tenants only para actualizar expectativas de precios: " + e);
		}
		
		// 4. Buscar a todos los land-lords (retirados o no) y owners-only para que actualicen sus expectativas		
		try {
			FarmerFactory farmerFactory = FarmerFactory.getInstance(manager.getScenario());
			List<Farmer> ownerFarmers = farmerFactory.findByOperationalStatus(Arrays.asList(OperationalStatus.LandLord, OperationalStatus.OwnerOnly, OperationalStatus.RetiredOwner));
			
			ISchedule schedule = RunEnvironment.getInstance().getCurrentSchedule();
			Double tick = schedule.getTickCount();
			Double nextTick = tick + this.relativeStartTime + ((this.relativeEndTime - this.relativeStartTime) / 2.0);
			ScheduleParameters parameters = ScheduleParameters.createOneTime(nextTick);
			
			for (Farmer ownerFarmer: ownerFarmers) {
				schedule.schedule(parameters, this, "executeStage2", ownerFarmer);
			}			
		} catch (Exception e) {
			throw new FulminantError("Error al buscar los owners para actualizar expectativas de precios: " + e);
		}		
	}
	
	/**
	 * Actualizacion de expectativas:
	 * 
	 * 1) Si es un tenant/only, entonces me fijo que al menos pueda seguir operando un campo teniendo en cuenta el costo de implantacion.
	 * 		1a. Si no puede siquiera pagar el costo de implantacion, se va a la quiebra y no necesita formar expectativas
	 * 		1b. Si puede pagar el costo de implantacion de al menos uno, entonces calcular WTP
	 * 
	 * 2) Si es un owner/tenant, entonces me fijo si puede implantar al menos su propio campo
	 *      2a. Si no puede, entonces deja su campo, calcula WTA y se convierte en Land Lord
	 *      2b. Si puede, entonces calcula su progress rate
	 *      		Si no esta conforme, entonces calcula WTA (pero no cambia de estado)
	 *      		Si esta conforme 
	 *      			Si puede implantar la SMA en ALGUNA de las regiones entonces calcula WTP
	 *      			Sino, deja todos sus campos alquilados y se convierte en Owner-Only.
	 */
	public void executeStage1(Farmer farmer) {
		try {
			boolean needsToFormWTA = false;
			boolean needsToFormWTP = false;
			
			Iterator<Region> regions = farmer.getCandidateRegions().iterator();
			Manager manager = Manager.getInstance();
			CroppingCycle croppingCycle = manager.getCurrentCroppingCycle();
			
			if (farmer.getOperationalStatus().equals(OperationalStatus.TenantOnly)) {
				// Es un tenant-only
				// Si es un tenant-only y no puede ni siquiera implantar el area minima en ninguna region, entonces se funde
				// Me fijo si puede pagar el costo de implantacion en alguna de las regiones posibles
				boolean canContinue = false;
				while (regions.hasNext() && !canContinue) {
					Region region = regions.next();
					Double minimumLeasableArea = farmer.getMinimumLeasableArea(region);				
					Double expectedOperationCostPerHA = region.getExpectedOperationCost(croppingCycle) * farmer.getManagerialAbility(); // Utilizo el managerial ability del farmer y un factor de escala de 1 (ya que asumo que se queda sin nada)
					Double expectedOperationCost = expectedOperationCostPerHA * minimumLeasableArea;
					canContinue = (expectedOperationCost < farmer.getWorkingCapital());
				}			
				
				if (canContinue) {
					// Puede continuar, entonces debe calcular el WTP
					needsToFormWTP = true;
				} else {
					// No puede seguir operando aun sin conocer el precio de alquiler. Se va a la quiebra				
					for (Farm farm: farmer.getOperatedFarms()) {
						farmer.removeOperatedFarm(farm, false);
						farm.setDefaultAspirationLevel();
					}
				}
			} else {
				// Es un owner-tenant. Primero me fijo si puede operar sus propios campos.
				// Si no puede, pasa a ser land lord		
				Double ownFarmsExpectedOperationCost = 0.0;
				Double ownFarmsTotalSize = 0.0;
				for (Farm farm: farmer.getOwnedFarms()) {
					ownFarmsExpectedOperationCost += farm.getExpectedOperationCost(farmer);
					ownFarmsTotalSize += farm.getFarmSize();
				}
				Double workingCapital = farmer.getWorkingCapital();
						
				if (workingCapital >= ownFarmsExpectedOperationCost) {
					// Puede seguir operando sus campos (todos juntos)
					// Calculo su progress rate
					boolean isProgressAcceptable = this.progressEvaluationMechanism.evaluate(farmer);
					if (! isProgressAcceptable) {
						// No esta conforme, entonces formo WTA
						needsToFormWTA = true;
					} else {
						// Esta conforme con su progreso
						// Me fijo si puedo pagar el costo de implantacion de la MLA en alguna de las regiones posibles
						Function fcScalingFunction = new FixedCostScalingFunction();
						Double fixedCostScaleFactor = fcScalingFunction.apply(ownFarmsTotalSize);
						boolean canOperateMoreFarms = false;
						while (regions.hasNext() && !canOperateMoreFarms) {
							Region region = regions.next();
							Double minimumLeasableArea = farmer.getMinimumLeasableArea(region);
							Double expectedOperationCostPerHA = region.getExpectedOperationCost(croppingCycle) * farmer.getManagerialAbility() * fixedCostScaleFactor; // Utilizo el managerial ability del farmer y un factor de escala segun el tamano de sus propios campos
							Double expectedOperationCost = expectedOperationCostPerHA * minimumLeasableArea;
							canOperateMoreFarms = (expectedOperationCost < (workingCapital - ownFarmsExpectedOperationCost));
						}
						
						if (canOperateMoreFarms) {
							needsToFormWTP = true;
						} else {
							// Deja los campos alquilados y se convierte en Owner-Only
							List<Farm> operatedFarms = farmer.getOperatedFarms();
							for (Farm farm: operatedFarms) {
								if (! farm.getOwner().equals(farmer)) {
									farmer.removeOperatedFarm(farm, false);
									farm.setDefaultAspirationLevel();
								}
							}
						}
					}				
				} else {
					// Debe dejar su propio campo y todos los demas que este operando y pasar a ser land lord
					List<Farm> operatedFarms = farmer.getOperatedFarms();
					for (Farm farm: operatedFarms) {
						farmer.removeOperatedFarm(farm, false);
						farm.setDefaultAspirationLevel();				
					}
					
					needsToFormWTA = true;
				}
			}
			
			if (needsToFormWTA) calculateWTA(farmer);
			if (needsToFormWTP) calculateWTP(farmer);
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}
	
	/**
	 * Actualizacion de expectativas:
	 * 
	 * 1) Si es un landlord, entonces ver si regresa al mercado. Si lo hace y le sobra plata, entonces calcula el WTP (para alquilar mas campos). Si no regresa, calcula WTA
	 * 2) Si es un owner only entonces:
	 * 		2a. Si no puede seguir operando su campo, entonces pasa a LiveOff y debe formar WTA
	 * 		2b. Si puede seguir operando su campo 
	 * 				Si no esta conforme con su progreso, debe formar WTA
	 * 				Si esta conforme con su progreso y le sobre dinero, debe formar WTP (potencial arrendatario) 
	 *  
	 */
	public void executeStage2(Farmer farmer) {
		try {
			boolean needsToFormWTA = false;
			boolean needsToFormWTP = false;		
			
			Iterator<Region> regions = farmer.getCandidateRegions().iterator();
			Manager manager = Manager.getInstance();
			CroppingCycle croppingCycle = manager.getCurrentCroppingCycle();
			Double ownFarmsExpectedOperationCost = 0.0;
			for (Farm farm: farmer.getOwnedFarms()) {
				ownFarmsExpectedOperationCost += farm.getExpectedOperationCost(farmer);
			}
			Double workingCapital = farmer.getWorkingCapital();
					
			if (farmer.getOperationalStatus().isLiveOff()) {
				if (farmer.getOperationalStatus().equals(OperationalStatus.RetiredOwner)) {
					// Farmers who are passive land lords (retired owners) do not return to farming activity
					farmer.addLiveOffCroppingCycle();
					needsToFormWTA = true;
				} else {
					// Farmers who are active land lords have the chance to come back if they have enough capital
					if (workingCapital >= ownFarmsExpectedOperationCost) {
						// The farmer has enough capital to operate his farm again
						Function probabilityFunction = null;
						if (manager.getScenario().hasAttribute(Scenario.Attribute.VariableReenteringProbability)) {
							probabilityFunction = new SigmoidFunction(-0.03, -887.3, 50.0, 0.9);
						} else {
							probabilityFunction = new ConstantFunction(0.25);
						}
						
						Randomizer randomizer = manager.getScenario().getRandomizer();
						Double p = probabilityFunction.apply((double) farmer.getLiveOffCroppingCycles());
						Double r = randomizer.getRandomDouble();
						if (r <= p) {
							// Resurrection
							for (Farm farm: farmer.getOwnedFarms()) {
								Farmer operator = farm.getOperator();
								if (operator != null) {
									// Puede ser que el tenant haya quebrado. Pero si tiene operador, lo echo						
									operator.removeOperatedFarm(farm, true); // Remove farm from operator's list
								}					
								farmer.addOperatedFarm(farm); // Add farm to owner's list
								farm.setDefaultAspirationLevel(); // Set default aspiration level for farm
							}
							
							// Calcula el WTP solo si puede implantar la SMA segun el tamano de sus propios campos
							// en alguna de las regiones.
							Double minimumExpectedOperationCost = null;					
							while (regions.hasNext()) {						
								Region region = regions.next();
								Double minimumLeasableArea = farmer.getMinimumLeasableArea(region);
								Double expectedOperationCostPerHA = region.getExpectedOperationCost(croppingCycle) * farmer.getManagerialAbility(); // Utilizo el managerial ability del farmer y un factor de escala de 1 (ya que asumo que se queda sin nada)
								Double expectedOperationCost = expectedOperationCostPerHA * minimumLeasableArea;
								if ((minimumExpectedOperationCost == null) || (expectedOperationCost < minimumExpectedOperationCost)) {
									minimumExpectedOperationCost = expectedOperationCost;
								}
							}
							
							if ((workingCapital - ownFarmsExpectedOperationCost) >= minimumExpectedOperationCost) {
								needsToFormWTP = true;
							}
						} else {
							// The "coin" decided that the farmer will continue as a landlord one more year
							farmer.addLiveOffCroppingCycle();
							needsToFormWTA = true;
						}
					} else {
						// Liquid capital is not enough. The farmer will continue as a landlord one more year
						farmer.addLiveOffCroppingCycle();
						needsToFormWTA = true;
					}
				}
			} else if (farmer.getOperationalStatus().equals(OperationalStatus.OwnerOnly)){
				if (workingCapital >= ownFarmsExpectedOperationCost) {
					// Puede seguir alquilando su campo. 
					// Primero me fijo si esta conforme con su progreso
					boolean isProgressAcceptable = this.progressEvaluationMechanism.evaluate(farmer);
					if (! isProgressAcceptable) {
						// No esta conforme, entonces formo WTA
						needsToFormWTA = true;				
					} else {
						// Esta conforme con su progreso. Entonces evalua si "va por mas"
						// Calcula el WTP solo si puede implantar la SMA segun el tamano de sus propios campos
						// en alguna de las regiones.
						Double minimumExpectedOperationCost = null;					
						while (regions.hasNext()) {						
							Region region = regions.next();
							Double minimumLeasableArea = farmer.getMinimumLeasableArea(region);
							Double expectedOperationCostPerHA = region.getExpectedOperationCost(croppingCycle) * farmer.getManagerialAbility(); // Utilizo el managerial ability del farmer y un factor de escala de 1 (ya que asumo que se queda sin nada)
							Double expectedOperationCost = expectedOperationCostPerHA * minimumLeasableArea;
							if ((minimumExpectedOperationCost == null) || (expectedOperationCost < minimumExpectedOperationCost)) {
								minimumExpectedOperationCost = expectedOperationCost;
							}
						}
						
						if ((workingCapital - ownFarmsExpectedOperationCost) >= minimumExpectedOperationCost) {
							needsToFormWTP = true;
						}					
					}				
				} else {
					// Debe dejar de operar su campo y pasa a ser un land lord. Debe calcular WTA
					for (Farm farm: farmer.getOwnedFarms()) {					
						farmer.removeOperatedFarm(farm, false);
						farm.setDefaultAspirationLevel();
					}
					
					needsToFormWTA = true;
				}
			}
	
			if (needsToFormWTA) calculateWTA(farmer);
			if (needsToFormWTP) calculateWTP(farmer);
		} catch (Exception e) {
			throw new FulminantError("Unexpected error while executing scheduled method: " + e);
		}
	}
	
	public void calculateWTA(Farmer farmer) {
		// El calculo del WTA es una estrategia propia de cada farmer
		WTACalculationMechanism wtaCalculationMechanism = farmer.getWTACalculationMechanism();
		Map<Farm,Double> wtas = wtaCalculationMechanism.calculateWTA(farmer);
		farmer.setWillingnessToAccept(wtas);
	}
	
	public void calculateWTP(Farmer farmer) {
		/* 
		 *  Para calcular el WTP, necesito primero hacerme una idea de lo que pretendo ganar. De acuerdo a lo
		 *  documentado se que:
		 *  
		 *  NM = GGM - IC
		 *  NM = ALEC - IC
		 *  NM = ALEC - (LR + OIC)
		 *  
		 *  De esta ecuacion, se desprende.
		 *  
		 *  LR = ALEC - NM - OIC, [1]
		 *  
		 *  LR = Land Rental price
		 *  ALEC = Aspitarion Level modified by Expected Context
		 *  NM = Net Margin
		 *  OIC = Other Indirect Costs (sin considerar el Land Rental Price) 
		 *  
		 *  El valor de NM se puede estimar como
		 *  
		 *  NM = T * (FDC + LR), [2]
		 *  
		 *  T = Desired profitability rate (por ahora es un valor fijo e igual al 8%)
		 *  FDC = Fixed direct cost (del AM mas caro)
		 *  
		 *  De las ecuaciones [1] y [2] despejo LR
		 *  
		 *  LR = ALEC - T * (FDC+LR) - OIC;
		 *  LR + T*LR = ALEC - T*FDC - OIC;
		 *  LR = (ALEC - T*FDC - OIC) / (1+T)
		 *  
		 *  Finalmente, WTP = LR
		 */
		Manager manager = Manager.getInstance();
		Scenario scenario = manager.getScenario();
		Double desiredProfitabilityRate = (Double) scenario.getParameter("WTP.DesiredProfitabilityRate");
		
		Map<SoilType,Double> wtp = new HashMap<SoilType,Double>();
		List<Farm> operatedFarms = farmer.getOperatedFarms();
		for (SoilType soilType: this.soilTypes) {
			Double soilTypeWTP = null;
			
			// 1. Busco todos los campos operados con ese tipo de suelo (asumo que todos los lotes de un campo tienen el mismo tipo de suelo)
			List<Farm> farms = new ArrayList<Farm>();
			for (Farm farm: operatedFarms) {
				if (farm.getSoilType().equals(soilType)) {
					farms.add(farm);
				}
			}
			
			if (! farms.isEmpty()) {
				// 2. Hay al menos un campo con ese tipo de suelo. De manera que puedo calcular un 
				//    precio de alquiler estimativo para cada una de ellos
				List<Double> farmsWTP = new ArrayList<Double>();
				for (Farm farm: farms) {
					// 2a. Obtencion del ALEC
					Double alec = null;
					if (farm.getOwner().equals(farmer)) {
						// Si el campo es propio, no considerar el nivel de aspiracion del mismo sino el promedio
						// de los niveles de aspiracion de los pares cuyo regimen de rotacion (segun el regimen de tenencia)
						// sea el mismo que el regimen de rotacion del farmer como Tenant
						Integer nPeers = 0;
						Double sumAlecs = 0.0;
						Network<Agent> network = manager.getNeighbourhood();
						Iterator<Agent> adjacents = network.getAdjacent(farm).iterator();
						while (adjacents.hasNext()) {
							Farm peer = (Farm) adjacents.next();
							Farmer peerOperator = peer.getOperator();
							if (peerOperator != null) {
								RotatorType peerRotatorType = peerOperator.equals(peer.getOwner()) ? peerOperator.getOwnerRotatorType() : peerOperator.getTenantRotatorType();
								if (farmer.getTenantRotatorType().equals(peerRotatorType)) {
									sumAlecs += peer.getAspirationLevel().getLevel();
									nPeers++;
								}
							}
						}
						
						if (nPeers > 0) {
							alec = sumAlecs / nPeers.doubleValue();
						}
					} else {
						alec = farm.getAspirationLevel().getLevel();	
					}
						
					if (alec != null) {
						// 2b. Obtencion del FDC (para el AM mas caro del cropping cycle)
						Double fdc = farm.getExpectedOperationCost(farm.getOwner()); // Ya viene en u$s/ha
			
						// 2c. Calculo de otros costos indirectos (sin uso de economias de escala para ser coherente con lo anterior)
						//     Como el WTP va a servir para "alquilar", entonces hago de cuenta que soy siempre un arrendatario, aunque el campo sea mio 
						//	   Entonces el unico costo aplicable es el Management cost
						Double oic = IndirectCost.ManagementCost.getCost(scenario, farmer, farm, false) / farm.getFarmSize();
						
						// 2d. Calculo de WTP para ese campo
						// LR = (ALCC - T*FDC - OIC) / (1+T)
						Double farmWTP = (alec - desiredProfitabilityRate * fdc - oic) / (1.0 + desiredProfitabilityRate);
						farmsWTP.add(farmWTP);
					}
				}
				
				// De todos los WTP, obtengo el de valor mas bajo
				if (! farmsWTP.isEmpty()) {
					soilTypeWTP = farmsWTP.get(0);
					for (Double fWTP: farmsWTP) {
						if (fWTP < soilTypeWTP) {
							soilTypeWTP = fWTP;
						}
					}
				}
			}
			
			wtp.put(soilType, soilTypeWTP);
		}
		
		farmer.setWillingnesstoPay(wtp);
	}
}