package fiuba.lamm.agromodel.areaupdate.mcp.pe;

import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.math.custom.*;
import java.util.*;
import repast.simphony.space.graph.*;

public class NOutOfMEvaluationMechanism extends ProgressEvaluationMechanism {
	protected Integer croppingCyclesForProgressEvaluation;
	protected Integer minimumPositiveCycles; /* Cantidad minima de ciclos donde tiene que alcanzar el objetivo (ej: 3 de 5) */

	
	public NOutOfMEvaluationMechanism(Integer croppingCyclesForProgressEvaluation, Integer minimumPositiveCycles) {
		this.croppingCyclesForProgressEvaluation = croppingCyclesForProgressEvaluation;
		this.minimumPositiveCycles = minimumPositiveCycles;
	}
	
	public Double evaluateCurrentProgress(Farmer farmer) {
		List<Double> totalNetFarmingIncomes = farmer.getTotalNetFarmingIncomes();
		if (totalNetFarmingIncomes.isEmpty()) {
			throw new FulminantError("La lista de total net farming incomes esta vacia. Cropping cycle = " + Manager.getInstance().getCurrentCroppingCycle() + "; farmerId = " + farmer.getFarmerId() + "; operational status = " + farmer.getOperationalStatus());
		}
		
		// Tiene que haber al menos 2 ciclos para poder comparar. Si hay un solo ciclo tomo como valor de referencia el 0.
		// Con lo cual el primer progress rate seguramente va a dar +INFINITO o -INFINITO
		Double previousTotalNetFarmingIncome = (totalNetFarmingIncomes.size() >= 2) ? totalNetFarmingIncomes.get(totalNetFarmingIncomes.size() - 2) : 0.0;
		Double actualTotalNetFarmingIncome = totalNetFarmingIncomes.get(totalNetFarmingIncomes.size() - 1);
		CapitalProgressRateFunction progressRateFunction = new CapitalProgressRateFunction();
		Double progressRate = progressRateFunction.apply(new double[] { previousTotalNetFarmingIncome, actualTotalNetFarmingIncome });		
		return progressRate;		
	}
	
	public Double calculateMinimumProgressRate(Farmer farmer) {
		if (farmer.getOwnMinimumProgressRate() != null) {
			// 1. Calculo el progress rate de los pares
			Network<Agent> network = Manager.getInstance().getNeighbourhood();
			List<Farmer> peers = new ArrayList<Farmer>();
			for (Farm farm: farmer.getOperatedFarms()) {
				Iterator<Agent> adjacents = network.getAdjacent(farm).iterator();
				while (adjacents.hasNext()) {
					Farm adjacentFarm = (Farm) adjacents.next();
					Farmer operator = adjacentFarm.getOperator();
					if ((operator != null) && !peers.contains(operator) && !operator.equals(farmer)) {
						peers.add(operator);
					}
				}
			}
			
			// 2. Evaluo el progress rate a considerar. Me fijo solamente en los pares que tienen progreso finito
			if (! peers.isEmpty()) {
				Integer nPeers = 0;
				Double sum = 0.0;
				for (Farmer peer: peers) {
					List<Double> progressRates = peer.getProgressRates();
					Double lastProgressRate = progressRates.get(progressRates.size() - 1);
					if (! lastProgressRate.isInfinite()) {
						sum += lastProgressRate;
						nPeers++;
					}
				}
				
				if (nPeers > 0) {
					Double peersAverageProgressRate = (sum / (double) nPeers);
					Double farmerMinimumProgressRate = farmer.getOwnMinimumProgressRate();
						
					return Math.min(peersAverageProgressRate, farmerMinimumProgressRate);
				} else {
					// No hay ningun par que tenga progreso finito. Tomo el own minimum progress rate
					return farmer.getOwnMinimumProgressRate();
				}
			} else {
				// No hay pares a considerar. Utilizo el progress rate propio.
				return farmer.getOwnMinimumProgressRate();
			}
		} else {
			// El farmer no tiene minimum progress rate propio definido. Asi que no le interesa tampoco el
			// de los pares
			return null;
		}
	}
	
	public boolean evaluate(Farmer farmer) {				
		if (farmer.getOwnMinimumProgressRate() != null) {
			List<Double> progressRates = farmer.getProgressRates();			
			List<Double> minimumProgressRates = farmer.getMinimumProgressRates();
			
			if (progressRates.size() >= croppingCyclesForProgressEvaluation) {				
				// Busco la cantidad de ciclos aceptables
				int acceptableCycles = 0;
				for (int i = 1; i <= croppingCyclesForProgressEvaluation; i++) {
					Double croppingCycleProgressRate = progressRates.get(progressRates.size() - i);
					Double croppingCycleMinimumProgressRate = minimumProgressRates.get(minimumProgressRates.size() - i);
					if (croppingCycleProgressRate >= croppingCycleMinimumProgressRate) {
						acceptableCycles++;
					}
				}
				
				// Si la cantidad de ciclos aceptables es mayor o igual al minimo, devuelvo true, sino false.
				return (acceptableCycles >= minimumPositiveCycles);
			} else {
				// Hay menos ciclos de los necesarios. Se considera que el progreso es aceptable
				return true;
			}
		} else {
			// El farmer no tiene minimum progress rate definido, por lo tanto se considera que el progreso
			// siempre es aceptable.
			return true;
		}
	}
}