package fiuba.lamm.agromodel.areaupdate.mcp.wta;

import java.util.*;
import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.decision.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.finance.*;
import fiuba.lamm.agromodel.math.custom.*;

public class OptimizedWTACalculationMechanism extends WTACalculationMechanism {
	public Map<Farm,Double> calculateWTA(Farmer farmer) {
		Map<Farm,Double> wtas = new HashMap<Farm,Double>();
		List<Farm> farms = farmer.getOwnedFarms();
		for (Farm farm: farms) {
			wtas.put(farm, calculateWTA(farm));
		}
		
		return wtas;
	}
	
	public Double calculateWTA(Farm farm) {
		/* Obtencion de variables relacionados al farmer */
		Farmer farmer = farm.getOwner();
		RotatorType tenantRotatorType = farmer.getTenantRotatorType();
		List<Double> ownedFarmOnlyIndirectCosts = farmer.getOwnFarmIndirectCosts(farm);
				
		/* Para los AMPs, aplico las mismas restricciones que para en LandUseSelectionMechanism, exepto
		 * que el RotatorType considerado es el de Tenant y no el de Owner
		 */		
		LandUseSelectionMechanism landUseSelectionMechanism = farmer.getLandUseSelectionMechanism();
		FitnessFunction fitnessFunction = null;		
		if (landUseSelectionMechanism instanceof OptimizationLandUseSelectionMechanism) {
			OptimizationLandUseSelectionMechanism m = (OptimizationLandUseSelectionMechanism) landUseSelectionMechanism;
			fitnessFunction = m.getFitnessFunction();
		} else {
			throw new FulminantError("Se esta intentando calcular el WTA con teoria de la utilidad pero el metodo de seleccion de uso de la tierra no es mediante optimizacion");
		}
		Map<LandUseSelectionRestriction,Object> restrictions = new HashMap<LandUseSelectionRestriction,Object>();
		restrictions.put(LandUseSelectionRestriction.RegionRestriction, farm.getRegion());
		restrictions.put(LandUseSelectionRestriction.TechnologyRestriction, farm.getUnknownManagements());
		restrictions.put(LandUseSelectionRestriction.MinimumActivityGroupProportionRestriction, tenantRotatorType.getMinPlotsPerActivityGroup());
		restrictions.put(LandUseSelectionRestriction.MaximumActivityGroupProportionPRestriction, tenantRotatorType.getMaxPlotsPerActivityGroup());
		List<ActivityManagementProportion> amps = landUseSelectionMechanism.getAMPFilteredList(restrictions);
		
		// 1. Generacion de lista de margenes netos
		Manager manager = Manager.getInstance();
		Scenario scenario = manager.getScenario();
		Integer croppingCyclesForUtilityCalculation = (Integer) scenario.getParameter("WTA.CroppingCyclesForUtilityCalculation");
		Double bestAMPProportionForUtilityCalculation = (Double) scenario.getParameter("WTA.BestAMPProportionForUtilityCalculation");
		Integer maxCroppingCycleAmps = (int) Math.round(bestAMPProportionForUtilityCalculation * (double) amps.size());
		
		SoilType soilType = farm.getSoilType();		
		List<Double> topNetMargins = new ArrayList<Double>();				
		for (int i = 1; i <= croppingCyclesForUtilityCalculation; i++) {
			Double minNetMargin = Double.POSITIVE_INFINITY;
			Double outcomeFactor = farm.getOutcomeFactor(i);
			List<Double> croppingCycleNetMargins = new ArrayList<Double>();
			Double croppingCycleIndirectCosts = ownedFarmOnlyIndirectCosts.get(ownedFarmOnlyIndirectCosts.size() - 1);
			
			for (ActivityManagementProportion amp: amps) {
				Double ggmPerHectare = manager.getGlobalGrossMargin(amp, soilType, i - 1);
				Double ggm = ggmPerHectare * farm.getFarmSize();
				if (outcomeFactor != null) {
					ggm *= outcomeFactor;
				}
				Double netMargin = ggm - croppingCycleIndirectCosts;
				
				if ((croppingCycleNetMargins.size() < maxCroppingCycleAmps) || (netMargin > minNetMargin)) {
					// Agrego el net margin a la lista
					
					if (croppingCycleNetMargins.size() < maxCroppingCycleAmps) {
						// Todavia le falta, asi que lo agrego sin sacar nada
						croppingCycleNetMargins.add(netMargin);
						
						// Si el NetMargin es menor al minino, lo seteo como el nuevo minimo
						if (netMargin < minNetMargin) {
							minNetMargin = netMargin;
						}
					} else {
						// La lista esta llena pero el net margin actual es mayor que alguno de los ya insertados
						// Borro el menor (o uno de ellos si es que hay mas de uno)
						croppingCycleNetMargins.remove(minNetMargin);
						croppingCycleNetMargins.add(netMargin);
						
						// Ahora debo calcular el nuevo net margin minimo
						Collections.sort(croppingCycleNetMargins);
						minNetMargin = croppingCycleNetMargins.get(0);
					}
				}				
			}
			
			// Ya tengos los mejores AMPs de ese ciclo. Ahora los agrego a la lista general
			topNetMargins.addAll(croppingCycleNetMargins);						
		}
		
		// 2. Seteo del valor de referencia
		Double refValue = null;
		Double currentIndirectCost = IndirectCost.getTotalIndirectCostsForOwnFarm(manager.getScenario(), farmer, farm);
		Double currentManagementCost = IndirectCost.ManagementCost.getCost(manager.getScenario(), farmer, farm, false);		
		if (fitnessFunction.getLandUseSelectionMechanismType().equals(LandUseSelectionMechanism.Type.EUOptimizationSelectionMechanism)) {
			// Seteo de parametro w0 (utilizando criterio de minimo capital liquido)
			Double farmerWorkingCapital = farmer.getWorkingCapital();
			Double expectedOperationCost = farm.getExpectedOperationCost(farmer) * farm.getFarmSize();
			Double workingCapital = Math.max(farmerWorkingCapital, expectedOperationCost);
			Double maxMortgageLoan = 0.4 * soilType.getLandSalePrice() * farm.getFarmSize();
			refValue = workingCapital + maxMortgageLoan;		
		} else if (fitnessFunction.getLandUseSelectionMechanismType().equals(LandUseSelectionMechanism.Type.CPTOptimizationSelectionMechanism)) {	
			/*
			 *  El valor de referencia es el AL menos los costos indirectos.
			 *  
			 *  En el caso de los land lords, el nivel de aspiracion que se toma del campo es el default o el del actual arrendatario
			 *  
			 */			
			refValue = farm.getAspirationLevel().getLevel() * farm.getFarmSize() - currentIndirectCost;
		}
		fitnessFunction.setCurrentReferenceValue(refValue);
		
		// 3. Generacion de variables para el calculo de la EU y obtencion de dicho valor
		//    Para el calculo de los costos indirectos, debemos asumir que el operador es el owner	
		//    No es posible que un land lord haya echado al tenant y llegue a la instancia de calcular WTA		
		double x[] = new double[topNetMargins.size()];
		for (int i = 0; i < topNetMargins.size(); i++) {
			x[i] = topNetMargins.get(i);			
		}
		Double expectedUtility = fitnessFunction.apply(x);		

		// 4. Obtencion del CE
		Double ce = fitnessFunction.getCertaintyEquivalent(expectedUtility);
		Double wta = (ce + (currentIndirectCost - currentManagementCost)) / farm.getFarmSize();
		if (wta.isNaN()) {
			throw new FulminantError("WTA is not a number. Farmer = " + farmer.getFarmerId() + "; Cropping cycle = " + manager.getCurrentCroppingCycle() + "; Certainty equivalent = " + ce + "; Expected utility = " + expectedUtility + "; Top Net Margins = " + topNetMargins.size() + "; Ref value = " + refValue);
		}
		
		return wta;
	}
}