package fiuba.lamm.agromodel.decision;

import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.utils.*;
import java.util.*;

public class ActivityManagementProportion implements Comparable<ActivityManagementProportion> {
	protected Map<ActivityManagement,Integer> proportion;
	
	/**
	 * Generate a list of activity managements with every possible proportion
	 * 
	 * @param managements		List of all possible managements
	 * @param numberOfPlots		Number of plots to distribute the activity managements
	 * @return
	 */
	public static List<ActivityManagementProportion> getPossibleProportions(List<ActivityManagement> managements, int numberOfPlots) {
		List<ActivityManagementProportion> proportions = new ArrayList<ActivityManagementProportion>();
		
		// 1. Descomponer el numero de parcelas en una particion de sumas que tenga como maximo
		//    managements.size() cantidad de terminos
		List<NonUniqueSet<Integer>> partitions = new ArrayList<NonUniqueSet<Integer>>();
		int maxTerms = Math.min(numberOfPlots, managements.size());
		for (int i = 1; i <= maxTerms; i++) {
			LinkedList<Integer> movingList = new LinkedList<Integer>();
			int reminder = numberOfPlots;
			for (int j = 1; j < i; j++) {
				movingList.add(1);
				reminder--;
			}
			movingList.add(reminder);
			
			// Esta lista es la primera particion posible
			NonUniqueSet<Integer> set = new NonUniqueSet<Integer>(movingList);
			partitions.add(set);
			
			int min = movingList.getFirst();
			int max = movingList.getLast();
			int currentIndex = movingList.size() - 1;
			while ((movingList.size() > 1) && (min <= max)) {
				// Paso una unidad de un elemento a otro
				int nextIndex = currentIndex - 1;
				
				// Paso una unidad de currentIndex a nextIndex
				movingList.set(nextIndex, movingList.get(nextIndex) + 1);
				movingList.set(currentIndex, movingList.get(currentIndex) - 1);
				NonUniqueSet<Integer> candidate = new NonUniqueSet<Integer>(movingList);
				
				// Si nextIndex == 0, entonces llegue al principio de la lista.
				// Actualizo el valor min y muevo el indice al final de la lista
				if (nextIndex == 0) {
					min = movingList.getFirst();
					max = movingList.getLast();
					nextIndex = movingList.size() - 1;
				}
			
				boolean contains = partitions.contains(candidate);
				if (!contains) partitions.add(candidate);
				currentIndex = nextIndex;
			}
		}
		
		// 2. Ya tengo las posibles particiones. Ahora lo que tengo que hacer es generar todoas
		//    las posibles combinaciones segun las particiones generadas		
		for (int i = 0; i < partitions.size(); i++) {
			List<Integer> partition = new ArrayList<Integer>(partitions.get(i).collection());
			List<ActivityManagementProportion> combinations = new ArrayList<ActivityManagementProportion>();
			
			// Genero todas las posibles combinaciones para el primer valor de proporcion
			Integer p = partition.get(0);
			for (int k = 0; k < managements.size(); k++) {
				ActivityManagement am = managements.get(k);
				ActivityManagementProportion amp = new ActivityManagementProportion();
				amp.setProportion(am, p);
				combinations.add(amp);
			}
			
			// Ahora cruzo genero todas las demas combinaciones con las proporciones restantes
			for (int j = 1; j < partition.size(); j++) {
				p = partition.get(j);
				List<ActivityManagementProportion> newCombinations = new ArrayList<ActivityManagementProportion>();
				
				for (int k = 0; k < managements.size(); k++) {
					ActivityManagement am = managements.get(k);
					
					// Agrego la proporcion (am, p) en todos los casos que sea posible
					for (int l = 0; l < combinations.size(); l++) {
						ActivityManagementProportion amp = combinations.get(l);
						if (amp.getProportion(am) == 0) {
							// Puedo agregarlo porque ese a/m no habia sido considerado
							ActivityManagementProportion amp2 = new ActivityManagementProportion(amp);
							amp2.setProportion(am, p);
							
							// Me fijo si puedo insertarlo o esta duplicado
							if (! newCombinations.contains(amp2)) newCombinations.add(amp2);
						}
					}
				}
				
				// Guardo en combinations las nuevas proporciones
				combinations = newCombinations;
			}
			
			proportions.addAll(combinations);
		}
			
		return proportions;
	}
	
	public ActivityManagementProportion() {
		this.proportion = new HashMap<ActivityManagement, Integer>();
	}
	
	public ActivityManagementProportion(List<Plot> plots) {
		this.proportion = new HashMap<ActivityManagement, Integer>();
		for (Plot plot: plots) {
			ActivityManagement management = plot.getCurrentActivityManagement();
			if (! this.proportion.containsKey(management)) {
				this.proportion.put(management, 1);
			} else {
				this.proportion.put(management, this.proportion.get(management) + 1);
			}
		}
	}
		
	public ActivityManagementProportion(ActivityManagementProportion amp) {
		this.proportion = new HashMap<ActivityManagement,Integer>(amp.proportion);
	}
	
	/**
	 * Checks if AMP contains a specific AM having a proportion greater than zero.
	 */
	public boolean contains(ActivityManagement am) {
		Integer proportion = this.getProportion(am);
		return ((proportion != null) && (proportion > 0));
	}
	
	/**
	 * Checks if AMP contains any of the AMs within the list having a proportion greater than zero.
	 */
	public boolean containsAny(List<ActivityManagement> ams) {
		boolean containsAny = false;
		int i = 0;
		while ((i < ams.size()) && ! containsAny) {
			ActivityManagement am = ams.get(i);
			containsAny = this.contains(am);
			i++;
		}
		
		return containsAny;		
	}
	
	/**
	 * Returns all the distinct ActivityManagements 
	 * @return
	 */
	public List<ActivityManagement> getDistinctActivityManagements() {
		return new ArrayList<ActivityManagement>(this.proportion.keySet());
	}
	
	/**
	 * Returns all the ActivityManagements. Each ActivityManagement appears
	 * as many times as its proportion 
	 * @return
	 */
	public List<ActivityManagement> getAllActivityManagements() {
		List<ActivityManagement> distinct = this.getDistinctActivityManagements();
		List<ActivityManagement> all = new ArrayList<ActivityManagement>();
		
		for (ActivityManagement management: distinct) {
			Integer p = this.proportion.get(management);
			for (int i = 0; i < p; i++) all.add(management);
		}
		
		return all;
	}
	
	public List<Activity> getActivities() {
		List<Activity> activitys = new ArrayList<Activity>();
		List<ActivityManagement> managements = getDistinctActivityManagements();
		for (int i = 0; i < managements.size(); i++) {
			ActivityManagement management = managements.get(i);
			Activity activity = management.getActivity();
			if (! activitys.contains(activity)) activitys.add(activity);
		}
		
		return activitys;
	}
	
	public Integer getProportion(ActivityManagement management) {
		if (this.proportion.containsKey(management)) {
			return this.proportion.get(management);
		}
		
		return 0;
	}	
	
	public Integer getProportion(Activity activity) {
		int p = 0;
		Iterator<ActivityManagement> managements = this.proportion.keySet().iterator();
		while (managements.hasNext()) {
			ActivityManagement management = managements.next();
			if (activity.equals(management.getActivity())) p += this.proportion.get(management);
		}
		
		return p;
	}
	
	public void setProportion(ActivityManagement management, Integer newProportion) {
		this.proportion.put(management, newProportion);
	}
	
	/**
	 * Picks at random any ActivityManagement whose Activity is equals the parameter activity. The first 
	 * ActivityManagement found is picked and its proportion is decreased by 1. If not, null is returned.
	 * 
	 * @param activity	The activity which is part of the ActivityManagement to be picked
	 * @return				The activity/management that stisfies the condition above
	 */
	public ActivityManagement pickActivityManagement(Activity activity) {
		List<ActivityManagement> candidates = new ArrayList<ActivityManagement>();
		Iterator<ActivityManagement> managements = this.proportion.keySet().iterator();
		while (managements.hasNext()) {
			ActivityManagement am = managements.next();
			if ((activity == null) || am.getActivity().equals(activity)) {
				Integer p = getProportion(am);
				if (p > 0) candidates.add(am);			
			}
		}
	
		ActivityManagement management = null;		
		Randomizer randomizer = Manager.getInstance().getScenario().getRandomizer();
		if (! candidates.isEmpty()) {
			int index = randomizer.getRandomInt(candidates.size());
			
			// Pick management and decrease proportion
			management = candidates.get(index);
			setProportion(management, getProportion(management) - 1);	
		}
		
		return management;
	}
	
	/**
	 * Picks at random any ActivityManagement. Its proportion is decreased by 1.
	 * If no activties with proportion greatear than 0 are found, it returns null.
	 */	
	public ActivityManagement pickActivityManagement() {
		return pickActivityManagement(null);
	}
	
	public Region getAssociatedRegion() {
		List<ActivityManagement> ams = this.getDistinctActivityManagements();
		return ams.get(0).getRegion();
	}
	
	public boolean equals(Object object) {
		if (object instanceof ActivityManagementProportion) {
			ActivityManagementProportion anotherAmp = (ActivityManagementProportion) object;
			
			boolean equals = true;
			List<ActivityManagement> managements = anotherAmp.getDistinctActivityManagements();
			int i = 0;
			while ((i < managements.size()) && equals) {
				ActivityManagement management = managements.get(i);
				Integer thisProportion = getProportion(management);
				Integer anotherProportion = anotherAmp.getProportion(management);
				equals = anotherProportion.equals(thisProportion);
				
				i++;
			}
			
			return equals; 
		}
		
		return false;
	}
	
	public int compareTo(ActivityManagementProportion amp) {
		int compareValue = 0;
		List<ActivityManagement> thisManagements = this.getAllActivityManagements();
		List<ActivityManagement> ampManagements = amp.getAllActivityManagements();
		
		Collections.sort(thisManagements);
		Collections.sort(ampManagements);
		
		int i = 0;
		while ((i < thisManagements.size()) && (compareValue == 0)) {
			ActivityManagement am1 = thisManagements.get(i);
			ActivityManagement am2 = ampManagements.get(i);
			compareValue = am1.compareTo(am2);
			
			i++;
		}
		
		return compareValue;
	}
	
	public int hashCode() {
		return this.proportion.hashCode();
	}
	
	public String toString() {
		return this.proportion.toString();
	}
}