package fiuba.lamm.agromodel.decision;

import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.decision.factories.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.factories.*;
import java.sql.*;
import java.util.*;
import repast.simphony.space.graph.*;

public class AspirationLevel {
	protected List<Double> globalGrossMargins;
	protected List<Double> levelsAdjustedByExpectedContexts;
	protected List<Double> levelsAdjustedByContextComparison;
	protected List<ActivityManagementProportion> activityManagementProportions;
	protected double level;
	
	public static synchronized AspirationLevel getDefaultAspirationLevel(SoilType soilType) {
		Manager manager = Manager.getInstance();
		Map<SoilType,Double> defaultLevels = manager.getDefaultAspirationLevels();
		if (defaultLevels == null) {
			defaultLevels = AspirationLevel.getDefaultAspirationLevels();
			manager.setDefaultAspirationLevels(defaultLevels);
		}
		
		return new AspirationLevel(defaultLevels.get(soilType));
	}
	
	public static Map<SoilType,Double> getDefaultAspirationLevels() {
		Map<SoilType,Double> defaultLevels = new HashMap<SoilType,Double>();
		Manager manager = Manager.getInstance();
		Scenario scenario = manager.getScenario();
				
		try {
			List<SoilType> soilTypes = SoilTypeFactory.getInstance(scenario).findAll();
			CroppingCycle croppingCycle = manager.getCurrentCroppingCycle();
			for (SoilType soilType: soilTypes) {
				double factor = 0.0;
				
				List<ContextType> contextTypes = ContextTypeFactory.getInstance(scenario).findAll();
				for (ContextType contextType: contextTypes) {
					ContextTypeStatus expectedContextStatus = croppingCycle.getExpectedContextTypeStatus(soilType, contextType);
					factor += contextType.getAdjustmentFactor(soilType, expectedContextStatus);
				}		
				Double expectedGrossMargin = soilType.getExpectedGrossMargin();
				if (expectedGrossMargin != null) {
					Double expectedLevel = expectedGrossMargin * (1.0 + factor);
					defaultLevels.put(soilType, expectedLevel);
				}			
			}
		} catch (SQLException sqle) {
			throw new FulminantError("Error al buscar los factories para actualizar los niveles de aspiracion default: " + sqle);
		}
		
		return defaultLevels;
	}
	
	public AspirationLevel(double level) {
		this.level = level;
		this.globalGrossMargins = new ArrayList<Double>();
		this.levelsAdjustedByExpectedContexts = new ArrayList<Double>();
		this.levelsAdjustedByContextComparison = new ArrayList<Double>();
		this.activityManagementProportions = new ArrayList<ActivityManagementProportion>();
	}
		
	/**
	 * Updates aspiration level before the beginning of the campaign according to the expected climate context
	 */
	public void updateBeforeCroppingCycle(Farm farm) {
		SoilType soilType = farm.getSoilType();
		Manager manager = Manager.getInstance();
		CroppingCycle croppingCycle = manager.getCurrentCroppingCycle();
		double factor = 0.0;

		try {
			List<ContextType> contextTypes = ContextTypeFactory.getInstance(manager.getScenario()).findAll();
			for (ContextType contextType: contextTypes) {
				ContextTypeStatus expectedContextStatus = croppingCycle.getExpectedContextTypeStatus(soilType, contextType);
				factor += contextType.getAdjustmentFactor(soilType, expectedContextStatus);
			}			
		} catch (SQLException sqle) {
			throw new FulminantError("Error al buscar los factories para actualizar los niveles de aspiracion default al inicio del cropping cycle: " + sqle);
		}		
		this.level *= (1.0 + factor);		
		
		// Store aspiration level adjusted by context comparison
		this.levelsAdjustedByExpectedContexts.add(new Double(this.level));
	}
	
	/**
	 * Updates the aspiration level taking into comparing the following items. This is done after the campaign
	 * has ended and the real contexts and economic results are known.
	 * 
	 * - Comparison between expected contexts and real contexts (for now only climate context)
	 * - Comparison of achieved outcome among peers
	 * - Optimism
	 */
	public void updateAfterCroppingCycle(Farm farm) {
		SoilType soilType = farm.getSoilType(); 
		Manager manager = Manager.getInstance();
		CroppingCycle croppingCycle = manager.getCurrentCroppingCycle();
		Network<Agent> network = manager.getNeighbourhood();
		Farmer operator = farm.getOperator();
		
		// 0. Store proportions of operated activity managements;
		List<Plot> plots = farm.getPlots();
		this.activityManagementProportions.add(new ActivityManagementProportion(plots));
		
		// 1. Comparison between expected context and real context
		double contextAdjustmentFactor = 0.0;
		try {
			List<ContextType> contextTypes = ContextTypeFactory.getInstance(manager.getScenario()).findAll();
			for (ContextType contextType: contextTypes) {
				ContextTypeStatus expectedContextStatus = croppingCycle.getExpectedContextTypeStatus(soilType, contextType);
				ContextTypeStatus actualContextStatus = croppingCycle.getActualContextTypeStatus(soilType, contextType);
				contextAdjustmentFactor += contextType.getStepAdjustmentFactor(soilType, expectedContextStatus, actualContextStatus);
			}
		} catch (SQLException sqle) {
			throw new FulminantError("Error al buscar los factories para actualizar los niveles de aspiracion luego del cropping cycle: " + sqle);
		}
		
		// 1e. Ahora genero el nuevo nivel de aspiracion
		this.level *= (1.0 + contextAdjustmentFactor);
		
		// 1f. Store aspiration level adjusted by context comparison
		this.levelsAdjustedByContextComparison.add(new Double(this.level));
		
		// 2. Outcome comparison among peers
		Double myGlobalGrossMarginPerHa = farm.getLastGlobalGrossMargin() / farm.getFarmSize();
		Iterator<Agent> adjacents = network.getAdjacent(farm).iterator();
		double adjustedGlobalGrossMarginPerHa = myGlobalGrossMarginPerHa;
		double peersGlobalGrossMarginPerHa = 0.0;
		int numberOfPeers = 0;		
		while (adjacents.hasNext()) {
			Farm peer = (Farm) adjacents.next();
			double ggm = peer.getLastGlobalGrossMargin();
			
			peersGlobalGrossMarginPerHa += ggm / peer.getFarmSize();
			numberOfPeers++;
		}
		double meanPeersGlobalGrossMarginPerHa = peersGlobalGrossMarginPerHa / (double) numberOfPeers;
				
		if (myGlobalGrossMarginPerHa < meanPeersGlobalGrossMarginPerHa) {
			double comparisonAdjustmentFactor = operator.getPeersAdjFactor();
			adjustedGlobalGrossMarginPerHa = myGlobalGrossMarginPerHa * (1.0 - comparisonAdjustmentFactor) + meanPeersGlobalGrossMarginPerHa * comparisonAdjustmentFactor;
		}
		
		// 2b. Store agent's global gross margin (without adjustment)
		this.globalGrossMargins.add(new Double(myGlobalGrossMarginPerHa));
		
		double lambda = 0.0;
		if (adjustedGlobalGrossMarginPerHa > this.level) {
			lambda = operator.getLambdaUp();
		} else if (adjustedGlobalGrossMarginPerHa < this.level) {	
			lambda = operator.getLambdaDown();
		}
		this.level = this.level * lambda + adjustedGlobalGrossMarginPerHa * (1.0 - lambda);
		
		// 3. Adjustment by optimism
		this.level *= (1.0 + operator.getOptimism());
	}
	
	// Getters
	public Double getLevel() {
		return this.level;
	}
	
	public List<Double> getGlobalGrossMargins() {
		return this.globalGrossMargins;
	}
	
	public List<Double> getAspirationLevelsAdjustedByExpectedContexts() {
		return this.levelsAdjustedByExpectedContexts;
	}
	
	public List<Double> getAspirationLevelsAdjustedByContextComparison() {
		return this.levelsAdjustedByContextComparison;
	}
	
	public List<ActivityManagementProportion> getActivityManagementProportions() {
		return this.activityManagementProportions;
	}
}