package fiuba.lamm.agromodel.decision;

import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.factories.*;
import fiuba.lamm.agromodel.math.*;
import fiuba.lamm.agromodel.math.custom.*;
import java.util.*;

public abstract class LandUseSelectionMechanism {
	protected Type landUseSelectionMechanismId;
	protected LandUseSelectionMechanism finalLandUseSelectionMechanism;
	
	public enum Type { 
		HabitSelectionMechanism	(0), 
		RandomSelectionMechanism	(1),
		MarginWeightedImitationSelectionMechanism	(2),
		EUOptimizationSelectionMechanism	(3),		
		CPTOptimizationSelectionMechanism	(4);
		
		private int id;
		Type(int id) {
			this.id = id;
		}
		
		public static Type byId(int id) {
			Type type = null;
			int i = 0;
			Type types[] = Type.values();
			
			while ((i < types.length) && (type == null)) {
				if (types[i].id == id) type = types[i];
				i++;
			}
			
			return type;
		}
	};
	
	public enum Parameter {
		RiskAversion,
		Alpha,
		Lambda,
		GammaGains,
		GammaLosses;		
	};
	
	/**
	 * @param searchTriggeringMechanismId
	 * @return	The search triggering mechanism specified by the id
	 */
	public static LandUseSelectionMechanism getInstance(int landUseSelectionMechanismId, Map<LandUseSelectionMechanism.Parameter, Object> mechanismParameters) {
		Type type = Type.byId(landUseSelectionMechanismId);
		LandUseSelectionMechanism mechanism = null;
		
		if (type.equals(Type.HabitSelectionMechanism)) {
			mechanism = new HabitLandUseSelectionMechanism();
		} else if (type.equals(Type.RandomSelectionMechanism)) {
			mechanism = new RandomLandUseSelectionMechanism();
		} else if (type.equals(Type.MarginWeightedImitationSelectionMechanism)) {
			mechanism = new MarginWeightedImitationSelectionMechanism();
		} else if (type.equals(Type.EUOptimizationSelectionMechanism)) {
			Double riskAversion = (Double) mechanismParameters.get(Parameter.RiskAversion);
			EUValueFunction valueFunction = new EUValueFunction(riskAversion);
			FitnessFunction fitnessFunction = new EUFitnessFunction(valueFunction, new ConstantVectorialField(1.0));			
			mechanism = new OptimizationLandUseSelectionMechanism(fitnessFunction);
		} else if (type.equals(Type.CPTOptimizationSelectionMechanism)) {
			Double alpha = (Double) mechanismParameters.get(Parameter.Alpha);
			Double lambda = (Double) mechanismParameters.get(Parameter.Lambda);
			Double gammaGains = (Double) mechanismParameters.get(Parameter.GammaGains);
			Double gammaLosses = (Double) mechanismParameters.get(Parameter.GammaLosses);
			CPTValueFunction valueFunction = new CPTValueFunction(lambda, alpha);
			FitnessFunction fitnessFunction = new CPTFitnessFunction(valueFunction, new ConstantVectorialField(1.0), gammaGains, gammaLosses);
			mechanism = new OptimizationLandUseSelectionMechanism(fitnessFunction);
		} else {
			throw new FulminantError("Unknown land use selection mechanism: " + landUseSelectionMechanismId);
		}
		
		return mechanism;
	}
	
	/**
	 * Generates a list of all the available proportions that match the selection criteria imposed 
	 * 
	 * @param restrictions		The restrictions imposed to select the new ActivityManagementProportion
	 * @return					A list of all the proportions that match the selection criteria
	 */
	public List<ActivityManagementProportion> getAMPFilteredList(Map<LandUseSelectionRestriction, Object> restrictions) {
		// 1. Select by region (mandatory)
		Region region = (Region) restrictions.get(LandUseSelectionRestriction.RegionRestriction);
		Manager manager = Manager.getInstance();
		Scenario scenario = manager.getScenario();
		List<ActivityManagementProportion> proportions = manager.getActivityManagementProportions(region);
		List<ActivityManagementProportion> filteredProportions = new ArrayList<ActivityManagementProportion>();

		// 1b. Select managements available to farmer
		// TODO: Me parece que todo este bloque esta demas
		List<ActivityManagement> availableManagements = null;
		if (restrictions.containsKey(LandUseSelectionRestriction.MinimumActivityGroupProportionRestriction) || restrictions.containsKey(LandUseSelectionRestriction.MaximumActivityGroupProportionPRestriction)) {
			List<ActivityManagement> regionManagements    = ActivityManagementFactory.getInstance(scenario).findByRegion(region);
			if (scenario.hasAttribute(Scenario.Attribute.HasTechnologyInnovations)) {
				availableManagements = new ArrayList<ActivityManagement>();
				
				List<?> unknownManagements = (List<?>) restrictions.get(LandUseSelectionRestriction.TechnologyRestriction);
				for (ActivityManagement activityManagement: regionManagements) {
					if (activityManagement.isAvailable() && ! unknownManagements.contains(activityManagement)) {
						availableManagements.add(activityManagement);
					}
				}
			} else {
				// All managements are available.
				availableManagements = new ArrayList<ActivityManagement>(regionManagements);
			}
		}
			
		// 2. Apply other filters
		for (int i = 0; i < proportions.size(); i++) {
			ActivityManagementProportion amp = proportions.get(i);
			boolean matches = true;
			
			// 2.1 Unknown technology restriction. The policy is to EXCLUDE those activity-managements not known by the farmer
			if (matches && scenario.hasAttribute(Scenario.Attribute.HasTechnologyInnovations)) {
				// 2.1.1 Check if AMP contains any unavailable management
				List<ActivityManagement> managements = amp.getDistinctActivityManagements();
				int j = 0;
				while ((j < managements.size()) && matches) {
					ActivityManagement management = managements.get(j);
					matches = management.isAvailable();
					j++;
				}
				
				// 2.1.2 Check if AMP contains any unknown management
				List<?> unknownManagements = (List<?>) restrictions.get(LandUseSelectionRestriction.TechnologyRestriction);
				j = 0;
				while ((j < unknownManagements.size()) && matches) {
					ActivityManagement management = (ActivityManagement) unknownManagements.get(j);
					matches = (amp.getProportion(management) == 0);
					j++;
				}				
			}
				
			// 2.2 Maximum & minimum restriction
			if (matches && (restrictions.containsKey(LandUseSelectionRestriction.MinimumActivityGroupProportionRestriction) || restrictions.containsKey(LandUseSelectionRestriction.MaximumActivityGroupProportionPRestriction))) {
				Integer minimum = (Integer) restrictions.get(LandUseSelectionRestriction.MinimumActivityGroupProportionRestriction);
				Integer maximum = (Integer) restrictions.get(LandUseSelectionRestriction.MaximumActivityGroupProportionPRestriction);
						
				// 2.2a Calculate group proportions for current AMP.
				Map<ActivityGroup,Integer> groupProportions = new HashMap<ActivityGroup,Integer>(); 
				for (ActivityManagement activityManagement: availableManagements) {
					ActivityGroup activityGroup  = activityManagement.getActivity().getActivityGroup();
					Integer managementProportion = amp.getProportion(activityManagement);
					Integer groupProportion      = groupProportions.containsKey(activityGroup) ? groupProportions.get(activityGroup) : 0;
					groupProportions.put(activityGroup, managementProportion + groupProportion);
				}
				
				// 2.2b Check minimum and maximum proportions for each ActivityGroup
				Iterator<ActivityGroup> activityGroups = groupProportions.keySet().iterator();
				while (activityGroups.hasNext() && matches) {
					ActivityGroup activityGroup = activityGroups.next();
					Integer groupProportion     = groupProportions.get(activityGroup);
					matches = (((minimum == null) || (groupProportion >= minimum)) && ((maximum == null) || (groupProportion <= maximum)));
				}
			}
						
			if (matches) filteredProportions.add(amp);
		}
				
		return filteredProportions;
	}
	
	public LandUseSelectionMechanism getLastLandUseSelectionMechanism() {
		return this.finalLandUseSelectionMechanism;
	}
	
	/**
	 * This method finds the next AMP for the given farm using this mechanism if it's possible. If not, every mechanism
	 * must support default mechanisms in order to return a result. The mechanism used must be set in attribute
	 * finalLandUseSelectionMechanism
	 * 
	 * @param farm	The farm that is being tested
	 * @return		New AMP for given farm
	 */
	public abstract ActivityManagementProportion search(Farm farm, Map<LandUseSelectionRestriction, Object> restrictions);

	public String toString() {
		return this.landUseSelectionMechanismId.toString();
	}
}
