package fiuba.lamm.agromodel.decision;

import fiuba.lamm.agromodel.ActivityManagement;
import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.execution.Scenario;
import fiuba.lamm.agromodel.factories.ActivityManagementFactory;
import fiuba.lamm.agromodel.finance.DirectCost;
import fiuba.lamm.agromodel.math.custom.*;
import fiuba.lamm.agromodel.utils.*;
import java.util.*;

public class OptimizationLandUseSelectionMechanism extends LandUseSelectionMechanism {
	protected FitnessFunction fitnessFunction;
		
	public OptimizationLandUseSelectionMechanism(FitnessFunction fitnessFunction) {
		this.landUseSelectionMechanismId = fitnessFunction.getLandUseSelectionMechanismType();
		this.fitnessFunction = fitnessFunction;
	}
	
	public FitnessFunction getFitnessFunction() {
		return this.fitnessFunction;
	}
	
	public ActivityManagementProportion search(Farm farm, Map<LandUseSelectionRestriction, Object> restrictions) {
		// Busco la mejor alternativa dado las restricciones del campo y del operador
		List<ActivityManagementProportion> bestAmps = new ArrayList<ActivityManagementProportion>();
		List<ActivityManagementProportion> ampFullList = getAMPFilteredList(restrictions);
		Manager manager = Manager.getInstance();
		Scenario scenario = manager.getScenario();
		Integer croppingCyclesForUtilityCalculation = (Integer) scenario.getParameter("LandUseSelection.CroppingCyclesForUtilityCalculation");
		Double maxExpectedUtility = Double.NEGATIVE_INFINITY;

		Double refValue = null;
		if (fitnessFunction.getLandUseSelectionMechanismType().equals(Type.EUOptimizationSelectionMechanism)) {
			/*
			 *  Calcular w0:
			 *  	  w0 es una medida del activo del operador. En el caso de un propietario 
			 *        es el capital liquido mas el 40% del valor de su campo (teniendo en cuenta su area) al momento de optimizar.
			 *        En el caso de un arrendatario, el w0 es el capital liquido mas el costo de 
			 *        implantar durante 2 ciclos el AM mas caro en el 100% del campo (teniendo en cuenta su area)
			 *        
			 *        El valor de x debe ser el ggm multiplicado por el tamano del campo
			 */	
			Farmer operator = farm.getOperator();	
			Double workingCapital = operator.getWorkingCapital();			
			if (operator.equals(farm.getOwner())) {
				Double maxMortgageLoan = 0.4 * farm.getSoilType().getLandSalePrice() * farm.getFarmSize();
				refValue = workingCapital + maxMortgageLoan;
			} else {
				// Tenant
				// El costo de implantacion no tiene en cuenta economia de escalas
				ActivityManagement mostExpensiveManagement = ActivityManagementFactory.getInstance(manager.getScenario()).findMostExpensive(manager.getCurrentCroppingCycle(), farm.getRegion());
				Double maxLoan = 0.0;				
				while (mostExpensiveManagement != null) {
					maxLoan += mostExpensiveManagement.getDirectCost(DirectCost.FixedCost) * farm.getFarmSize();
					mostExpensiveManagement = mostExpensiveManagement.getNextActivityManagement();
				}
				maxLoan *= 2.0; // Dos veces el costo de implantacion del AM mas caso
								  
				refValue = workingCapital + maxLoan;
			}
		} else if (fitnessFunction.getLandUseSelectionMechanismType().equals(Type.CPTOptimizationSelectionMechanism)){
			// Calculo del valor de referencia de acuerdo al nivel de aspiracion
			refValue = farm.getAspirationLevel().getLevel() * farm.getFarmSize();
		}		
		this.fitnessFunction.setCurrentReferenceValue(refValue);
		
		for (ActivityManagementProportion amp: ampFullList) {
			double globalGrossMargins[] = manager.getLastGlobalGrossMargins(amp, farm.getSoilType(), croppingCyclesForUtilityCalculation);
			
			if (globalGrossMargins != null) {
				// Multiplico los global gross margins posibles por el tamano del campo para que sean
				// comparables a w0, ya que los global gross margins estan calculados para campos de 1ha
				// Ademas, debo restar los costos indirectos
				double x[] = new double[globalGrossMargins.length];
				int i = 0;
				for (double ggm: globalGrossMargins) {
					x[i] = ggm * farm.getFarmSize();
					i++;
				}
			
				Double expectedUtility = this.fitnessFunction.apply(x);
				if (expectedUtility >= maxExpectedUtility) {
					if (expectedUtility > maxExpectedUtility) bestAmps.clear();
					bestAmps.add(amp);
					maxExpectedUtility = expectedUtility;				
				}				
			}
		}
		
		// Selecciono un AMP de la lista de "mejores" y seteo el mecanismo definitivo de seleccion
		ActivityManagementProportion amp = null;
		if (bestAmps.size() > 1) {
			// Elijo uno de los AMPs al azar
			Map<ActivityManagementProportion, Double> proportions = new HashMap<ActivityManagementProportion, Double>();
			for (ActivityManagementProportion candidate: bestAmps) {
				proportions.put(candidate, 1.0 / (double) bestAmps.size());
			}
			
			Randomizer randomizer = scenario.getRandomizer();
			amp = randomizer.getRandomEvent(proportions);
		} else {
			amp = bestAmps.get(0);
		}
		
		// Set expected utility form selected AMP
		farm.setExpectedUtility(maxExpectedUtility);
		
		// Set optimization as current land use selection mechanism and return AMP
		this.finalLandUseSelectionMechanism = this;		
		return amp;
	}
}