package fiuba.lamm.agromodel.diffusion.ising;

import java.util.*;
import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.decision.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.factories.*;
import fiuba.lamm.agromodel.finance.*;
import fiuba.lamm.agromodel.math.*;
import fiuba.lamm.agromodel.math.custom.*;

public class UtilityBasedIsingComponentsCalculationMechanism extends IsingComponentsCalculationMechanism {
	public Double getInnovationUtility(Farm farm, ActivityManagement management, Integer croppingCyclesForUtilityCalculation) {
		Manager manager = Manager.getInstance();
		Scenario scenario = manager.getScenario();
		ActivityManagement managements[] = { management, management.getCompetingManagement() };
		Map<ActivityManagement,Double> certaintyEquivalents = new HashMap<ActivityManagement,Double>();
		SoilType soilType = farm.getSoilType();
		Farmer operator = farm.getOperator();
		Farmer owner = farm.getOwner();
					
		// 1. Obtengo la funcion de fitness. Cuando el farmer no tiene LUSM con optimizacion, entonces el
		//    es un laggard que no tiene en cuenta la percepcion propia
		LandUseSelectionMechanism landUseSelectionMechanism = operator.getLandUseSelectionMechanism();
		OptimizationLandUseSelectionMechanism m = (OptimizationLandUseSelectionMechanism) landUseSelectionMechanism;
		FitnessFunction fitnessFunction = m.getFitnessFunction();			
		
		// 2. Calculo el valor de referencia
		Double refValue = null;
		if (fitnessFunction.getLandUseSelectionMechanismType().equals(LandUseSelectionMechanism.Type.EUOptimizationSelectionMechanism)) {
			/*
			 *  Calcular w0:
			 *  	  w0 es una medida del activo del operador. En el caso de un propietario 
			 *        es el capital liquido mas el 40% del valor de su campo (teniendo en cuenta su area) al momento de optimizar.
			 *        En el caso de un arrendatario, el w0 es el capital liquido mas el costo de 
			 *        implantar durante 2 ciclos el AM mas caro en el 100% del campo (teniendo en cuenta su area)
			 *        
			 *        El valor de x debe ser el ggm multiplicado por el tamano del campo
			 */
			Double workingCapital = operator.getWorkingCapital();
			if (operator.equals(owner)) {
				Double maxMortgageLoan = 0.4 * soilType.getLandSalePrice() * farm.getFarmSize();
				refValue = workingCapital + maxMortgageLoan;
			} else {
				// Tenant
				// El costo de implantacion no tiene en cuenta economia de escalas
				ActivityManagement mostExpensiveManagement = ActivityManagementFactory.getInstance(manager.getScenario()).findMostExpensive(manager.getCurrentCroppingCycle(), farm.getRegion());
				Double maxLoan = 0.0;				
				while (mostExpensiveManagement != null) {
					maxLoan += mostExpensiveManagement.getDirectCost(DirectCost.FixedCost) * farm.getFarmSize();
					mostExpensiveManagement = mostExpensiveManagement.getNextActivityManagement();
				}
				maxLoan *= 2.0; // Dos veces el costo de implantacion del AM mas caro
								  
				refValue = workingCapital + maxLoan;
			}
		} else if (fitnessFunction.getLandUseSelectionMechanismType().equals(LandUseSelectionMechanism.Type.CPTOptimizationSelectionMechanism)){
			// Calculo del valor de referencia de acuerdo al nivel de aspiracion
			refValue = farm.getAspirationLevel().getLevel() * farm.getFarmSize();
		}		
		fitnessFunction.setCurrentReferenceValue(refValue);
			
		// 3. Calculo del CE para cada AMP
		Integer numberOfPlots = (Integer) scenario.getParameter("Farm.NumberOfPlots");
		for (ActivityManagement mgmt: managements) {
			ActivityManagementProportion amp = new ActivityManagementProportion();
			amp.setProportion(mgmt, numberOfPlots);
			double ggms[] = manager.getLastGlobalGrossMargins(amp, soilType, croppingCyclesForUtilityCalculation);
			
			// Compute final wealth for each global gross margin
			double wfs[] = new double[ggms.length];
			for (int i = 0; i < ggms.length; i++) {
				wfs[i] = ggms[i] * farm.getFarmSize();
			}
			
			// Compute expected utility and certainty equivalent
			Double utility = fitnessFunction.apply(wfs);
			Double ce = fitnessFunction.getCertaintyEquivalent(utility);
			certaintyEquivalents.put(mgmt, ce);			
		}
			
		// 4. Calculo del du a priori
		Double innovationCE = certaintyEquivalents.get(management);
		Double competingCE = certaintyEquivalents.get(management.getCompetingManagement());
		Double du = null;
		if (competingCE != 0.0) {
			du = (innovationCE - competingCE) / competingCE;	
		} else if (innovationCE == 0.0) {
			du = 0.0; // Ya que la innovation tiene la misma utilidad que el producto de la competencia
		} else {
			du = Math.signum(innovationCE); // De acuerdo al signo, asigno +1 o -1
		}
					
		// 5. Calculo del du normalizado
		Function utilityScalingFunction = new UtilityScalingFunction(-0.1, 0.1);
		Double ndu = utilityScalingFunction.apply(du);		
		return ndu;	
	}
}