package fiuba.lamm.agromodel.execution;

import java.io.*;
import java.util.*;

public class BatchExecutor {
	protected List<ExecutionHandler> handlers;
	protected List<Short> scenarioId;
	protected Map<Short,Map<RunInfo,String>> scenarioRunInfo;
	protected NotificationHandler notificationHandler;
	protected String repositoryUrl;
	protected String repositoryUserName;
	protected String repositoryPassword;
	protected String ftpHost;
	protected String ftpUserName;
	protected String ftpPassword;
	protected String ftpPathName;
	protected String ftpUploadUrl;
	protected Boolean loadCommonTables;	
	protected Boolean loadScenarioTables;
	protected Integer randomSeed;
	protected Short maxThreads;	
	protected String recipientName;
	protected String recipientEmail;
	
	protected BatchExecutor(List<Short> scenarioId) {
		/*
		 * Carga de parametros para la ejecucion distribuida
		 */
		String projectRootLocation = java.lang.System.getenv("PROJECT_ROOT");
		if ((projectRootLocation != null) && !projectRootLocation.trim().equals("")) {
			File projectRoot = new File(projectRootLocation);
			File configFile = new File(projectRoot, "config/distributed.batch.properties");
			
			Properties properties = new Properties();
			try {
				properties.load(new FileReader(configFile));
			} catch (IOException ioe) {
				throw new FulminantError("Error reading configuration file: " + ioe);
			}
			
			this.scenarioId = scenarioId;
			this.scenarioRunInfo = new HashMap<Short,Map<RunInfo,String>>();
			this.repositoryUrl = properties.getProperty("fiuba.lamm.agromodel.batch.repository.url");
			this.repositoryUserName = properties.getProperty("fiuba.lamm.agromodel.batch.repository.user");
			this.repositoryPassword = properties.getProperty("fiuba.lamm.agromodel.batch.repository.password");
			this.ftpHost = properties.getProperty("fiuba.lamm.agromodel.batch.ftp.host");
			this.ftpUserName = properties.getProperty("fiuba.lamm.agromodel.batch.ftp.user");
			this.ftpPassword = properties.getProperty("fiuba.lamm.agromodel.batch.ftp.password");
			this.ftpPathName = properties.getProperty("fiuba.lamm.agromodel.batch.ftp.path");
			this.ftpUploadUrl = properties.getProperty("fiuba.lamm.agromodel.batch.ftp.upload.url");
			this.recipientName = properties.getProperty("fiuba.lamm.agromodel.batch.notification.recipient.name");
			this.recipientEmail = properties.getProperty("fiuba.lamm.agromodel.batch.notification.recipient.email");
			this.loadCommonTables = new Boolean(properties.getProperty("fiuba.lamm.agromodel.batch.load.common.tables"));
			this.loadScenarioTables = new Boolean(properties.getProperty("fiuba.lamm.agromodel.batch.load.scenario.tables"));
			this.randomSeed = new Integer(properties.getProperty("fiuba.lamm.agromodel.batch.run.randomseed"));
			this.maxThreads = new Short(properties.getProperty("fiuba.lamm.agromodel.batch.run.maxthreads"));			
		} else {
			throw new FulminantError("Environment variable PROJECT_ROOT not set");			
		}		
	}
	
	protected void init() {
		/*
		 *  Handlers para la ejecucion de escenario. Se ejecutan en este orden y si uno falla
		 *  los siguientes no se ejecutan.
		 */
		this.handlers = new ArrayList<ExecutionHandler>();
		handlers.add(new AuthenticationHandler(this.repositoryUrl, this.repositoryUserName, this.repositoryPassword));
		handlers.add(new WorkingCopyUpdateHandler(this.repositoryUserName, this.repositoryPassword));
		if (this.loadCommonTables) handlers.add(new CommonTablesLoaderHandler());
		if (this.loadScenarioTables) handlers.add(new ScenarioLoaderHandler(this.scenarioId));
		handlers.add(new SimulationExecutionHandler(this.scenarioId, this.randomSeed, this.maxThreads, this.scenarioRunInfo));
		handlers.add(new ResultsUploadHandler(this.repositoryUserName, this.repositoryPassword, this.ftpHost, this.ftpUserName, this.ftpPassword, this.ftpPathName, this.scenarioId, this.scenarioRunInfo));
		
		/*
		 * Handler de notificacion. Se ejecuta siempre cualquiera fuere el resultado de
		 * la ejecucion
		 */
		this.notificationHandler = new NotificationHandler(this.ftpUploadUrl, this.recipientName, this.recipientEmail, scenarioId);
	}
	
	protected void run() {
		// Execute simulation
		String error = null;
		try {			
			for (ExecutionHandler handler: handlers) {
				System.out.print("Executing " + handler.getClass().getSimpleName() + "... ");
				Long handlerStartTime = System.currentTimeMillis();
				handler.execute();
				Long handlerEndTime = System.currentTimeMillis();
				Double elapsedSeconds = (handlerEndTime.doubleValue() - handlerStartTime.doubleValue()) / 1000.0;
				System.out.println(" [" + fiuba.lamm.agromodel.utils.System.elapsedTime(elapsedSeconds) + "]");
			}
		} catch (ExecutionException ee) {
			StringWriter writer = new StringWriter();
			ee.printStackTrace(new PrintWriter(writer));
			error = writer.toString();
		}
		
		// Notify execution result
		try {
			if (error != null) {
				this.notificationHandler.setErrorMessage(error);
			}
			
			this.notificationHandler.execute();
		} catch (ExecutionException ee) {
			System.err.println("Error notifying results: " + ee);
			ee.printStackTrace(System.err);
			System.exit(1);
		}
	}
	
	protected static List<Short> getScenarioList(String scenarioIdString) {
		if ((scenarioIdString == null) || !scenarioIdString.matches("\\d+([,\\-]\\d+)*")) {
			return null;
		}
		
		List<Short> list = new ArrayList<Short>();
		String s1[] = scenarioIdString.split(",");
		for (String a: s1) {
			if (a.matches("\\d+\\-\\d+")) {
				String s2[] = a.split("\\-");
				Short from = new Short(s2[0]);
				Short to = new Short(s2[1]);
				
				for (short b = from; b <= to; b++) {
					Short scenarioId = new Short(b);
					if (!list.contains(scenarioId)) list.add(scenarioId);
				}
			} else {
				Short scenarioId = new Short(a);
				if (!list.contains(scenarioId)) list.add(scenarioId);				
			}
		}
		
		return list;
	}
	
	protected static void showUsageAndExit() {
		String usageMessage = "Parameters are: USER_NAME PASSWORD LOAD_COMMON_TABLES LOAD_SCENARIOS SCENARIO_ID [RANDOM_SEED]\n";
		usageMessage += "\nUSER_NAME = User login for SVN authentication";
		usageMessage += "\nPASSWORD = User password for SVN authentication";
		usageMessage += "\nE-MAIL = User e-mail for notification";
		usageMessage += "\nLOAD_COMMON_TABLES = [ 1 = load common tables before running simulatios, 0 = otherwise ]";
		usageMessage += "\nLOAD_SCENARIOS = [ 1 = load scenarios before running simulatios, 0 = otherwise ]";
		usageMessage += "\nSCENARIO_ID = A list of scenarios separated by commas. To specify a range a hyphen may be used";
		usageMessage += "\nRANDOM_SEED = Random seed used in simulations";
		usageMessage += "\nMAX_PROCESSORS = Maximum number of processors available for simulation execution\n";
		System.err.println(usageMessage);
		System.exit(1);
	}
	
	public static void main(String args[]) {
		// 1. Parse parameters		
		List<Short> scenarioId = BatchExecutor.getScenarioList(args[0]); 
		if (scenarioId == null) {
			BatchExecutor.showUsageAndExit();
		}
		
		// 2. Create batch executor and run
		BatchExecutor executor = new BatchExecutor(scenarioId);
		executor.init();
		executor.run();
	}
}