package fiuba.lamm.agromodel.execution;

import fiuba.lamm.agromodel.utils.compression.*;
import java.io.*;
import java.util.*;

import org.apache.commons.net.ftp.*;
import org.tmatesoft.svn.core.*;
import org.tmatesoft.svn.core.auth.*;
import org.tmatesoft.svn.core.wc.*;

public class ResultsUploadHandler implements ExecutionHandler {
	protected String userName;
	protected String password;
	protected String ftpHost;
	protected String ftpUserName;
	protected String ftpPassword;
	protected String ftpPathName;
	protected List<Short> scenarioId;
	protected Map<Short,Map<RunInfo,String>> scenarioRunInfo;
	
	public ResultsUploadHandler() { } // Just for testing purposes
		
	public ResultsUploadHandler(String userName, String password, String ftpHost, String ftpUserName, String ftpPassword, String ftpPathName, List<Short> scenarioId, Map<Short,Map<RunInfo,String>> scenarioRunInfo) {
		this.userName = userName;
		this.password = password;
		this.ftpHost = ftpHost;
		this.ftpUserName = ftpUserName;
		this.ftpPassword = ftpPassword;
		this.ftpPathName = ftpPathName;
		this.scenarioId = scenarioId;
		this.scenarioRunInfo = scenarioRunInfo;
	}
	
	public void execute() throws ExecutionException {
		// 1. Setup
		File projectRoot = new File(System.getenv("PROJECT_ROOT"));
		File outputDir = new File(projectRoot, "output");
		if (! outputDir.isDirectory()) {
			throw new ExecutionException("Output directory not found: " + outputDir.getAbsolutePath());
		}
		
		// 2. Create package files
		List<File> packageFiles = createPackageFiles(outputDir);
		
		// 3. Update repository
		updateRepository(packageFiles);
		
		// 4. Synchronize with external repository
		if (this.ftpHost != null) {
			uploadResults(packageFiles);
		}
	}
	
	public List<File> createPackageFiles(File outputDir) throws ExecutionException {
		String outputFilePrefix[] = { "amp", "manager", "farms", "farmers", "plots", "soiltypes" };
				
		// 1. Check temporary output files
		File tmpOutputDir = new File(outputDir, "tmp");
		if (! tmpOutputDir.isDirectory()) {
			throw new ExecutionException("Temporary output directory not found: " + tmpOutputDir.getAbsolutePath());
		}
		Map<Short, List<File>> tmpOutputFiles = new HashMap<Short, List<File>>();
		for (Short sId: scenarioId) {
			// 1a. Add output files
			List<File> scenarioTmpOutputFiles = new ArrayList<File>();
			int i = 0;
			while (i < outputFilePrefix.length) {
				String prefix = outputFilePrefix[i];
				File tmpFile = new File(tmpOutputDir, prefix + "_scenario_" + sId + ".csv");
				if (tmpFile.isFile()) {
					scenarioTmpOutputFiles.add(tmpFile);
				}
				
				i++;
			}
			
			// 1b. Add run info
			Map<RunInfo,String> runInfo = this.scenarioRunInfo.get(sId);
			if (runInfo.containsKey(RunInfo.Output)) {
				File scenarioRunInfoFile = new File(tmpOutputDir, "run_info_scenario_" + sId + ".log");
				try {
					PrintWriter writer = new PrintWriter(new FileWriter(scenarioRunInfoFile));
					writer.print(runInfo.get(RunInfo.Output));
					writer.close();
					scenarioTmpOutputFiles.add(scenarioRunInfoFile);
				} catch (IOException ioe) {
					throw new ExecutionException("Error while writer scenario run info file for scenario " + sId + ": " + ioe);
				}
			}
			
			tmpOutputFiles.put(sId, scenarioTmpOutputFiles);
		}
		
		// 2. Create compressed files for each scenario and move them to output directory
		List<File> compressedFiles = new ArrayList<File>();
		File finalOutputDir = new File(outputDir, "final");
		if (! finalOutputDir.isDirectory()) {
			throw new ExecutionException("Final output directory not found: " + finalOutputDir.getAbsolutePath());
		}	
		for (Short sId: tmpOutputFiles.keySet()) {
			File outputFile = new File(finalOutputDir, sId + ".rar");
			
			// 2.1 Eliminacion del rar existente (si es que habia)
			if (outputFile.isFile()) {
				outputFile.delete();
			}
			
			// 2.2 Creacion del nuevo rar
			List<File> inputFiles = tmpOutputFiles.get(sId);
			try {
				RarCompressor compressor = new RarCompressor();
				compressor.compress(inputFiles, outputFile, true);
				compressedFiles.add(outputFile);
			} catch (IOException ioe) {
				throw new FulminantError("Error while creating RAR file for scenario " + sId + ": " + ioe);
				
			}
		}
		
		return compressedFiles;
	}
	
	public void uploadResults(List<File> packageFiles) throws ExecutionException {
		FTPClient client = new FTPClient();
		
		try {
			// 1. Connect to FTP server and set working directory
			client.connect(ftpHost);
			if (! client.login(ftpUserName, ftpPassword)) {
				throw new ExecutionException("Error connecting to " + ftpHost + " using account " + ftpUserName + "/" + ftpPassword);
			}
			client.changeWorkingDirectory(ftpPathName);
			
			// 2. Results upload
			for (File packageFile: packageFiles) {
				client.setFileType(FTP.BINARY_FILE_TYPE);
				InputStream inputStream = new FileInputStream(packageFile);
				try {
					if (! client.storeFile(packageFile.getName(), inputStream)) {						
						throw new ExecutionException("Error uploading file " + packageFile.getAbsolutePath() + " to " + client.printWorkingDirectory() + ": " + client.getReplyString());
					}
				} finally {
					inputStream.close();
				}
			}
			
			// 3. Logout
			client.logout();
		} catch (IOException ioe) {
			throw new ExecutionException("Error uploading results to FTP server: " + ioe);
		}
	}
	
	public void updateRepository(List<File> packageFiles) throws ExecutionException {		
		// 1. Add new files
		BasicAuthenticationManager authenticationManager = new BasicAuthenticationManager(userName, password);
		SVNWCClient svnWCClient = new SVNWCClient(authenticationManager, null);
		for (File file: packageFiles) {
			try {
				svnWCClient.doInfo(file, SVNRevision.HEAD);
			} catch (SVNException svne) {
				// File does not exists
				try {
					svnWCClient.doAdd(file, false, false, false, SVNDepth.FILES, true, false);
				} catch (SVNException svne2) {
					throw new ExecutionException("Error adding file: " + file.getAbsolutePath());
				}
			}
		}
		
		// 2. Commit compressed files
		String commitMessage = "Execution of ";
		if (scenarioId.size() > 1) {
			commitMessage += " scenarios " + scenarioId.get(0);
			for (int i = 1; i < scenarioId.size(); i++) {
				commitMessage += "," + scenarioId.get(i);
			}
		} else {
			commitMessage = " scenario " + scenarioId.get(0);
		}
		
		SVNCommitClient svnCommitClient = new SVNCommitClient(authenticationManager, null);
		try {
			svnCommitClient.doCommit(packageFiles.toArray(new File[0]), false, commitMessage, null, null, false, false, SVNDepth.FILES);
		} catch (SVNException svne) {
			throw new ExecutionException("Error storing output into the repository: " + svne);
		}		
	}
}