package fiuba.lamm.agromodel.execution;

import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.output.*;
import fiuba.lamm.agromodel.networks.*;
import fiuba.lamm.agromodel.utils.*;
import java.io.*;
import java.util.*;

public class Scenario implements Comparable<Scenario> {
	public enum Attribute {
		IsTest									(1), // Is for testing purpose
		ProgressRateNOutOfM						(2), // Indica si utiliza el metodo de N out of M para calcular el progress rate
		HasTechnologyInnovations				(4), // Has at least one AM introduces as an innovation
		HasScaleEconomics						(8), // Margin calculation should use scale factors
		HasIncomeTax							(16), // Income tax should be applied
		HasVariableWithdrawal					(32), // Agent's withdrawal is a function of income 
		HasWealthWeightedSelectionMechanism		(64), // Market clearing uses a wealth-weighted selection policy for bidding
		HasMemoryForWithdrawal					(128), // If HasVariableWithdrawal, then the agent may have memory (5 years) or not 
		VariableReenteringProbability			(256), // If ON, landlords have a time-dependent probability of regaining control of their farms. If OFF, probability = 0.25
		AllowFarmersMigration   				(512), // Indica si se consideran o no los campos de regiones no preferidas en el mercado de alquiler
		UseAdvisorsInDiffusionProcess			(1024), // If ON, AACREA & external advisors are introduced in the diffusion of innovations process
		UseAdvancedUtilityInDiffusionProcess	(2048), // If ON, salesmen+uncertainty are introduced in the diffusion of innovations process
		HasGrossMarginPerturbation				(4096), // Activate Luck factor for each farm on each cropping cycle
		AllowFarmerRetirement					(8192), // Indica si se acepta el retiro de los farmers
		HasYieldAdjustment					    (16384), // Indica si se realizar ajuste de rindes mediante la interaccion con MIKE		
		DisableTechnologyAdjustmentUpdate		(32768); // Deshabilita la actualizacion de GGMs por mejoras tecnologicas
		
		private int bit;
		private Attribute(int bit) {
			this.bit = bit;
		}
		
		public int getBit() {
			return this.bit;
		}
		
		public static Map<Attribute,Boolean> getAttributes(int bits) {
			Map<Attribute,Boolean> attributes = new HashMap<Attribute,Boolean>();
			for(Attribute attribute: Attribute.values()) {
				boolean active = ((attribute.bit & bits) != 0);
				attributes.put(attribute, active);
			}
			
			return attributes;
		}
	}
	
	protected Short scenarioId;
	protected String name;
	protected String description;
	protected Map<Scenario.Attribute,Boolean> attributes;
	protected Map<String,?> parameters;
	protected ScenarioEnsemble scenarioEnsemble;
	protected Manager manager;
	protected Randomizer randomizer;
	protected FarmNetworkModel farmNetworkModel;

	public Scenario(Short scenarioId, String name, String description, Integer attributes, Map<String,?> parameters, String shapeFileName, ScenarioEnsemble scenarioEnsemble) {
		this.scenarioId = scenarioId;
		this.name = name;
		this.description = description;
		this.attributes = Scenario.Attribute.getAttributes(attributes);
		this.parameters = new HashMap<String,Object>(parameters);
		this.scenarioEnsemble = scenarioEnsemble;
		this.manager = null;
		this.randomizer = null;
		
		if (shapeFileName != null) {
			File projectRoot = new File(java.lang.System.getProperty("fiuba.lamm.agromodel.workdir"));
			File gisDir = new File(projectRoot, "input" + File.separator + "GIS");
			File shapeFile = new File(gisDir, shapeFileName);
			
			this.farmNetworkModel = new GISNetworkModel((Double) this.parameters.get("Farm.SpatialNetworkInfluenceRadius"), shapeFile);
		} else {
			this.farmNetworkModel = new MooreNetworkModel();
		}
	}
	
	// Accessors
	public Short getScenarioId() {
		return this.scenarioId;
	}
	public String getName() {
		return this.name;
	}
	public String getDescription() {
		return this.description;
	}
	public Boolean hasAttribute(Scenario.Attribute attribute) {
		return this.attributes.get(attribute);
	}
	public Object getParameter(String paramName) {		
		return this.parameters.get(paramName);
	}
	public ScenarioEnsemble getScenarioEnsemble() {
		return this.scenarioEnsemble;
	}
	public FarmNetworkModel getFarmNetworkModel() {
		return this.farmNetworkModel;		
	}
	
	public Manager getManager() {
		return this.manager;
	}
	public void setManager(Manager manager) {
		if ((this.manager != null) && (manager == null)) {
			// Destroy manager before assignment
			this.manager.destroy();	
		}
		this.manager = manager;
	}
	
	public void setRandomSeed(Integer randomSeed) {
		this.randomizer = new Randomizer(randomSeed);
	}
	public Randomizer getRandomizer() {
		return this.randomizer;
	}
	
	public Verbosity getVerbosityLevel() {
		if (this.hasAttribute(Attribute.IsTest)) {
				return Verbosity.Verification;
		} else if (this.scenarioEnsemble != null) {
				return Verbosity.Reduced;
		} else {
				return Verbosity.Normal;
		}		
	}
	
	public int hashCode() {
		return this.scenarioId;		
	}
	
	public boolean equals(Object o) {
		if (o instanceof Scenario) {
			Scenario s = (Scenario) o;
			return s.scenarioId.equals(this.scenarioId);
		}
		
		return false;
	}
	
	public int compareTo(Scenario scenario) {
		return this.scenarioId.compareTo(scenario.getScenarioId());
	}
}