package fiuba.lamm.agromodel.execution.ensemble;

import fiuba.lamm.agromodel.execution.*;

import java.io.*;
import java.util.*;

public class BatchExecutor {
	protected List<ExecutionHandler> handlers;
	protected Short scenarioEnsembleId;
	protected Map<Short,Map<RunInfo,String>> scenarioRunInfo;
	protected NotificationHandler notificationHandler;
	protected String repositoryUrl;
	protected String repositoryUserName;
	protected String repositoryPassword;
	protected String ftpHost;
	protected String ftpUserName;
	protected String ftpPassword;
	protected String ftpPathName;
	protected String ftpUploadUrl;
	protected Boolean loadCommonTables;	
	protected Boolean loadScenarioEnsembleTables;
	protected Integer randomSeed;
	protected Short maxThreads;
	protected String recipientName;
	protected String recipientEmail;
	
	protected BatchExecutor(Short scenarioEnsembleId) {
		/*
		 * Carga de parametros para la ejecucion distribuida
		 */
		String projectRootLocation = java.lang.System.getenv("PROJECT_ROOT");
		if ((projectRootLocation != null) && !projectRootLocation.trim().equals("")) {
			File projectRoot = new File(projectRootLocation);
			File configFile = new File(projectRoot, "config/distributed.ensemble.batch.properties");
			
			Properties properties = new Properties();
			try {
				properties.load(new FileReader(configFile));
			} catch (IOException ioe) {
				throw new FulminantError("Error reading configuration file: " + ioe);
			}
			
			this.scenarioEnsembleId = scenarioEnsembleId;
			this.scenarioRunInfo = new HashMap<Short,Map<RunInfo,String>>();
			this.repositoryUrl = properties.getProperty("fiuba.lamm.agromodel.batch.repository.url");
			this.repositoryUserName = properties.getProperty("fiuba.lamm.agromodel.batch.repository.user");
			this.repositoryPassword = properties.getProperty("fiuba.lamm.agromodel.batch.repository.password");
			this.ftpHost = properties.getProperty("fiuba.lamm.agromodel.batch.ftp.host");
			this.ftpUserName = properties.getProperty("fiuba.lamm.agromodel.batch.ftp.user");
			this.ftpPassword = properties.getProperty("fiuba.lamm.agromodel.batch.ftp.password");
			this.ftpPathName = properties.getProperty("fiuba.lamm.agromodel.batch.ftp.path");
			this.ftpUploadUrl = properties.getProperty("fiuba.lamm.agromodel.batch.ftp.upload.url");
			this.recipientName = properties.getProperty("fiuba.lamm.agromodel.batch.notification.recipient.name");
			this.recipientEmail = properties.getProperty("fiuba.lamm.agromodel.batch.notification.recipient.email");
			this.loadCommonTables = new Boolean(properties.getProperty("fiuba.lamm.agromodel.batch.load.common.tables"));
			this.loadScenarioEnsembleTables = new Boolean(properties.getProperty("fiuba.lamm.agromodel.batch.load.ensemble.tables"));
			this.randomSeed = new Integer(properties.getProperty("fiuba.lamm.agromodel.batch.run.randomseed"));
			this.maxThreads = new Short(properties.getProperty("fiuba.lamm.agromodel.batch.run.maxthreads"));			
		} else {
			throw new FulminantError("Environment variable PROJECT_ROOT not set");			
		}		
	}
	
	protected void init() {
		/*
		 *  Handlers para la ejecucion de escenario. Se ejecutan en este orden y si uno falla
		 *  los siguientes no se ejecutan.
		 */		
		// Initialize factories
		fiuba.lamm.agromodel.utils.System.init();
		
		// Instantiate handlers
		this.handlers = new ArrayList<ExecutionHandler>();
		handlers.add(new AuthenticationHandler(this.repositoryUrl, this.repositoryUserName, this.repositoryPassword));
		handlers.add(new WorkingCopyUpdateHandler(this.repositoryUserName, this.repositoryPassword));
		if (this.loadCommonTables) handlers.add(new CommonTablesLoaderHandler());
		if (this.loadScenarioEnsembleTables) handlers.add(new ScenarioEnsembleLoaderHandler(this.scenarioEnsembleId));
		handlers.add(new SimulationExecutionHandler(this.scenarioEnsembleId, this.randomSeed, this.maxThreads, this.scenarioRunInfo));
		handlers.add(new ResultsUploadHandler(this.repositoryUserName, this.repositoryPassword, this.ftpHost, this.ftpUserName, this.ftpPassword, this.ftpPathName, this.scenarioEnsembleId, this.scenarioRunInfo));
		
		/*
		 * Handler de notificacion. Se ejecuta siempre cualquiera fuere el resultado de
		 * la ejecucion
		 */
		this.notificationHandler = new NotificationHandler(this.ftpUploadUrl, this.recipientName, this.recipientEmail, this.scenarioEnsembleId);
	}
	
	protected void run() {
		// Execute simulation
		String error = null;
		try {			
			for (ExecutionHandler handler: handlers) {
				System.out.print("Executing " + handler.getClass().getSimpleName() + "... ");
				Long handlerStartTime = System.currentTimeMillis();
				handler.execute();
				Long handlerEndTime = System.currentTimeMillis();
				Double elapsedSeconds = (handlerEndTime.doubleValue() - handlerStartTime.doubleValue()) / 1000.0;
				System.out.println(" [" + fiuba.lamm.agromodel.utils.System.elapsedTime(elapsedSeconds) + "]");
			}
		} catch (ExecutionException ee) {
			StringWriter writer = new StringWriter();
			ee.printStackTrace(new PrintWriter(writer));
			error = writer.toString();
		}
		
		// Notify execution result
		try {
			if (error != null) {
				this.notificationHandler.setErrorMessage(error);
			}
			
			this.notificationHandler.execute();
		} catch (ExecutionException ee) {
			System.err.println("Error notifying results: " + ee);
			ee.printStackTrace(System.err);
			System.exit(1);
		}
	}
		
	public static void main(String args[]) {
		// 1. Parse parameters		
		Short scenarioEnsembleId = null;
		try {
			scenarioEnsembleId = new Short(args[0]);
		} catch (NumberFormatException nfe) {
			System.err.println("Invalid Scenario Ensemble Id: " + nfe);
			System.exit(1);
		}
		
		// 2. Create batch executor and run
		BatchExecutor executor = new BatchExecutor(scenarioEnsembleId);
		executor.init();
		executor.run();
	}
}