package fiuba.lamm.agromodel.execution.ensemble;

import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.execution.factories.*;
import fiuba.lamm.agromodel.utils.compression.*;
import java.io.*;
import java.util.*;

import org.apache.commons.net.ftp.*;

public class ResultsUploadHandler implements ExecutionHandler {
	protected String userName;
	protected String password;
	protected String ftpHost;
	protected String ftpUserName;
	protected String ftpPassword;
	protected String ftpPathName;
	protected Short scenarioEnsembleId;
	protected List<Scenario> scenarios;
	protected Map<Short,Map<RunInfo,String>> scenarioRunInfo;
	
	public ResultsUploadHandler() { } // Just for testing purposes
		
	public ResultsUploadHandler(String userName, String password, String ftpHost, String ftpUserName, String ftpPassword, String ftpPathName, Short scenarioEnsembleId, Map<Short,Map<RunInfo,String>> scenarioRunInfo) {
		this.userName = userName;
		this.password = password;
		this.ftpHost = ftpHost;
		this.ftpUserName = ftpUserName;
		this.ftpPassword = ftpPassword;
		this.ftpPathName = ftpPathName;
		this.scenarioEnsembleId = scenarioEnsembleId;
		this.scenarios = new ArrayList<Scenario>();
		this.scenarioRunInfo = scenarioRunInfo;
	}
	
	public void execute() throws ExecutionException {
		// 1. Initialize scenarios
		this.scenarios.addAll(ScenarioFactory.getInstance().findByScenarioEnsembleId(this.scenarioEnsembleId));
		
		// 2. Setup
		File projectRoot = new File(System.getenv("PROJECT_ROOT"));
		File outputDir = new File(projectRoot, "output");
		if (! outputDir.isDirectory()) {
			throw new ExecutionException("Output directory not found: " + outputDir.getAbsolutePath());
		}
		
		// 3. Create package files
		File packageFile = createPackageFile(outputDir);
		
		// 4. Synchronize with external repository
		if (this.ftpHost != null) {
			uploadResults(packageFile);
		}
	}
	
	public File createPackageFile(File outputDir) throws ExecutionException {
		String outputFilePrefix[] = { "amp", "manager", "farms", "farmers", "plots", "soiltypes" };
		
		// 1. Check temporary output files
		File tmpOutputDir = new File(outputDir, "tmp");
		if (! tmpOutputDir.isDirectory()) {
			throw new ExecutionException("Temporary output directory not found: " + tmpOutputDir.getAbsolutePath());
		}
		Map<Short, List<File>> tmpOutputFiles = new HashMap<Short, List<File>>();
		for (Scenario scenario: this.scenarios) {
			// 1a. Add output files
			List<File> scenarioTmpOutputFiles = new ArrayList<File>();
			int i = 0;
			while (i < outputFilePrefix.length) {
				String prefix = outputFilePrefix[i];
				File tmpFile = new File(tmpOutputDir, prefix + "_scenario_" + scenario.getScenarioId() + ".csv");
				if (tmpFile.isFile()) {
					scenarioTmpOutputFiles.add(tmpFile);
				}
				
				i++;
			}
			
			// 1b. Add run info
			Map<RunInfo,String> runInfo = this.scenarioRunInfo.get(scenario.getScenarioId());
			if (runInfo.containsKey(RunInfo.Output)) {
				File scenarioRunInfoFile = new File(tmpOutputDir, "run_info_scenario_" + scenario.getScenarioId() + ".log");
				try {
					PrintWriter writer = new PrintWriter(new FileWriter(scenarioRunInfoFile));
					writer.print(runInfo.get(RunInfo.Output));
					writer.close();
					scenarioTmpOutputFiles.add(scenarioRunInfoFile);
				} catch (IOException ioe) {
					throw new ExecutionException("Error while writer scenario run info file for scenario " + scenario.getScenarioId() + ": " + ioe);
				}
			}
			
			tmpOutputFiles.put(scenario.getScenarioId(), scenarioTmpOutputFiles);
		}
		
		// 2. Create compressed files for each scenario, then for scenario ensemble and move it them to output directory
		File finalOutputDir = new File(outputDir, "final");		
		if (! finalOutputDir.isDirectory()) {
			throw new ExecutionException("Final output directory not found: " + finalOutputDir.getAbsolutePath());
		}
		File ensembleOutputFile = new File(finalOutputDir, this.scenarioEnsembleId + ".rar");
		if (ensembleOutputFile.isFile()) {
			ensembleOutputFile.delete();
		}
		
		List<File> scenarioCompressedFiles = new ArrayList<File>();			
		for (Short sId: tmpOutputFiles.keySet()) {
			File outputFile = new File(tmpOutputDir, sId + ".zip");
			
			// 2.1 Eliminacion del zip existente (si es que habia)
			if (outputFile.isFile()) {
				outputFile.delete();
			}
							
			// 2.2 Creacion del nuevo zip
			List<File> inputFiles = tmpOutputFiles.get(sId);
			try {
				RarCompressor compressor = new RarCompressor();
				compressor.compress(inputFiles, outputFile, true);
				scenarioCompressedFiles.add(outputFile);
			} catch (IOException ioe) {
				throw new FulminantError("Error while creating ZIP output file for scenario " + sId + " of ensemble " + this.scenarioEnsembleId + ": " + ioe);
			}
		}

		// 2.3 Creacion del zip para el ensemble
		try {
			RarCompressor compressor = new RarCompressor();
			compressor.compress(scenarioCompressedFiles, ensembleOutputFile, true);
		} catch (IOException ioe) {
			throw new FulminantError("Error while creating ZIP output file for ensemble " + this.scenarioEnsembleId + ": " + ioe);
		}
		
		return ensembleOutputFile;
	}
	
	public void uploadResults(File packageFile) throws ExecutionException {
		FTPClient client = new FTPClient();
		
		try {
			// 1. Connect to FTP server and set working directory
			client.connect(ftpHost);
			if (! client.login(ftpUserName, ftpPassword)) {
				throw new ExecutionException("Error connecting to " + ftpHost + " using account " + ftpUserName + "/" + ftpPassword);
			}
			client.changeWorkingDirectory(ftpPathName);
			
			// 2. Results upload
			client.setFileType(FTP.BINARY_FILE_TYPE);
			InputStream inputStream = new FileInputStream(packageFile);
			try {
				if (! client.storeFile(packageFile.getName(), inputStream)) {						
					throw new ExecutionException("Error uploading file " + packageFile.getAbsolutePath() + " to " + client.printWorkingDirectory() + ": " + client.getReplyString());
				}
			} finally {
				inputStream.close();
			}
			
			// 3. Logout
			client.logout();
		} catch (IOException ioe) {
			throw new ExecutionException("Error uploading results to FTP server: " + ioe);
		}
	}	
}