package fiuba.lamm.agromodel.execution.ensemble;

import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.execution.ExecutionException;
import fiuba.lamm.agromodel.execution.factories.*;
import java.io.*;
import java.util.*;
import java.util.concurrent.*;
import javax.xml.stream.*;

public class SimulationExecutionHandler implements ExecutionHandler, Callable<Map<Short,Map<RunInfo,String>>> {
	protected Short scenarioEnsembleId;
	protected Queue<Scenario> scenarios;
	protected Integer randomSeed;
	protected Short maxThreads;
	protected Map<Short,Map<RunInfo,String>> results;
	protected CountDownLatch gate;
	
	public SimulationExecutionHandler(Short scenarioEnsembleId, Integer randomSeed, Short maxThreads, Map<Short,Map<RunInfo,String>> results) {
		// 2. Assign variables
		this.scenarioEnsembleId = scenarioEnsembleId;
		this.scenarios = new ConcurrentLinkedQueue<Scenario>();
		this.randomSeed = randomSeed;
		this.maxThreads = maxThreads;
		this.results = results;
		this.gate = new CountDownLatch(maxThreads);		
	}
	
	public void execute() throws ExecutionException {
		// 1. Initialize scenarios
		this.scenarios.addAll(ScenarioFactory.getInstance().findByScenarioEnsembleId(this.scenarioEnsembleId));
		
		// 2. Create threads and launch them
		ExecutorService service = Executors.newFixedThreadPool(this.maxThreads);
		List<Future<Map<Short,Map<RunInfo,String>>>> futureResults = new ArrayList<Future<Map<Short,Map<RunInfo,String>>>>();
		for (int i = 0; i < this.maxThreads.intValue(); i++) {
			futureResults.add(service.submit(this));
		}
		service.shutdown();

		// 3. Wait for all the threads
		try {
			this.gate.await();
		} catch (InterruptedException ie) {
			throw new ExecutionException("Execution interrupted: " + ie.getMessage());			
		}
		
		// 4. Check if any exception was raised
		String error = new String();
		for (Future<Map<Short,Map<RunInfo,String>>> future: futureResults) {
			try {
				Map<Short,Map<RunInfo,String>> threadResults = future.get();
				this.results.putAll(threadResults);
				
				Iterator<Short> sIds = threadResults.keySet().iterator();
				while (sIds.hasNext()) {
					Short sId = sIds.next();					
					Map<RunInfo,String> threadRunInfo = threadResults.get(sId);					
					if (threadRunInfo.containsKey(RunInfo.Errors)) {
						error += "Scenario " + sId + " finished with errors: " + threadRunInfo.get(RunInfo.Errors) + "\n";
					}					
				}
			} catch (java.util.concurrent.ExecutionException ee) {
				throw new ExecutionException("Error fetching execution results: " + ee);
			} catch (InterruptedException ie) {
				throw new ExecutionException("Error fetching execution results: " + ie);
			}
		}
		if (error.length() > 0) {
			throw new ExecutionException("Error executing scenarios:\n" + error);
		}
	}

	public void runScenario(Scenario scenario, File projectRoot, File sweepFile) throws ExecutionException {
		// 1. Create new temporary XML file 
		XMLOutputFactory outputFactory = XMLOutputFactory.newInstance();
		try {
			OutputStream stream = new FileOutputStream(sweepFile);
			XMLStreamWriter writer = outputFactory.createXMLStreamWriter(stream);
			writer.writeStartDocument("1.0");
			writer.writeStartElement("sweep");
			writer.writeAttribute("runs", "1");
			
			// ScenarioId
			writer.writeStartElement("parameter");
			writer.writeAttribute("name", "scenarioId");
			writer.writeAttribute("type", "constant");
			writer.writeAttribute("constant_type", "number");
			writer.writeAttribute("value", scenario.getScenarioId().toString());
			writer.writeEndElement();
			
			// Random Seed
			writer.writeStartElement("parameter");
			writer.writeAttribute("name", "randomSeed");
			writer.writeAttribute("type", "constant");
			writer.writeAttribute("constant_type", "number");
			writer.writeAttribute("value", this.randomSeed.toString());
			writer.writeEndElement();
			
			writer.writeEndElement(); // </sweep>
			writer.writeEndDocument();
			writer.flush();
			writer.close();			
			stream.close();
		} catch (IOException ioe) {
			throw new ExecutionException("Error creating temporary sweep file for scenario " + scenario.getScenarioId() + ": " + ioe);		
		} catch (XMLStreamException xmlse) {
			throw new ExecutionException("Error creating temporary sweep file for scenario " + scenario.getScenarioId() + ": " + xmlse);
		}
		
		// 2. Run simulation in batch mode
		File launchersDir = new File(projectRoot, "launchers");
		if (! launchersDir.isDirectory()) {
			throw new ExecutionException("Execution of scenario " + scenario.getScenarioId() + ". Launchers directory not found");			
		}
		File batchExecutionScript = new File(launchersDir, "batch.sh");
		if (! batchExecutionScript.isFile()) {
			throw new ExecutionException("Execution of scenario " + scenario.getScenarioId() + ". Batch execution script not found");
		}
		
		ProcessBuilder pb = new ProcessBuilder(batchExecutionScript.getAbsolutePath(), sweepFile.getName());
		pb.directory(launchersDir);
		try {
			Process p = pb.start();
			p.waitFor();			
		} catch (InterruptedException ie) {
			throw new ExecutionException("Error executing scenario " + scenario.getScenarioId() + " in batch mode: " + ie);
		} catch (IOException ioe) {
			throw new ExecutionException("Error executing scenario " + scenario.getScenarioId() + " in batch mode: " + ioe);
		}
	}
	
	public Map<Short,Map<RunInfo,String>> call() {		
		Map<Short,Map<RunInfo,String>> results = new HashMap<Short,Map<RunInfo,String>>();
		
		Scenario scenario = this.scenarios.poll();
		while (scenario != null) {
			File sweepFile = null;
			File projectRoot = new File(System.getenv("PROJECT_ROOT"));

			try {
				// 1. Create temporary sweep file
				// TODO: en el futuro, crear el batch programaticamente				
				File batchDir = new File(projectRoot, "batch");
				if (! batchDir.isDirectory()) {
					throw new ExecutionException("Execution of scenario " + scenario.getScenarioId() + ". Batch directory not found");
				}		
				sweepFile = new File(batchDir, "ensemble_" + this.scenarioEnsembleId + "_scenario_" + scenario.getScenarioId() + ".xml");
				if (sweepFile.isFile()) sweepFile.delete();
			
				// 2. Run scenario and finish task
				runScenario(scenario, projectRoot, sweepFile);
			} catch (ExecutionException ee) {
				Map<RunInfo,String> runInfo = new HashMap<RunInfo,String>();
				runInfo.put(RunInfo.Errors, ee.getMessage());
				results.put(scenario.getScenarioId(), runInfo);
			} finally {
				// 3. Delete temporary sweep file
				if (sweepFile != null) {
					sweepFile.delete();
				}			
					
				// 4. Get simulation results
				File outputFile = new File(projectRoot, "ensemble_" + this.scenarioEnsembleId + "_scenario_" + scenario.getScenarioId() + ".xml.out");
				File errorFile = new File(projectRoot, "ensemble_" + this.scenarioEnsembleId + "_scenario_" + scenario.getScenarioId() + ".xml.err");				
				Map<RunInfo,String> runInfo = RunInfo.getInstance(outputFile, errorFile);
				results.put(scenario.getScenarioId(), runInfo);
				outputFile.delete();
				errorFile.delete();
			}
			
			// 5. Get next scenario
			scenario = this.scenarios.poll();
		}
		
		// 6. Finish execution, decrease CountDownLatch and return results
		this.gate.countDown();		
		return results;
	}
}