package fiuba.lamm.agromodel.factories;

import java.sql.*;
import java.util.*;

import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.finance.*;
import fiuba.lamm.agromodel.utils.*;

public class ActivityManagementFactory extends Factory {	
	protected static Map<Scenario,ActivityManagementFactory> factories;	
	protected Map<Short, ActivityManagement> cache;
	protected Map<CroppingCycleRegionCompositeKey, ActivityManagement> mostExpensiveManagementsCache;
	
	public static synchronized ActivityManagementFactory getInstance(Scenario scenario) {
		if (factories == null) {
			factories = new TreeMap<Scenario,ActivityManagementFactory>();			
		}
		if (! factories.containsKey(scenario)) {
			try {
				factories.put(scenario, new ActivityManagementFactory(scenario));
			} catch (SQLException sqle) {
				throw new FulminantError("Error creation factory for scenario " + scenario);
			}
		}
		
		return factories.get(scenario);
	}
	
	protected ActivityManagementFactory(Scenario scenario) throws SQLException {
		super();
		
		this.cache = new TreeMap<Short, ActivityManagement>();
		this.mostExpensiveManagementsCache = new HashMap<CroppingCycleRegionCompositeKey, ActivityManagement>();

		Connection connection = null;
		try {
			connection = getConnection();
			
			String rootQuery = null;
			List<Object> rootParameters = new ArrayList<Object>();
			if (scenario != null) {
				// Busco todos los AMs (excepto ganaderia) permitidos en el escenario que no sean de "segunda".
				if (scenario.hasAttribute(Scenario.Attribute.HasTechnologyInnovations)) {
					// En este caso incluye las innovaciones.
					rootQuery = "select am.*, ams.firstAvailableCroppingCycleId from activitymgmt_scenario ams join activitymgmt_specifications am on (am.activityMgmtId = ams.activityMgmtId and ams.scenarioId = ?) join activity a on (a.activityId = am.activityId and a.activityType = ?) where am.activityMgmtId not in (select nextActivityMgmtId from activitymgmt_specifications where nextActivityMgmtId is not null)";
				} else {
					// En este caso se excluyen las innovaciones.
					rootQuery = "select am.*, ams.firstAvailableCroppingCycleId from activitymgmt_scenario ams join activitymgmt_specifications am on (am.activityMgmtId = ams.activityMgmtId and ams.scenarioId = ? and ams.firstAvailableCroppingCycleId = 1) join activity a on (a.activityId = am.activityId and a.activityType = ?) where am.activityMgmtId not in (select nextActivityMgmtId from activitymgmt_specifications where nextActivityMgmtId is not null)";
				}				
				rootParameters.add(scenario.getScenarioId());
			} else {
				// Busco todos los AMs (excepto ganaderia) que no sean de "segunda".
				rootQuery = "select am.*, null as firstAvailableCroppingCycleId from activitymgmt_specifications am join activity a on(a.activityId = am.activityId and a.activityType = ?) where am.activityMgmtId not in (select nextActivityMgmtId from activitymgmt_specifications where nextActivityMgmtId is not null)";
			}
			
			// TODO: Estos queries solo buscan AMs que sean de Agricultura. En el futuro se deberan incluir aquellos de Ganaderia.
			rootParameters.add("Agriculture");
			
			PreparedStatement stRoot = connection.prepareStatement(rootQuery);
			PreparedStatement stNext = connection.prepareStatement("select * from activitymgmt_specifications where activityMgmtId = ?");
			PreparedStatement stInputs = connection.prepareStatement("select inputTypeId, inputAmount from activitymgmt_inputs where activityMgmtId = ?");
			PreparedStatement stLabors = connection.prepareStatement("select laborTypeId, laborAmount from activitymgmt_labors where activityMgmtId = ?");
			
			InputTypeFactory inputTypeFactory = InputTypeFactory.getInstance();
			LaborTypeFactory laborTypeFactory = LaborTypeFactory.getInstance();
			
			Map<Short,Short> competingManagements = new HashMap<Short,Short>();
			List<Map<String, Short>> successors = new ArrayList<Map<String, Short>>();
			int i = 1;
			for (Object parameterValue: rootParameters) {
				stRoot.setObject(i, parameterValue);
				i++;
			}
			
			ResultSet rs = stRoot.executeQuery();
			while (rs.next()) {
				Short activityMgmtId = rs.getShort("activityMgmtId");
				Activity activity = ActivityFactory.getInstance().findByPrimaryKey(rs.getByte("activityId"));
				Region region = RegionFactory.getInstance().findByPrimaryKey(rs.getByte("regionId"));
				
				Map<InputType, Double> inputs = new HashMap<InputType, Double>();
				stInputs.setShort(1, activityMgmtId);
				ResultSet rsInputs = stInputs.executeQuery();
				while (rsInputs.next()) {
					InputType inputType = inputTypeFactory.findByPrimaryKey(rsInputs.getByte("inputTypeId"));
					inputs.put(inputType, rsInputs.getDouble("inputAmount"));
				}
				rsInputs.close();
				
				Map<LaborType, Integer> labors = new HashMap<LaborType, Integer>();
				stLabors.setShort(1, activityMgmtId);
				ResultSet rsLabors = stLabors.executeQuery();
				while (rsLabors.next()) {
					LaborType laborType = laborTypeFactory.findByPrimaryKey(rsLabors.getByte("laborTypeId"));
					labors.put(laborType, rsLabors.getInt("laborAmount"));
				}
				rsLabors.close();
								
				Short nextActivityMgmtId = rs.getShort("nextActivityMgmtId");
				if (! rs.wasNull()) {
					Map<String, Short> linkTable = new TreeMap<String, Short>();
					linkTable.put("previous", activityMgmtId);
					linkTable.put("next", nextActivityMgmtId);
					successors.add(linkTable);
				}
				
				Short firstAvailableCroppingCycleId = rs.getShort("firstAvailableCroppingCycleId");
				if  (rs.wasNull()) {
					firstAvailableCroppingCycleId = null;
				}
				
				Short competingManagementId = rs.getShort("competingManagementId");
				if (! rs.wasNull()) {
					competingManagements.put(activityMgmtId, competingManagementId);
				}
				
				ActivityManagement am = new ActivityManagement(
					activityMgmtId,
					activity,
					region,
					rs.getString("activityMgmtName"),
					rs.getString("plantingDate"),
					rs.getString("genotype"),
					rs.getInt("nFertilizerDose"),
					rs.getDouble("rowSpacing"),
					rs.getDouble("plantingDensity"),
					inputs,
					labors,
					firstAvailableCroppingCycleId
				);
				
				cache.put(activityMgmtId, am);
			}
			rs.close();
			stRoot.close();
			
			// Ahora busco los sucesores
			while (! successors.isEmpty()) {
				Map<String, Short> linkTable = successors.remove(0);
				Short prevActivityMgmtId = linkTable.get("previous");
				Short activityMgmtId = linkTable.get("next");
				
				stNext.setShort(1, activityMgmtId);
				ResultSet rsNext = stNext.executeQuery();
				if (rsNext.next()) {
					Activity activity = ActivityFactory.getInstance().findByPrimaryKey(rsNext.getByte("activityId"));
					ActivityManagement previousActivityManagement = cache.get(prevActivityMgmtId);
					Region region = RegionFactory.getInstance().findByPrimaryKey(rsNext.getByte("regionId"));
					
					Map<InputType, Double> inputs = new HashMap<InputType, Double>();
					stInputs.setShort(1, activityMgmtId);
					ResultSet rsInputs = stInputs.executeQuery();
					while (rsInputs.next()) {
						InputType inputType = inputTypeFactory.findByPrimaryKey(rsInputs.getByte("inputTypeId"));
						inputs.put(inputType, rsInputs.getDouble("inputAmount"));
					}
					rsInputs.close();
					
					Map<LaborType, Integer> labors = new HashMap<LaborType, Integer>();
					stLabors.setShort(1, activityMgmtId);
					ResultSet rsLabors = stLabors.executeQuery();
					while (rsLabors.next()) {
						LaborType laborType = laborTypeFactory.findByPrimaryKey(rsLabors.getByte("laborTypeId"));
						labors.put(laborType, rsLabors.getInt("laborAmount"));
					}
					rsLabors.close();
									
					Short nextActivityMgmtId = rsNext.getShort("nextActivityMgmtId");
					if (! rsNext.wasNull()) {
						Map<String, Short> newLinkTable = new TreeMap<String, Short>();
						newLinkTable.put("previous", activityMgmtId);
						newLinkTable.put("next", nextActivityMgmtId);
						successors.add(newLinkTable);
					}
					
					ActivityManagement am = new ActivityManagement(
						activityMgmtId,
						activity,
						region,
						rsNext.getString("activityMgmtName"),
						rsNext.getString("plantingDate"),
						rsNext.getString("genotype"),
						rsNext.getInt("nFertilizerDose"),
						rsNext.getDouble("rowSpacing"),
						rsNext.getDouble("plantingDensity"),
						inputs,
						labors,
						previousActivityManagement.getFirstAvailableCroppingCycleId()
					);
					
					ActivityManagement previousManagement = cache.get(prevActivityMgmtId);
					previousManagement.setNextActivityManagement(am);
				}				
				rsNext.close();
			}
			
			// Manejos competitivos
			List<ActivityManagement> managements = new ArrayList<ActivityManagement>(cache.values());
			for (ActivityManagement management: managements) {
				if (competingManagements.containsKey(management.getActivityMgmtId())) {
					ActivityManagement competingManagement = cache.get(competingManagements.get(management.getActivityMgmtId()));
					management.setCompetingManagement(competingManagement);
				}
			}
			
			stNext.close();
			stInputs.close();
			stLabors.close();
		} catch (SQLException sqle) {
			throw new FulminantError("Error buscar las actividades/manejos: " + sqle);			
		} finally {
			if (connection != null) {
				try {
					connection.close();
				} catch (SQLException sqle) {
					throw new FulminantError("Error buscar las actividades/manejos: " + sqle);
				}
			}
		}
	}
	
	public ActivityManagement findByPrimaryKey(Short activityMgmtId) {
		return cache.get(activityMgmtId);
	}	
	
	public List<ActivityManagement> findAll() {
		return new ArrayList<ActivityManagement>(cache.values());
	}
	
	public List<ActivityManagement> findInnovationsByRegion(Region region) {
		List<ActivityManagement> managements = findByRegion(region);
		List<ActivityManagement> innovations = new ArrayList<ActivityManagement>();
		for (ActivityManagement management: managements) {
			if (management.isInnovation()) {
				innovations.add(management);
			}
		}
		
		return innovations;
	}
	
	public List<ActivityManagement> findByRegion(Region region) {
		List<ActivityManagement> managements = new ArrayList<ActivityManagement>();
		List<ActivityManagement> all = findAll();
		for (int i = 0; i < all.size(); i++) {
			ActivityManagement am = all.get(i);
			if (am.getRegion().equals(region)) managements.add(am);
		}
		
		return managements;
	}
	
	public ActivityManagement findMostExpensive(CroppingCycle croppingCycle, Region region) {
		CroppingCycleRegionCompositeKey key = new CroppingCycleRegionCompositeKey(croppingCycle, region);
		ActivityManagement am = mostExpensiveManagementsCache.get(key);
		if (am == null) {
			// Busco el cultivo-manejo mas caro para esa temporada y region
			double mostExpensiveCost = 0.0;
			List<ActivityManagement> managements = findByRegion(region);
			for (int i = 0; i < managements.size(); i++) {
				/*
				 * El costo de implantacion para el calculo del cultivo-manejo mas caro
				 * surje de la suma de los siguientes costos:
				 * 
				 * 1. Costos de labores
				 * 2. Costos de insumos
				 * 
				 */
				ActivityManagement activityManagement = managements.get(i);
				ActivityManagement management = activityManagement;
				double implantationCost = 0.0;
				while (management != null) {
					implantationCost += management.getDirectCost(DirectCost.FixedCost);
					management = management.getNextActivityManagement();
				}
				
				if (implantationCost > mostExpensiveCost) {
					am = activityManagement;
					mostExpensiveCost = implantationCost;
				}				
			}
			
			mostExpensiveManagementsCache.put(key, am);
		}
		
		return am;
	}
}