package fiuba.lamm.agromodel.factories;

import java.sql.*;
import java.util.*;
import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.agents.Manager;
import fiuba.lamm.agromodel.decision.*;
import fiuba.lamm.agromodel.decision.factories.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.finance.*;

public class CroppingCycleFactory extends Factory {
	protected String sqlFindByPrimaryKey;
	protected String sqlFindYields;
	protected String sqlFindInputPrices;
	protected String sqlFindActivityData;
	protected String sqlFindSoilTypeData;
	protected String sqlFindRegionData;
	protected String sqlFindAdjustmentIndices;
	protected String sqlFindContextStatusData;	
	protected Scenario scenario;
	protected static Map<Scenario,CroppingCycleFactory> factories;
	
	public static synchronized CroppingCycleFactory getInstance(Scenario scenario) {
		if (factories == null) {
			factories = new TreeMap<Scenario,CroppingCycleFactory>();			
		}
		if (! factories.containsKey(scenario)) {
			factories.put(scenario, new CroppingCycleFactory(scenario));
		}
		
		return factories.get(scenario);
	}
	
	protected CroppingCycleFactory(Scenario scenario) {
		super();
		this.scenario = scenario;
		this.sqlFindByPrimaryKey = "select * from croppingcycle where croppingCycleId = ? and scenarioId = ?";
		this.sqlFindYields = "select soilTypeId, yield from yieldlookup where croppingCycleId = ? and activityMgmtId = ? and scenarioId = ?";
		this.sqlFindInputPrices = "select inputTypeId, inputUnitPrice from croppingcycle_inputprices where croppingCycleId = ? and scenarioId = ?";
		this.sqlFindActivityData = "select outputPrice from croppingcycle_activity where croppingCycleId = ? and activityId = ? and scenarioId = ?";
		this.sqlFindSoilTypeData = "select landSalePrice, rentalPrice, expectedGrossMargin, farmTaxes, organicMatter, waterStorageCapacity, soilCurveNumber from croppingcycle_soiltype where croppingCycleId = ? and soilTypeId = ? and scenarioId = ?";
		this.sqlFindRegionData = "select transportationCost from croppingcycle_region where croppingCycleId = ? and regionId = ? and scenarioId = ?";
		this.sqlFindAdjustmentIndices = "select yieldAdjustmentIndex, directCostAdjustmentIndex from croppingcycle_historicaladjustment where croppingCycleId = ? and activityMgmtId = ? and scenarioId = ?";
		this.sqlFindContextStatusData = "select expectedStatusId, actualStatusId from croppingcycle_contexts where croppingCycleId = ? and scenarioId = ? and soilTypeId = ? and contextTypeId = ?";
	}
	
	public static void clearScenarioCache(Scenario scenario) {
		factories.remove(scenario);
	}

	public List<CroppingCycle> findAll() {
		throw new UnsupportedOperationException("CroppingCycleFactory does not support this method");
	}
	
	public CroppingCycle findFirst() throws SQLException {
		return findByPrimaryKey(new Short((short) 1), false);		
	}
	
	public CroppingCycle findNext(CroppingCycle currentCroppingCycle) throws SQLException {
		return findByPrimaryKey(new Short((short) (currentCroppingCycle.getCroppingCycleId().intValue() + 1)), currentCroppingCycle.isActive());
	}
		
	public CroppingCycle findByPrimaryKey(Short croppingCycleId, boolean currentCroppingCycleIsActive) throws SQLException {
		CroppingCycle croppingCycle = null;
		Connection connection = getConnection();
		
		PreparedStatement st = connection.prepareStatement(sqlFindByPrimaryKey);
		st.setShort(1, croppingCycleId);
		st.setShort(2, this.scenario.getScenarioId());
		ResultSet rs = st.executeQuery();
				
		if (rs.next()) {
			croppingCycle = new CroppingCycle(
					rs.getShort("croppingCycleId"),
					rs.getString("croppingCycleName"),
					rs.getDouble("managementCost"),
					rs.getDouble("structuralCost"),
					rs.getDouble("utaPrice"),
					new Byte((byte) 1).equals(rs.getByte("active"))
			);
			
			// Check if cropping cycle is the first avtive one
			if (! currentCroppingCycleIsActive && croppingCycle.isActive()) {
				croppingCycle.setIsFirstActive(true);
			}
			
			// Update cropping cycle dependencies
			updateCroppingCycleDependencies(croppingCycle);			
		}
		rs.close();
		st.close();
		connection.close();
				
		return croppingCycle;
	}
	
	protected void updateCroppingCycleDependencies(CroppingCycle croppingCycle) {
		Scenario scenario = Manager.getInstance().getScenario();
		
		try {		
			// 1. Actualizacion de precios y contextos de activities.
			//    ESTO DEBE HACERCE ANTES QUE LA ACTUALIZACION DE PRECIOS DE ALQUILER, YA
			//    QUE LOS PRECIOS DE ALQUILER DEPENDEN DEL PRECIO DE LA SOJA
			List<Activity> activitys = ActivityFactory.getInstance().findAll();
			updateActivitys(croppingCycle, activitys);
			
			// 2. Actualizacion de costos de labores
			Iterator<LaborType> labors = LaborTypeFactory.getInstance().findAll().iterator();
			while (labors.hasNext()) {
				LaborType labor = labors.next();
				labor.setUtaPrice(croppingCycle.getUtaPrice());
			}
			
			// 3. Actualizacion de costos de insumos
			Map<Byte, Double> inputPrices = getInputTypePrices(croppingCycle);
			Iterator<InputType> inputs = InputTypeFactory.getInstance().findAll().iterator();
			while (inputs.hasNext()) {
				InputType input = inputs.next();
				if (inputPrices.containsKey(input.getInputTypeId())) { 
					input.setInputTypePrice(inputPrices.get(input.getInputTypeId()));
				}
			}
			
			// 4. Actualizo los datos de costos para los cultivos-manejo
			List<ActivityManagement> activityManagements = ActivityManagementFactory.getInstance(scenario).findAll();
			for (int i = 0; i < activityManagements.size(); i++) {
				ActivityManagement activityManagement = activityManagements.get(i);
				
				while (activityManagement != null) {
					Map<DirectCost, Double> costsInfo = new HashMap<DirectCost, Double>();
					
					/* 4a. Costos fijos */
					// Costos de mantenimiento (expresados en U$S/ha)
					double fixedCosts = 0.0;
					Map<InputType, Double> inputAmounts = activityManagement.getNeededInputs();
					Map<LaborType, Integer> laborAmounts = activityManagement.getNeededLabors();
					
					Iterator<InputType> neededInputs = inputAmounts.keySet().iterator();
					while (neededInputs.hasNext()) {
						InputType input = neededInputs.next();
						Double amount = inputAmounts.get(input);
						fixedCosts += input.getInputTypePrice() * amount;
					}
					
					Iterator<LaborType> neededLabors = laborAmounts.keySet().iterator();
					while (neededLabors.hasNext()) {
						LaborType labor = neededLabors.next();
						Integer amount = laborAmounts.get(labor);
						fixedCosts += labor.getCostPerUnit() * amount.doubleValue();
					}
					costsInfo.put(DirectCost.FixedCost, fixedCosts);
					
					/* 4b. Costos variables */
					// Costos de cosecha y comercializacion (es un valor porcentual aplicable al IB del cultivo-manejo
					Activity activity = activityManagement.getActivity();
					double marketingAndHarvestingCost = activity.getHarvestCostRate() + activity.getMarketingCostRate();
					costsInfo.put(DirectCost.MarketingAndHarvestingCost, marketingAndHarvestingCost);
					
					activityManagement.setCroppingCycleCosts(croppingCycle.getCroppingCycleId(), costsInfo);
					
					/* 4c. Actualizacion de rendimientos */
					setYields(croppingCycle, activityManagement);
					
					/* 4d. Actualizacion de indices */
					setAdjustmentIndices(croppingCycle, activityManagement);
					
					/* 4e. Seteo la disponibilidad del cultivo-manejo */
					
					// Me fijo si tiene un cultivo-manejo que le suceda
					activityManagement = activityManagement.getNextActivityManagement();
				}
			}
			
			// 5. Actualizacion de costos de administracion
			IndirectCost.ManagementCost.setBasicCost(croppingCycle.getManagementCost());
			IndirectCost.StructuralCost.setBasicCost(croppingCycle.getStructuralCost());
				
			// 6. Actualizacion de valores asociados a tipos de suelo
			List<SoilType> soilTypes = SoilTypeFactory.getInstance(scenario).findAll();
			updateSoilTypes(croppingCycle, soilTypes);
			
			// 7. Actualizacion de valores asociados a regiones
			List<Region> regions = RegionFactory.getInstance().findAll();
			updateRegions(croppingCycle, regions);
			
			// 8. Actualizacion de contextos
			List<ContextType> contextTypes = ContextTypeFactory.getInstance(scenario).findAll();
			updateContextStatus(croppingCycle, soilTypes, contextTypes);
		} catch (SQLException sqle) {
			throw new FulminantError("Error updating cropping cycle dependencies: " + sqle);
		}
	}

	// Actualizacion de indices de ajuste
	public void setAdjustmentIndices(CroppingCycle croppingCycle, ActivityManagement activityManagement) throws SQLException {
		Connection connection = getConnection();
		PreparedStatement st = connection.prepareStatement(sqlFindAdjustmentIndices);
		st.setShort(1, croppingCycle.getCroppingCycleId());
		st.setShort(2, activityManagement.getActivityMgmtId());
		st.setShort(3, scenario.getScenarioId());
		
		Double yieldAdjustmentIndex = 1.0;
		Double directCostAdjustmentIndex = 1.0;
		ResultSet rs = st.executeQuery();
		if (rs.next()) {
			yieldAdjustmentIndex = rs.getDouble("yieldAdjustmentIndex");
			directCostAdjustmentIndex = rs.getDouble("directCostAdjustmentIndex");
		}
		
		rs.close();
		st.close();
		connection.close();

		activityManagement.setYieldAdjustmentIndex(croppingCycle.getCroppingCycleId(), yieldAdjustmentIndex);
		activityManagement.setDirectCostAdjustmentIndex(directCostAdjustmentIndex);
	}
	
	// Actualizacion de rendimientos
	public void setYields(CroppingCycle croppingCycle, ActivityManagement activityManagement) throws SQLException {
		Map<Byte, Double> yields = new HashMap<Byte, Double>();
		Connection connection = getConnection();
		PreparedStatement st = connection.prepareStatement(sqlFindYields);
		st.setShort(1, croppingCycle.getCroppingCycleId());
		st.setShort(2, activityManagement.getActivityMgmtId());
		st.setShort(3, scenario.getScenarioId());
		
		ResultSet rs = st.executeQuery();
		while (rs.next()) {
			Byte soilTypeId = rs.getByte(1);
			Double yield = rs.getDouble(2);
			
			yields.put(soilTypeId, yield);
		}
		rs.close();
		st.close();
		connection.close();

		activityManagement.setCroppingCycleYields(croppingCycle.getCroppingCycleId(), yields);
	}
	
	// Actualizacion de precios de insumos
	public Map<Byte, Double> getInputTypePrices(CroppingCycle croppingCycle) throws SQLException {
		Map<Byte, Double> inputPrices = new HashMap<Byte, Double>();
		Connection connection = getConnection();
		PreparedStatement st = connection.prepareStatement(sqlFindInputPrices);
		
		st.setShort(1, croppingCycle.getCroppingCycleId());
		st.setShort(2, scenario.getScenarioId());
		ResultSet rs = st.executeQuery();
		while (rs.next()) {
			inputPrices.put(rs.getByte(1), rs.getDouble(2));
		}
		
		rs.close();
		st.close();
		connection.close();
		
		return inputPrices;
	}
	
	/* Actualizacion de precios y contextos de los activities */
	public void updateActivitys(CroppingCycle croppingCycle, List<Activity> activitys) throws SQLException {
		Connection connection = getConnection();
		PreparedStatement st = connection.prepareStatement(this.sqlFindActivityData);
		
		st.setShort(1, croppingCycle.getCroppingCycleId());
		for (int i = 0; i < activitys.size(); i++) {
			Activity activity = activitys.get(i);
			st.setByte(2, activity.getActivityId());
			st.setShort(3, scenario.getScenarioId());
			
			ResultSet rs = st.executeQuery();
			if (rs.next()) {
				activity.setOutputPrice(croppingCycle.getCroppingCycleId(), rs.getDouble("outputPrice"));
			}
			rs.close();
		}
		
		st.close();
		connection.close();
	}
	
	// Actualizacion de tipos de suelos
	public void updateSoilTypes(CroppingCycle croppingCycle, List<SoilType> soilTypes) throws SQLException {
		Connection connection = getConnection();
		PreparedStatement st = connection.prepareStatement(this.sqlFindSoilTypeData);
		
		st.setShort(1, croppingCycle.getCroppingCycleId());
		for (int i = 0; i < soilTypes.size(); i++) {
			SoilType soilType = soilTypes.get(i);
			st.setByte(2, soilType.getSoilTypeId());
			st.setShort(3, scenario.getScenarioId());
			
			ResultSet rs = st.executeQuery();
			if (rs.next()) {
				if (croppingCycle.isFirstActive()) {
					// El precio de alquiler de un tipo de suelo viene expresado como
					// toneladas de soja por hectara. Para obtener el precio por hectarea, hay que multiplicar
					// por el precio de la soja
					Activity soyBean = ActivityFactory.getInstance().findByPrimaryKey(Activity.SOYBEAN);
					soilType.update(rs.getDouble("landSalePrice"), soyBean.getOutputPrice(croppingCycle.getCroppingCycleId()) * rs.getDouble("rentalPrice"), rs.getDouble("farmTaxes"), rs.getDouble("organicMatter"), rs.getDouble("waterStorageCapacity"), rs.getDouble("soilCurveNumber"));
				} else {
					soilType.update(rs.getDouble("landSalePrice"), rs.getDouble("farmTaxes"), rs.getDouble("organicMatter"), rs.getDouble("waterStorageCapacity"), rs.getDouble("soilCurveNumber"));
				}				
			}
			rs.close();
		}
		
		st.close();
		connection.close();
	}
	
	// Actualizacion de variables de regiones
	public void updateRegions(CroppingCycle croppingCycle, List<Region> regions) throws SQLException {
		Connection connection = getConnection();
		PreparedStatement st = connection.prepareStatement(this.sqlFindRegionData);
		
		st.setShort(1, croppingCycle.getCroppingCycleId());
		for (int i = 0; i < regions.size(); i++) {
			Region region = regions.get(i);
			st.setByte(2, region.getRegionId());
			st.setShort(3, scenario.getScenarioId());
			
			ResultSet rs = st.executeQuery();
			if (rs.next()) {
				region.setTransportationCost(rs.getDouble("transportationCost"));
			}
			
			rs.close();
		}
		
		st.close();
		connection.close();
	}	
	
	/* Actualizacion de contextos */
	public void updateContextStatus(CroppingCycle croppingCycle, List<SoilType> soilTypes, List<ContextType> contextTypes) throws SQLException {
		Connection connection = getConnection();
		ContextTypeStatusFactory contextTypeStatusFactory = ContextTypeStatusFactory.getInstance();
		PreparedStatement st = connection.prepareStatement(this.sqlFindContextStatusData);
		
		st.setShort(1, croppingCycle.getCroppingCycleId());
		st.setShort(2, scenario.getScenarioId());
		for (SoilType soilType: soilTypes) {
			for (ContextType contextType: contextTypes) {		
				st.setByte(3, soilType.getSoilTypeId());
				st.setShort(4, contextType.getContextTypeId());
				
				ResultSet rs = st.executeQuery();
				if (rs.next()) {
					ContextTypeStatus expectedContextStatus = contextTypeStatusFactory.findByPrimaryKey(rs.getShort("expectedStatusId")); 
					croppingCycle.setExpectedContextTypeStatus(soilType, contextType, expectedContextStatus);
					
					ContextTypeStatus actualContext = contextTypeStatusFactory.findByPrimaryKey(rs.getShort("actualStatusId"));
					croppingCycle.setActualContextTypeStatus(soilType, contextType, actualContext);
				} else {
					throw new FulminantError("No hay valores de contexto esperado/real para scenarioId = " + scenario.getScenarioId() + "; cropping cycle = " + croppingCycle.getCroppingCycleName() + "; soil type = " + soilType.getSoilTypeName() + "; contexto = " + contextType.getName());
				}
				rs.close();
			}
		}
		
		st.close();
		connection.close();
	}
}