package fiuba.lamm.agromodel.finance;

import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.execution.*;

import java.util.*;

/**
 * Esta clase hace las veces de con "contador" que calcula ciertos valores como por ejemplo:
 * 
 * - Global gross margin de un campo
 * - Rental margin de un campo
 * - Impuesto a las ganancias
 * 
 * @author santiago
 *
 */
public class Accountant {
	protected enum IncomeTaxCategory {
		Category1	(0.0, 3175.0, 0.0, 0.09),
		Category2	(3175.0, 6349.0, 286.0, 0.14),
		Category3	(6349.0, 9524.0, 730.0, 0.19),
		Category4	(9524.0, 19048.0, 1333.0, 0.23),
		Category5	(19048.0, 28571.0, 3524.0, 0.27),
		Category6	(28571.0, 38095.0, 6095.0, 0.31),
		Category7	(38095.0, Double.MAX_VALUE, 9048.0, 0.35);
		
		private Double minimumIncome;
		private Double maximumIncome;
		private Double fixedTaxAmount;
		private Double taxRate;
		
		private IncomeTaxCategory(Double minimumIncome, Double maximumIncome, Double fixedTaxAmount, Double fixedTaxRate) {
			this.minimumIncome = minimumIncome;
			this.maximumIncome = maximumIncome;
			this.fixedTaxAmount = fixedTaxAmount;
			this.taxRate = fixedTaxRate;
		}
		
		public Double getMinimumIncome() { 
			return this.minimumIncome;
		}
		
		public Double getMaximumIncome() {
			return this.maximumIncome;
		}
		
		public Double getFixedTaxAmount() {
			return this.fixedTaxAmount;
		}
		
		public Double getFixedTaxRate() {
			return this.taxRate;
		}
	};	
	
	public static Double getGlobalGrossMargin(List<Plot> plots, Double operatorManagerialAbility, Double fixedCostScaleFactor, Short croppingCycleId, Short technologyCroppingCycleId) {
		double profits = 0.0;
		double costs = 0.0;
		
		// 1. Calculate gross margin for each plot in the farm
		Region region = plots.get(0).getSoilType().getRegion(); // All the plots MUST belong to the same region. It is a constraint of the model.
		for (int i = 0; i < plots.size(); i++) {
			Plot plot = plots.get(i);
			// Find out what Activity/Management was assigned to each plot in the farm
			ActivityManagement activityManagement = plot.getCurrentActivityManagement();
			
			/* We calculate the gross margin for the Activity/Management
			 * in each farm plot during a given cropping cycle.
			 */ 
			while (activityManagement != null) {
				// 1a. Gross income calculation (physical yield times product price)
				double yield = 0.0;
				try {
					yield = activityManagement.getCroppingCycleYield(croppingCycleId, technologyCroppingCycleId, plot.getSoilType().getSoilTypeId()) / 1000.0; // Se expresa en ton/ha
				} catch (NullPointerException npe) {
					// No hay rendimiento para ese tipo de suelo. Tiro un error.
					throw new FulminantError("No hay rendimiento para el cropping cycle " + croppingCycleId + ", activity-management " + activityManagement + " y tipo de suelo " + plot.getSoilType());
				}
				double activityManagementProfit = yield * plot.getPlotSize() * activityManagement.getActivity().getOutputPrice(croppingCycleId);
								
				/* 1b. Direct costs calculation.
				 * This item includes costs associated with a specific activity/management.
				 * Direct costs include fixed and variable costs.
				 * Fixed costs are NOT a function of yield or output level;
				 * Variable costs are a function of yield or output level.
				 */
				// FIXED production costs
				double fixedCost = activityManagement.getDirectCost(DirectCost.FixedCost) * plot.getPlotSize();
				/* VARIABLE production costs include three items:
				 * 1. Harvesting costs
				 * 2. Marketing costs (1 and 2 are grouped,as they are both a percentage of gross income)
				 * 3. Transportation costs (function of amount of product to transport, distance to shipping port
				 *    and a transportation cost factor (cost per ton per km).
				 */
				double marketingAndHarvestingCost = activityManagement.getDirectCost(DirectCost.MarketingAndHarvestingCost) * activityManagementProfit;
				double transportationCost = region.getTransportationCost() * yield * plot.getPlotSize() * region.getDistanceToPort();
				
				/* 1c. A "managerial ability" factor multiplies direct costs.
				 * A managerial ability factor ranging between 0.95 and 1.05
				 * is assigned to a given farmer at the time of initialization.
				 * When the factor is < 1.0, an agent has higher managerial ability and his
				 * production costs are lower. Conversely, a factor > 1.0 implies lower managerial ability
				 * and higher costs.
				 * For example, a farmer's higher managerial ability may result in an ability to purchase inputs
				 * at a slightly discounted price, or sell products with lower marketing fees.
				 */
				double activityManagementCosts = operatorManagerialAbility * (fixedCost * fixedCostScaleFactor + marketingAndHarvestingCost + transportationCost); 
				
				// 1d. Add up gross income and direct costs for all Activity/Managements
				profits += activityManagementProfit;		// gross income
				costs += activityManagementCosts;			// direct production costs
			
				// 1e. Do the calculations for the next Activity/Management
				activityManagement = activityManagement.getNextActivityManagement();
			}
		}
		
		return (profits - costs);	
	}
	
	public static Double getRentalMargin(Scenario scenario, Farm farm) {
		double profit = 0.0;
		double cost = 0.0;
		
		// If owner is not the operator, then he has an income from leasing it
		if (! farm.getOwner().equals(farm.getOperator())) {
			// 1. Leasing income if the farm is operated
			if (farm.getOperator() != null) {
				// The profit is equal to the leasing cost(paid by the tenant)
				profit = IndirectCost.LeasingCost.getCost(scenario, null, farm, true);
			}
		
			// 2. Calculate indirect production costs
			cost = IndirectCost.getTotalIndirectCosts(scenario, farm.getOwner(), farm);			
		}
		
		return (profit - cost);
	}
	
	public static Double getIncomeTax(Double income) {
		double incomeTax = 0.0;
		
		int i = 0;
		IncomeTaxCategory categories[] = IncomeTaxCategory.values();
		while ((incomeTax == 0.0) && (i < categories.length)) {
			IncomeTaxCategory category = categories[i];
			if ((income > category.getMinimumIncome()) && (income <= category.getMaximumIncome())) {
				// Paga el impuesto indicado por esta alicuota							
				incomeTax = category.getFixedTaxAmount() + category.getFixedTaxRate() * (income - category.getMinimumIncome());
			}
			
			i++;
		}
		
		return incomeTax;
	}
}
