package fiuba.lamm.agromodel.finance;

import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.math.*;
import java.util.*;

public enum IndirectCost {
	// Definimos los componentes de costos indirectos.
	// Los valores del "costo basico" o valor de referencia se inicializan en 0 porque despues son leidos de la base de datos
	StructuralCost,
	TaxCost,
	ManagementCost,
	LeasingCost,
	AmortizationCost;
	
	private double basicCost; // Costo para la unidad minima de area
	private IndirectCost() {
		this.basicCost = 0.0;
	}
	
	public void setBasicCost(double basicCost) {
		this.basicCost = basicCost;
	}
	
	/**
	 * @param farmer
	 * @param farm
	 * @return The indirect cost of the farm for the farmer
	 */
	public double getCost(Scenario scenario, Farmer farmer, Farm farm, boolean ownFarmOnly) {
		Double calculatedCost = null;
		if (this.equals(IndirectCost.StructuralCost) || this.equals(IndirectCost.ManagementCost)) {
			double totalStructureSize = 0.0; // Area relevante para el calculo de economia de escala
			double productionUnit = 1500.0; // Tamano de unidad independiente de produccion
			
			Double scaleFactor = 1.0;
			if (this.equals(IndirectCost.StructuralCost)) {
				if (scenario.hasAttribute(Scenario.Attribute.HasScaleEconomics)) {
					// Total structure size calculation
					List<Farm> farms = farmer.getOwnedFarms(); // Campos que posee  
					for(int i = 0; i < farms.size(); i++) {
						Farm ownFarm = farms.get(i);
						totalStructureSize += ownFarm.getFarmSize(); // Area de campos que posee
					}
					
					// Scale function creation. Para costos de infraestructura
					Function scaleFunction = new RationalFunction(
						new double[] { productionUnit, 0.35 },
						new double[] { productionUnit, 1.0 }
					);
					
					scaleFactor = scaleFunction.apply(totalStructureSize);
				}				
			} else if (this.equals(IndirectCost.ManagementCost) && scenario.hasAttribute(Scenario.Attribute.HasScaleEconomics)) {
				double maxStructureSize = 10500.0; // A partir de esta cantidad de hectareas el costo es 0.5

				// Total structure size calculation
				List<Farm> farms = ownFarmOnly ? farmer.getOwnedFarms() : farmer.getOperatedFarms(); // Campos que opera
				for(int i = 0; i < farms.size(); i++) {
					Farm operatedFarm = farms.get(i);
					totalStructureSize += operatedFarm.getFarmSize(); // Area total operada
				}
				
				if (totalStructureSize >= maxStructureSize) {
					scaleFactor = 0.5;
				} else {
					double dy = 0.075;
					double ds = 0.025;
					double s = 0.2;
					double r = Math.floor(totalStructureSize / productionUnit);
					
					double x0 = r * productionUnit;
					double x1 = x0 + productionUnit;
					double y0 = 1.0 - r * dy;
					double y1 = y0 - (s - r * ds);
					
					Function scaleFunction = new LinearFunction(
							new double[] { x0, y0 },
							new double[] { x1, y1 }
					);
					
					scaleFactor = scaleFunction.apply(totalStructureSize);
				}
			}
			
			// Resultado final del calculo de costo de estructura
			calculatedCost = (this.basicCost * scaleFactor * farm.getFarmSize());
		} else if (this.equals(IndirectCost.LeasingCost)) {
			double leasingCost = 0.0;
			List<Plot> plots = farm.getPlots();
			for (int i = 0; i < plots.size(); i++) {
				Plot plot = plots.get(i);
				leasingCost += plot.getSoilType().getRentalPrice() * plot.getPlotSize();
			}
			
			calculatedCost = leasingCost;
		} else if (this.equals(IndirectCost.TaxCost)) {
			// El costo basico se obtiene de la base de datos
			// Hay un costo por hectara por cada tipo de suelo
			// Voy sumando en taxCost el costo particular de cada plot segun su tipo de suelo
			double taxCost = 0.0;
			List<Plot> plots = farm.getPlots();
			for (int i = 0; i < plots.size(); i++) {
				Plot plot = plots.get(i);
				taxCost += plot.getSoilType().getFarmTaxes() * plot.getPlotSize();	
			}
			
			calculatedCost = taxCost;
		} else if (this.equals(IndirectCost.AmortizationCost)){
			// TODO: Funcionalidad aun no implementada
			calculatedCost = 0.0;
		} else {
			throw new FulminantError("Unknown indirect cost: " + this.toString());
		}
		
		return calculatedCost;
	}
	
	public static double getTotalIndirectCosts(Scenario scenario, Farmer farmer, Farm farm) {
		return getTotalIndirectCosts(scenario, farmer, farm, false);
	}
	
	/*
	 * Este metodo calcula los costos indirectos totales asumiendo que el farmer opera sus propios campos
	 * y SOLO esos campos
	 */
	public static double getTotalIndirectCostsForOwnFarm(Scenario scenario, Farmer farmer, Farm farm) {
		return getTotalIndirectCosts(scenario, farmer, farm, true);
	}
	
	private static double getTotalIndirectCosts(Scenario scenario, Farmer farmer, Farm farm, boolean ownFarmOnly) {
		double totalCost = 0.0;
		if (farm.getOwner().equals(farmer)) {
			// The farmer is the owner, so he has structural, tax and amortization costs
			totalCost += IndirectCost.StructuralCost.getCost(scenario, farmer, farm, ownFarmOnly);
			totalCost += IndirectCost.TaxCost.getCost(scenario,farmer, farm, ownFarmOnly);
			totalCost += IndirectCost.AmortizationCost.getCost(scenario, farmer, farm, ownFarmOnly);
		} else {
			// The farmer is a tenant so he has leasing costs
			totalCost += IndirectCost.LeasingCost.getCost(scenario, farmer, farm, ownFarmOnly);
		}
		
		Farmer operator = farm.getOperator();
		// Si el campo no tiene operator se asume que el owner hace el calculo con la intencion
		// de ver el costo que tendria operar el mismo
		if (ownFarmOnly || (operator == null) || farmer.equals(operator)) {
			// The farmer is the operator, so he has management costs
			// If it has no operator the calculation must proceed as if (farmer == operator)
			totalCost += IndirectCost.ManagementCost.getCost(scenario, farmer, farm, ownFarmOnly);
		}
					
		return totalCost;
	}
}
