package fiuba.lamm.agromodel.networks;

import com.vividsolutions.jts.geom.*;
import fiuba.lamm.agromodel.agents.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.utils.*;
import java.io.*;
import java.net.*;
import java.util.*;
import org.geotools.data.*;
import org.geotools.data.shapefile.*;
import org.geotools.feature.*;
import repast.simphony.context.*;
import repast.simphony.context.space.gis.*;
import repast.simphony.context.space.graph.*;
import repast.simphony.query.space.gis.*;
import repast.simphony.space.gis.*;
import repast.simphony.space.graph.*;

public class GISNetworkModel extends FarmNetworkModel {
	protected Double maxNeighbourDistance; // In metres 
	protected File shapeFile;
	protected Geography<Agent> farmingRegions;
	protected Network<Agent> neighbourhood;
	
	public GISNetworkModel(Double maxNeighbourDistance, File shapeFile) {
		this.maxNeighbourDistance = maxNeighbourDistance;
		this.shapeFile = shapeFile;
		this.farmingRegions = null;
		this.neighbourhood = null;
	}
	
	public Network<Agent> createNetwork(Context<Agent> context,	List<Farm> farms) {
		NetworkFactory networkFactory = NetworkFactoryFinder.createNetworkFactory(new HashMap<String, Object>());
		this.neighbourhood = networkFactory.createNetwork("Neighborhood", context, false);
		
		GeographyFactory geographyFactory = GeographyFactoryFinder.createGeographyFactory(new HashMap<String, Object>());		
		this.farmingRegions = geographyFactory.createGeography("FarmingRegions", context, new GeographyParameters<Agent>());
		
		if (shapeFile.isFile()) {
			FeatureIterator featureIterator = null;
			try {
				URL url = shapeFile.toURI().toURL();
				ShapefileDataStore store = new ShapefileDataStore(url);
				featureIterator = store.getFeatureSource().getFeatures().features();
			} catch (MalformedURLException murle) {
				throw new FulminantError("Shape file url is invalid: " + murle);
			} catch (IOException ioe) {
				throw new FulminantError("Error opening shape file: " + ioe);
			}
			
			Map<Integer,Map<String,Object>> farmsAttributes = new HashMap<Integer,Map<String,Object>>();
			while (featureIterator.hasNext()) {
				Feature feature = featureIterator.next();
				Integer farmId = null;
				try {
					Double dFarmId = (Double) ((feature.getAttribute("farmId") != null) ? feature.getAttribute("farmId") : feature.getAttribute("farmID"));					
					farmId = new Integer(dFarmId.intValue());
					
					Map<String,Object> farmAttributes = new HashMap<String,Object>();
					farmAttributes.put("gisFarmId", new Integer((String) feature.getAttribute("ID")));
					farmAttributes.put("location", feature.getDefaultGeometry());
					farmsAttributes.put(farmId, farmAttributes);
				} catch (NumberFormatException nfe) {
					throw new FulminantError("Error al obtener el farmId del shape file: " + nfe);
				}
			}
			
			for (Farm farm: farms) {
				Map<String,Object> farmAttributes = farmsAttributes.get(farm.getFarmId());
				
				if (farmAttributes != null) {
					Geometry location = (Geometry) farmAttributes.get("location");
					this.farmingRegions.move(farm, location);
					farm.setGISFarmId((Integer) farmAttributes.get("gisFarmId"));
				} else {
					throw new FulminantError("El campo " + farm.getFarmId() + " no se encuentra representado dentro del espacio de GIS");
				}
			}
		} else {
			throw new FulminantError("GIS File does not exist: " + this.shapeFile.getAbsolutePath());
		}
		
		/*
		 *  Crear la network haciendo queries del espacio GIS. Tener en cuenta que un campo A
		 *  puede no considerar vecino a un campo B, pero si que B consedere vecino a A. Esto
		 *  es geometricamente posible por el tipo de query que se ejecuta. Un ejemplo trivial seria
		 *
		 *  A = campo con forma circular con centroide en (x,y) y radio = Rmax (radio maximo para buscar campos vecinos
		 *  B = campo con forma circular con centroide en (x + Rmax + delta, y) y radio Rmax/2
		 *  
		 *  Como se ve, A no considera vecino a B porque su radio es igual al radio maximo. Sin embargo,
		 *  B si considera vecino a A.
		 *  
		 *  En estos casos, al ser la red no-orientada, se considera que A y B son vecinos en forma bilateral.
		 */
		Iterable<Agent> agents = this.farmingRegions.getAllObjects();
		List<Farm> isolatedFarms = new ArrayList<Farm>();
		for (Agent agent: agents) {
			if (agent instanceof Farm) {
				Farm source = (Farm) agent;
				GeographyWithin<Agent> query = new GeographyWithin<Agent>(this.farmingRegions, maxNeighbourDistance, source);
				Iterator<Agent> neighbours = query.query().iterator();
				
				if (neighbours.hasNext()) {
					while (neighbours.hasNext()) {
						Agent neighbour = neighbours.next();
						if (neighbour instanceof Farm) {
							Farm target = (Farm) neighbour;
							if (isolatedFarms.contains(target)) {
								// Ahora si tiene al menos un vecino
								isolatedFarms.remove(target);
							}
							RepastEdge<Agent> edge = new RepastEdge<Agent>(source, target, false);
							if (! neighbourhood.containsEdge(edge)) {
								neighbourhood.addEdge(edge);
							}
						}
					}
				} else {
					Iterator<Agent> currentNeighbours = neighbourhood.getAdjacent(source).iterator();
					if (! currentNeighbours.hasNext()) {
						// No tiene vecinos que el pueda ver ni tampoco otros que lo hayan agregado a el como vecino
						isolatedFarms.add(source);
					}
				}
			}
		}
		if (! isolatedFarms.isEmpty()) {
			throw new FulminantError("Los campos " + isolatedFarms + " no tienen vecinos en " + (maxNeighbourDistance / 1000.0) + " km a la redonda");
		}
		
		/*
		double minDegreeFarmSize = 0.0;
		double maxDegreeFarmSize = 0.0;
		double minDegree = Double.MAX_VALUE;
		double maxDegree = 0.0;
		double degreeSum = 0.0;
		int numberOfAgents = 0;
		Iterable<Agent> nodes = neighbourhood.getNodes();
		for (Agent agent: nodes) {
			if (agent instanceof Farm) {
				Farm farm = (Farm) agent;
				double degree = neighbourhood.getDegree(agent);
				
				degreeSum += degree;
				if (degree < minDegree) {
					minDegree = degree;
					minDegreeFarmSize = farm.getFarmSize();
				}
				if (degree > maxDegree) {
					maxDegree = degree;
					maxDegreeFarmSize = farm.getFarmSize();
				}
				
				numberOfAgents++;
			}
		}
		java.lang.System.err.println("avg degree = " + (degreeSum / (double) numberOfAgents) + "; min degree = " + minDegree + "(" + minDegreeFarmSize + " has); max degree = " + maxDegree + "(" + maxDegreeFarmSize + " has)");
		*/
				
		return neighbourhood;
	}
	
	public void exportFarmsLandUse(File baseDir) {
		// 1. Crear FeatureType para archivo DBF
		FeatureType landUseFeatureType = null;
		try {
			landUseFeatureType = DataUtilities.createType("LandUse", "location:Geometry,ID:Integer,farmId:Integer,ampId:Integer");
		} catch (SchemaException se) {
			throw new FulminantError("Error al crear el esquema de datos para el dbf: " + se);
		}
		
		// 2. Buscar los campos guardar el uso de cada uno de ellos
		AMPSortedTable ampSortedTable = Manager.getInstance().getAMPSortedTable();
		List<Feature> landUseFeatures = new ArrayList<Feature>();
		for (Agent agent: this.farmingRegions.getAllObjects()) {
			if (agent instanceof Farm) {
				Farm farm = (Farm) agent;
				Geometry farmGeometry = this.farmingRegions.getGeometry(farm);
				Object attributes[] = { farmGeometry, farm.getGISFarmId(), farm.getFarmId(), ampSortedTable.getActivityManagementProportionId(farm.getCurrentActivityManagementProportion()) };
				try {
					Feature feature = landUseFeatureType.create(attributes, farm.getFarmId().toString());
					feature.setDefaultGeometry(farmGeometry);
					landUseFeatures.add(feature);
				} catch (IllegalAttributeException iae) {
					iae.printStackTrace(java.lang.System.err);
					throw new FulminantError("Illegal attributes: " + iae);
				}
			} else {
				throw new FulminantError("Agent is not a Farm instance: " + agent);
			}
		}
		DefaultFeatureCollection landUseFeatureCollection = (DefaultFeatureCollection) DefaultFeatureCollections.newCollection();
    	landUseFeatureCollection.addAll(landUseFeatures);
        		
		// 3. Crear el ShapeFile a partir de los features    	
		File newShapeFile = new File(baseDir, "farming_regions.shp");
        ShapefileDataStoreFactory dataStoreFactory = new ShapefileDataStoreFactory();
        Map<String, Serializable> params = new HashMap<String, Serializable>();
        try {
        	params.put("url", newShapeFile.toURI().toURL());
        	params.put("create spatial index", Boolean.TRUE);
        	ShapefileDataStore newDataStore = (ShapefileDataStore) dataStoreFactory.createNewDataStore(params);
        	newDataStore.createSchema(landUseFeatureType);
        	        	        	
        	Transaction transaction = new DefaultTransaction();     
        	String typeName = newDataStore.getTypeNames()[0];
        	FeatureStore featureStore = (FeatureStore) newDataStore.getFeatureSource(typeName);
        	featureStore.addFeatures(landUseFeatureCollection);
        	newDataStore.forceSchemaCRS(this.farmingRegions.getCRS());      	
        	
        	transaction.commit();
            transaction.close();
        } catch (IOException ioe) {
        	throw new FulminantError("Error writing shapefile: " + ioe);
        }
	}
}