package fiuba.lamm.agromodel.tests;

import au.com.bytecode.opencsv.*;
import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.decision.*;
import fiuba.lamm.agromodel.factories.*;
import fiuba.lamm.agromodel.math.*;
import fiuba.lamm.agromodel.utils.*;
import java.io.*;
import java.util.*;
import org.junit.*;
import static org.junit.Assert.*;

public class AMPSortedTableTest extends BaseTest {
	protected Integer numberOfPlots;
	protected List<Region> regions;
	protected Map<Region,List<ActivityManagement>> ams; 
	protected AMPSortedTable table;
	
	@BeforeClass
	public static void init() {
		fiuba.lamm.agromodel.utils.System.init();
	}
	
	@Before
	public void setUp() throws Exception {
		this.numberOfPlots = 6; // Default number of plots per farm
		this.regions       = RegionFactory.getInstance().findAll();		
		this.ams           = new HashMap<Region,List<ActivityManagement>>();		
		for (Region region: this.regions) {
			List<ActivityManagement> regionAms = ActivityManagementFactory.getInstance(null).findByRegion(region);
			this.ams.put(region, regionAms);
		}
		
		this.table = new AMPSortedTable(this.numberOfPlots);	
		this.table.build();
	}

	@After
	public void tearDown() throws Exception {
		this.numberOfPlots = null;
		this.regions       = null;
		this.ams           = null;
		this.table         = null;
	}

	@Test
	public void testTableSize() {		
		NaturalFunction factorialFunction = new FactorialFunction();		
		Integer expectedSize = 0;		
		for (Region region: this.regions) {
			List<ActivityManagement> regionAms = this.ams.get(region);
			Integer k = numberOfPlots; // Numeros de objetos a repartir
			Integer n = regionAms.size(); // Numeros de "personas" entre las cuales repartir los objetos
			expectedSize += factorialFunction.apply(n + k - 1) / (factorialFunction.apply(k) * factorialFunction.apply(n - 1));
		}
		
		List<ActivityManagementProportion> amps = this.table.getActivityManagementProportions();		
		assertEquals("Table size mismatch", expectedSize.longValue(), amps.size());		
	}	
	
	@Test
	public void testTableIds() {
		// 1. Open test directory
		try {
			File workingDirectory = getTestWorkingDirectory();
			
			// 2. Check each region
			boolean checked = false;
			for (Region region: this.regions) {
				// 2a. If file does not exist, skip region
				File csvFile = new File(workingDirectory, region.getRegionId() + ".txt");
				if (csvFile.isFile()) {
					checked = true;
					
					// 2b. Read CSV file
					CSVReader reader = new CSVReader(new FileReader(csvFile));
					List<?> lines = reader.readAll();
					
					// 2c. Check header
					List<ActivityManagement> regionAms = this.ams.get(region);
					String header[] = (String[]) lines.remove(0);					
					assertEquals("File for region " + region.getRegionName() + " has and invalid number of columns (" + header.length + ")", regionAms.size() + 1, header.length);
					for (int i = 1; i < header.length; i++) {
						assertEquals("Column mismatch for region " + region.getRegionName(), regionAms.get(i-1).getActivityMgmtName(), header[i]);
					}
					
					// 2d. Check data
					Map<String,ActivityManagement> amLookupTable = new HashMap<String,ActivityManagement>();
					for (ActivityManagement am: regionAms) {
						amLookupTable.put(am.getActivityMgmtName(), am);
					}
					for (Object o: lines) {
						String data[] = (String[]) o;
						ActivityManagementProportion amp = new ActivityManagementProportion();
						
						Integer id = new Integer(data[0]);
						for (int i = 1; i < header.length; i++) {
							Double proportion = new Double(data[i]);
							Integer amPlots = (int) Math.round(proportion * this.numberOfPlots.doubleValue());
							if (amPlots > 0) {
								ActivityManagement am = amLookupTable.get(header[i]);
								amp.setProportion(am, amPlots);
							}
						}
						
						Integer tableAmpId = this.table.getActivityManagementProportionId(amp);
						assertNotNull("AMP not found (" + amp + ")", tableAmpId);
						assertEquals("Id mismatch for AMP " + amp, tableAmpId, id);
					}
				}
			}
			
			// 3. Check that at least one region had been checked
			assertTrue("No checked regions", checked);
		} catch (IOException ioe) {
			fail(ioe.getMessage());
		}		
	}
}