package fiuba.lamm.agromodel.utils;

import fiuba.lamm.agromodel.*;
import fiuba.lamm.agromodel.decision.*;
import fiuba.lamm.agromodel.execution.*;
import fiuba.lamm.agromodel.factories.*;
import java.util.*;

public class AMPSortedTable implements Comparator<ActivityManagementProportion> {
	protected Integer numberOfPlots;
	protected List<Region> regions;
	protected Map<Region,List<ActivityManagement>> ams;
	protected Map<ActivityManagementProportion,Integer> idLookupTable;
	protected Map<Integer,ActivityManagementProportion> ampLookupTable;
	
	public AMPSortedTable(Scenario scenario) {
		this.numberOfPlots = (Integer) scenario.getParameter("Farm.NumberOfPlots");
		this.regions = RegionFactory.getInstance().findAll();
		Collections.sort(this.regions); // Sort regions
		
		this.ams     = new HashMap<Region,List<ActivityManagement>>();
		for (Region region: this.regions) {
			List<ActivityManagement> regionAms = ActivityManagementFactory.getInstance(scenario).findByRegion(region);
			Collections.sort(regionAms); // Sort activity-managements
			this.ams.put(region, regionAms);
		}
		
		this.idLookupTable  = new HashMap<ActivityManagementProportion,Integer>();
		this.ampLookupTable = new TreeMap<Integer,ActivityManagementProportion>();
	}
	
	public AMPSortedTable(Integer numberOfPlots) {
		this.numberOfPlots = numberOfPlots;
		this.regions = RegionFactory.getInstance().findAll();
		Collections.sort(this.regions); // Sort regions
		
		this.ams     = new HashMap<Region,List<ActivityManagement>>();		
		for (Region region: this.regions) {
			List<ActivityManagement> regionAms = ActivityManagementFactory.getInstance(null).findByRegion(region);
			Collections.sort(regionAms); // Sort activity-managements
			this.ams.put(region, regionAms);
		}
		
		this.idLookupTable  = new HashMap<ActivityManagementProportion,Integer>();
		this.ampLookupTable = new TreeMap<Integer,ActivityManagementProportion>();
	}
	
	public void build() {
		// 1. Find AMPs for each region and sort them. Regions array is already sorted.	
		List<ActivityManagementProportion> amps = new ArrayList<ActivityManagementProportion>();
		for (Region region: this.regions) {
			List<ActivityManagementProportion> regionAmps = ActivityManagementProportion.getPossibleProportions(this.ams.get(region), this.numberOfPlots);
			Collections.sort(regionAmps, this);
			amps.addAll(regionAmps);
		}
		
		// 2. Build table
		this.idLookupTable.clear();
		this.ampLookupTable.clear();
		Integer index = 1;
		for (ActivityManagementProportion amp: amps) {
			idLookupTable.put(amp, index);
			ampLookupTable.put(index, amp);
			index++;
		}
	}
	
	public Integer getActivityManagementProportionId(ActivityManagementProportion amp) {
		return this.idLookupTable.get(amp);
	}
	
	public ActivityManagementProportion getActivityManagementProportion(Integer ampId) {
		return this.ampLookupTable.get(ampId);
	}
	
	public List<Integer> getActivityManagementProportionIds() {
		return new ArrayList<Integer>(this.idLookupTable.values());
	}
	
	public List<ActivityManagementProportion> getActivityManagementProportions() {
		return new ArrayList<ActivityManagementProportion>(this.ampLookupTable.values());
	}
	
	public List<ActivityManagementProportion> getActivityManagementProportions(Region region) {
		List<ActivityManagement> regionAms = this.ams.get(region);
		List<ActivityManagementProportion> regionAmps = new ArrayList<ActivityManagementProportion>();
		
		for (ActivityManagementProportion amp: this.ampLookupTable.values()) {
			if (amp.containsAny(regionAms)) {
				regionAmps.add(amp);
			}
		}
		
		return regionAmps;
	}
	
	public int compare(ActivityManagement am1, ActivityManagement am2) {
		Region r1 = am1.getRegion();
		Region r2 = am2.getRegion();
		
		if (r1.equals(r2)) {
			return am1.getActivityMgmtId().compareTo(am2.getActivityMgmtId());
		} else {
			return r1.getRegionId().compareTo(r2.getRegionId());
		}
	}
	
	public int compare(ActivityManagementProportion amp1, ActivityManagementProportion amp2) {
		// Both AMPs must belong to the same region 
		Region r1 = amp1.getAssociatedRegion();
		Region r2 = amp2.getAssociatedRegion();
		
		if (r1.equals(r2)) {
			List<ActivityManagement> regionAms = this.ams.get(r1);
			int result = 0;
			int i      = 0;
			
			while ((i < regionAms.size()) && (result == 0)) {
				ActivityManagement am = regionAms.get(i);				
				Integer p1 = amp1.getProportion(am);
				Integer p2 = amp2.getProportion(am);
				result     = p1.compareTo(p2);
				
				i++;
			}
			
			return result;
		} else {
			throw new ClassCastException("AMP1 (" + amp1 + ") does not belong to the same region of AMP2(" + amp2 + ")");
		}
	}	
}